
from pipedream.asm.allocator import *
from pipedream.asm.ir        import *
from pipedream.utils         import *
from pipedream.utils         import abc
import collections
import typing as ty


__all__ = [
  'Register_Allocator_Base',
]


class Register_Allocator_Base(Register_Allocator):
  """
    Helper class for implementing register allocators
  """

  FREE  = 0
  TAKEN = Sentinel('REGISTER_IS_TAKEN')

  """LValue allocator that manages a fixed, finite set of registers."""

  def __init__(self, all_regs: [Register], callee_save_registers: [Register]):
    self._regs = collections.OrderedDict()

    for r in all_regs:
      assert isinstance(r, Register), r
      self._regs[r] = self.FREE

    for r in callee_save_registers:
      self.take(r)

  @classmethod
  def from_register_set(clss, register_set: 'Register_Set'):
    return clss(register_set.all_registers(), register_set.callee_save_registers())

  @classmethod
  def from_registers(clss, regs: [Register]):
    return clss(regs, [])

  ### QUERY REGISTER STATUS

  def is_free(self, reg: Register) -> bool:
    return self._regs[reg] == self.FREE

  @abc.override
  def is_taken(self, reg: Register) -> bool:
    return self._regs[reg] != self.FREE

  @abc.override
  def available(self, ty) -> bool:
    """Query if there is a free register of type *ty*."""

    for reg in self.free_registers(ty):
      # there is a free register with type *ty*
      return True

    return False

  ### CONTAINER OF REGISTERS

  def registers(self, clss: Register_Class) -> ty.Sequence[Register]:
    """
      Iterate over all registers in register class *clss* managed by this
      allocator.
    """

    for reg in self._regs:
      if reg in clss:
        yield reg

  def free_registers(self, clss: Register_Class = None) -> ty.Sequence[Register]:
    """Iterate over all free registers of class *clss*."""

    for reg, status in self._regs.items():
      if status == self.FREE and (clss is None or reg in clss):
        yield reg

  @abc.override
  def __contains__(self, reg) -> bool:
    return reg in self._regs

  ### ALLOCATE

  @abc.override
  def allocate(self, reg_class: Register_Class) -> Register:
    assert type(reg_class) is Register_Class

    for reg in self.free_registers(reg_class):
       self.take(reg)
       return reg

    raise AllocationError('No location of type ' + reg_class.name + ' available')

  @abc.override
  def take(self, reg: Register):
    if self.is_taken(reg):
      raise AllocationError(repr(reg) + ' is not free')

    for alias in reg.aliases:
      if self._regs[alias] == self.TAKEN:
        raise AllocationError(repr(reg) + ' is already taken (via ' + repr(alias) + ')')

    self._regs[reg] = self.TAKEN
    for alias in reg.aliases:
      self._regs[alias] += 1

  @abc.override
  def free(self, reg: Register):
    if self.is_free(reg):
      raise AllocationError(repr(reg) + ' is already free')

    self._regs[reg] = self.FREE
    for alias in reg.aliases:
      oldstatus = self._regs[alias]

      assert oldstatus != self.TAKEN
      assert oldstatus >= self.FREE

      if oldstatus != self.FREE:
        self._regs[alias] = oldstatus - 1
