
import collections
import enum
import functools
import types
import typing as ty

from pipedream.utils import abc


__all__ = [
  'Architecture',

  'Instruction_Set',
  'Register_Set',

  'IR_Builder',
  'Register_Allocator',

  'ASM_Dialect',

  'Instruction',
  'Machine_Instruction',

  'Operand',
  'Use_Def',
  'Operand_Visibility',
  'Register_Operand',
  'Immediate_Operand',
  'Memory_Operand',
  'Address_Operand',
  'Base_Displacement_Operand',
  'Base_Displacement_Address_Operand',
  'Base_Displacement_Memory_Operand',

  'Register',
  'Register_Class',

  'Label',
  'Loop',
]


class Architecture(abc.ABC):
  @staticmethod
  def for_name(name: str) -> 'Architecture':
    import pipedream.asm.x86

    if not hasattr(Architecture, '_REGISTRY_'):
      X86 = pipedream.asm.x86.X86_Architecture()

      Architecture._REGISTRY_ = types.MappingProxyType({
        X86.name: X86
      })

    try:
      return Architecture._REGISTRY_[name]
    except KeyError:
      raise ValueError('unknown arch ' + repr(name))

  @abc.abstractproperty
  def name(self) -> str:
    ...

  @abc.abstractmethod
  def instruction_set(self) -> 'Instruction_Set':
    ...

  @abc.abstractmethod
  def register_set(self) -> 'Register_Set':
    ...

  @abc.abstractmethod
  def asm_dialects(self) -> ['Asm_Dialect']:
    ...

  @abc.abstractmethod
  def make_asm_writer(self, dialect: 'ASM_Dialect', file: ty.IO[str]) -> 'Asm_Writer':
    return NotImplemented

  @abc.abstractmethod
  def make_ir_builder(self) -> 'IR_Builder':
    ...

  @abc.abstractmethod
  def make_register_allocator(self) -> 'Register_Allocator':
    ...

  @abc.abstractmethod
  def loop_overhead(self, num_iterations: int) -> 'Loop_Overhead':
    """
      Returns the number of instructions executed by *num_iterations* iterations of a `loop`.
    """
    # FIXME: this is an ugly wart. We need to know this to find appropriate
    #        iteration numbers for a benchmark kernel.
    #        We could move this to IR_Builder, but then it would be its
    #        responsibility to create Benchmark_Spec instances,
    #        and I would like Benchmark_Spec to be seperate.


class IR_Builder(abc.ABC):
  """
    Helper class for building IR.
  """

  #### generate operands

  @abc.abstractmethod
  def get_return_register(self) -> 'Register':
    """
      Get the register for returning integers from functions on this architecture.
    """

  @abc.abstractmethod
  def get_argument_register(self, idx: int) -> 'Register':
    """
      Get the Nth (zero based indexing) argument register for function calls.
      Raises an exception of there is no Nth arg register on this architecture.
    """

  @abc.abstractmethod
  def get_scratch_register(self, idx: int) -> 'Register':
    """
      Get the Nth scratch register code can use in benchmark kernels.
      Every architecure must provide at least three such registers.
    """

  @abc.abstractmethod
  def select_memory_base_register(self, insts: ty.List['Instruction'], free_regs: ty.Set['Register'],
                                  address_width: int) -> 'Register':
    """
      Select a register that can be used as a memory base register for all memory accesses
      in the given list of instructions :insts:.
    """

  @abc.abstractmethod
  def preallocate_benchmark(self, alloc: 'Register_Allocator',
                            instructions: ty.Sequence['Instruction']) -> ty.Tuple[object,
                                                                                  ty.List['Instruction']]:
    """
      Fill in some operands of benchmark kernel if required by architecture constraints.
      Returns a state object that must be passed to :free_stolen_benchmark_registers():
      to free them again.
    """

  @abc.abstractmethod
  def free_stolen_benchmark_registers(self, alloc: 'Register_Allocator', stolen_regs: object):
    """
      Free registers taken by :steal_benchmark_registers():
    """

  #### low-level IR generation

  @abc.abstractmethod
  def emit_benchmark_prologue(kernel: ty.Sequence['Instruction'],
                              free_regs: ty.Sequence['Register'],
                              ) -> ty.List['Instruction']:
    """
      Emit code to initialize registers for benchmark kernel.
    """

  @abc.abstractmethod
  def emit_dependency_breaker(self, reg: 'Register') -> ty.List['Instruction']:
    """
      Emit code to break data dependencies along register :reg:.
      The contents of :reg: are undefined after the depdency breaker code executes.

      If the architecture does not support dependency breakers an empty list is returned.
    """

  @abc.abstractmethod
  def emit_sequentialize_cpu(self, alloc: 'Register_Allocator') -> ty.List['Instruction']:
    """Emit instructions that sequentialize the CPU."""

  @abc.abstractmethod
  def emit_branch_if_not_zero(self, reg: 'Register', dst: 'Label') -> ty.List['Instruction']:
    """
      emit code to branch to label :dst: if register :reg: does not contains zero.
    """

  @abc.abstractmethod
  def emit_branch_if_zero(self, reg: 'Register', dst: 'Label') -> ty.List['Instruction']:
    """
      emit code to branch to label :dst: if register :reg: contains zero.
    """

  @abc.abstractmethod
  def emit_copy(self, src: 'Register', dst: 'Register') -> ty.List['Instruction']:
    """
      emit register to register copy.
    """

  @abc.abstractmethod
  def emit_push_to_stack(self, src: 'Register') -> ty.List['Instruction']:
    """
      emit code to push :src: to stack.
    """

  @abc.abstractmethod
  def emit_call(self, dst: 'Label') -> ty.List['Instruction']:
    """
      emit code to call the procedure given by :dst:.
    """

  @abc.abstractmethod
  def emit_return(self, reg: 'Register') -> ty.List['Instruction']:
    """
      emit code to return the value in :reg: from the current procedure.

      TODO: add helper to gen proc-enter to push link register for ARM.
    """

  @abc.abstractmethod
  def emit_put_const_in_register(self, const: int, reg: 'Register') -> ty.List['Instruction']:
    """
      emit code to put value :const: in register :reg:.
    """

  @abc.abstractmethod
  def emit_mul_reg_const(self, reg: 'Register', const: int) -> ty.List['Instruction']:
    """
      emit code to multiply the contents of a register by a constant.
      The result is stored in :reg:.
    """

  @abc.abstractmethod
  def emit_add_registers(self, src_reg: 'Register', src_dst_reg: 'Register') -> ty.List['Instruction']:
    """
      emit code to add contents of two registers.
      The result is stored in the second register :src_dst_reg:.
    """

  @abc.abstractmethod
  def emit_substract_one_from_reg_and_branch_if_not_zero(self, loop_counter: 'Register',
                                                         dst: 'Label') -> ty.List['Instruction']:
    """
      emit code to substract one from the contents of register :loop_counter:
      and store the result in :loop_counter:.
      Then branches to :dst: if :loop_counter: now contains zero.
    """


class ASM_Dialect:
  pass


class Register_Allocator:
  @abc.abstractmethod
  def take(self, reg: 'Register'):
    ...

  @abc.abstractmethod
  def free(self, reg: 'Register'):
    ...

  @abc.abstractmethod
  def is_free(self, reg: 'Register') -> bool:
    ...

  @abc.abstractmethod
  def sub_allocator(self) -> 'Register_Allocator':
    """
      Creates a new Register_Allocator that can only use the registers used
      that are currently free in :self:.
      The sub-allocator cannot allocate any arguments.
      take() and free() are not synchronized across the allocator and its
      sub-allocators.
    """
    ...


class Instruction_Set(abc.ABC):
  @abc.abstractmethod
  def instruction_groups(self) -> ['Instruction_Group']:
    ...

  @abc.abstractmethod
  def instructions(self) -> ['Instruction']:
    ...

  @abc.abstractmethod
  def benchmark_instructions(self) -> ['Instruction']:
    """
      List of instructions intended for benchmarking.
      Subset of self.instructions.
    """

  @abc.abstractmethod
  def instruction_for_name(self, name: str) -> 'Instruction':
    ...

  @abc.abstractmethod
  def instructions_for_tags(self, tags: str):
    ...

  def __iter__(self):
    return iter(self.instructions)


class Register_Set(abc.ABC):
  @abc.abstractmethod
  def stack_pointer_register(self) -> 'Register':
    ...

  @abc.abstractmethod
  def register_classes(self) -> ['Register_Class']:
    ...

  @abc.abstractmethod
  def all_registers(self) -> 'Register_Class':
    ...

  @abc.abstractmethod
  def argument_registers(self) -> 'Register_Class':
    ...

  @abc.abstractmethod
  def callee_save_registers(self) -> 'Register_Class':
    ...


class Asm_Dialect:
  """
    Defines how instructions and operands are printed.
  """

  @abc.abstractproperty
  def name(self) -> str:
    ...

  @abc.abstractmethod
  def instruction_mnemonic(self, inst: 'Instruction') -> str:
    ...

  @abc.abstractmethod
  def instruction_to_str(self, inst: 'Instruction') -> str:
    ...

  @abc.abstractmethod
  def operand_to_str(self, op: 'Operand') -> str:
    ...


class Instruction_Group:
  """
    Groups instructions that do `the same thing`.
    Can be more or less abstract.
    One instruction ca be in many groups.
  """

  @abc.abstractproperty
  def name(self) -> str:
    ...

  @abc.abstractmethod
  def instructions(self) -> ['Instruction']:
    ...


@functools.total_ordering
class Register_Class(abc.ABC):
  """
    A finite non-empty frozen set of machine registers.
    Two registers classes can overlap.
    If a register class is compleletely contained in another one it is a
    sub register class.
    The register class containment relation forms a DAG.

    Note that a register can be in many register classes.
  """

  def __init__(self, name, registers, register_aliases = {}):
    self._name        = name
    self._reg_iter    = tuple(registers)
    self._reg_set     = frozenset(self._reg_iter)
    self._reg_aliases = dict(register_aliases)

    assert all(isinstance(r, Register) for r in self._reg_iter)

  @property
  def name(self) -> str:
    return self._name

  def __getattr__(self, reg):
    """allow accessing registers as attributes of a register class."""

    try:
      return self._reg_aliases[reg]
    except KeyError:
      raise AttributeError(reg)

  ### treat register classes like sets

  def __contains__(self, value: 'Register') -> bool:
    return value in self._reg_set

  def __or__(self, that: 'Register_Class') -> 'Register_Class':
    if type(self) != type(that):
      return NotImplemented

    set  = self._reg_set  | that._reg_set
    iter = self._reg_iter + that._reg_iter

    return Register_Class(
      name             = self.name + '|' + that.name,
      registers        = tuple(t for t in iter if t in set),
      register_aliases = self._alias_merge(set, self._reg_aliases, that._reg_aliases),
    )

  def __and__(self, that: 'Register_Class') -> 'Register_Class':
    if type(self) != type(that):
      return NotImplemented

    set  = self._reg_set  & that._reg_set
    iter = self._reg_iter + that._reg_iter

    return Register_Class(
      name             = self.name + '&' + that.name,
      registers        = tuple(t for t in iter if t in set),
      register_aliases = self._alias_merge(set, self._reg_aliases, that._reg_aliases),
    )

  def __eq__(self, that):
    if type(self) != type(that):
      return NotImplemented

    return self._reg_set == that._reg_set

  def __lt__(self, that):
    if type(self) != type(that):
      return NotImplemented

    return self._reg_set < that._reg_set

  def __le__(self, that):
    if type(self) != type(that):
      return NotImplemented

    return self._reg_set <= that._reg_set

  def __hash__(self):
    return hash(self._reg_set)

  ### iteration

  def __iter__(self) -> ty.Sequence['Register']:
    return iter(self._reg_iter)

  def __getitem__(self, idx) -> 'Register':
    return self._reg_iter[idx]

  def __len__(self) -> int:
    return len(self._reg_iter)

  ### MISC

  def __str__(self):
    return type(self).__name__ + '(' + repr(self.name) + ')'

  ### private

  @staticmethod
  def _alias_merge(reg_set: ty.Set['Register'], alias1, alias2) -> ty.Dict[str, 'Register']:
    out = {}
    out.update({alias: reg for alias, reg in alias1.items() if reg in reg_set})
    out.update({alias: reg for alias, reg in alias2.items() if reg in reg_set})
    return out


class Register(abc.ABC):
  """A machine register."""

  @abc.abstractproperty
  def name(self) -> str:
    ...

  @abc.abstractproperty
  def width(self) -> int:
    """
      Width of register in bits
    """

  @abc.abstractmethod
  def as_width(self, bits: int) -> 'Register':
    """
      Get alias with a given width.
    """
    return NotImplemented

  @abc.abstractproperty
  def widest(self) -> 'X86_Register':
    """
      Get widest alias.
    """
    return NotImplemented

  @property
  @abc.abstractmethod
  def sub_registers(self):
    """An iterator over the registers immediate sub-registers."""

  @property
  @abc.abstractmethod
  def super_registers(self):
    """An iterator over the registers immediate super-registers."""

  @property
  @abc.abstractmethod
  def aliases(self):
    """
      An iterator over the registers that alias this register.
      I.e. an iterator of all registers that share storage/are affected by
      stores to this register (and vice versa).
    """

  @property
  def all_sub_registers(self):
    """An iterator over all sub-registers."""

    for reg in self.sub_registers:
      yield reg
      yield from reg.all_sub_registers

  @property
  def all_super_registers(self):
    """An iterator over all super-registers."""

    for reg in self.super_registers:
      yield reg
      yield from reg.all_super_registers


class Instruction(abc.ABC):
  """
    Descriptor for one instruction.
  """

  @abc.abstractmethod
  def name(self) -> str:
    ...

  @abc.abstractproperty
  def tags(self) -> ty.Sequence[str]:
    ...

  @abc.abstractproperty
  def operands(self) -> ['Operand']:
    ...

  @property
  def defs(self) -> ['Operand']:
    """ Operands written to """
    for op in self.operands:
      if op.is_def:
        yield op

  @property
  def uses(self) -> ['Operand']:
    """ Operands read from """
    for op in self.operands:
      if op.is_use:
        yield op

  def get_operand(self, name: str) -> 'Operand':
    for op in self.operands:
      if op.name == name:
        return op

    raise KeyError(name)

  def get_operand_idx(self, name: str) -> 'Operand':
    for idx, op in enumerate(self.operands):
      if op.name == name:
        return idx

    raise KeyError(name)

  @abc.abstractmethod
  def update_operand(self, idx_or_name: ty.Union[int, str], fn: ty.Callable[['Operand'], 'Operand']) -> 'Instruction':
    """
      Create new instruction where the operand indicated by :idx_or_name:
      has been replaced with the result of calling :fn: with the original operand.
    """

  def has_memory_operand(self) -> bool:
    for op in self.operands:
      if isinstance(op, Memory_Operand):
        return True
    return False

  @abc.abstractmethod
  def encodings(self) -> ['Instruction_Encoding']:
    """
      List of ways that an instruction can be encoded.
    """

  @abc.abstractproperty
  def can_benchmark(self) -> bool:
    """
      Can this instruction occur in a benchmark kernel?
    """

  def __repr__(self):
    txt = self.name

    if self.operands:
      txt += ' ' + ', '.join(map(str, self.operands))

    return txt


class Pseudo_Instruction(Instruction):
  """
    Helper pseudo instruction that is not directly encodable.
    Must be lowered to zero or more real machine instructions.
  """

  def encodings(self):
    return []

  @property
  @abc.override
  def can_benchmark(self) -> bool:
    return False


class Machine_Instruction(Instruction):
  """
    Real machine instruction that can be encoded to binary.
  """


class Instruction_Encoding:
  """
    A concrete encoding for an instruction, a pattern of bytes.
    All encodings for an instruction behave semantically the same, but
    have a different byte pattern.
  """

  def operands(self) -> ['Operand']:
    pass

  def encode(self) -> bytes:
    ...


class Use_Def(enum.IntFlag):
  USE     = 1
  DEF     = 2
  USE_DEF = USE | DEF

  def repr_name(self):
    return {self.USE: '?', self.DEF: '!', self.USE_DEF: '?!'}[self]


class Operand_Visibility(enum.IntFlag):
  ## operand shown in asm & instruction bits
  EXPLICIT   = 0
  ## operand shown in asm, implicit in instruction bits
  IMPLICIT   = 1
  ## operand not show in asm, implicit in instruction bits
  SUPPRESSED = 2


class Operand(abc.ABC):
  @abc.abstractproperty
  def name(self) -> str:
    ...

  @abc.abstractproperty
  def use_def(self) -> Use_Def:
    ...

  @abc.abstractproperty
  def visibility(self) -> Operand_Visibility:
    ...

  @property
  def is_use(self) -> bool:
    return bool(self.use_def & Use_Def.USE)

  @property
  def is_def(self) -> bool:
    return bool(self.use_def & Use_Def.DEF)

  @abc.abstractproperty
  def is_virtual(self) -> bool:
    """
      An operand is 'virtual' if it is not fully specified yet.
      I.e. a Register_Operand where 'register' returns None or a
      Immediate_Operand where 'value' returns None.
    """


class Register_Operand(Operand):
  @abc.abstractproperty
  def register_class(self) -> Register_Class:
    ...

  @abc.abstractproperty
  def register(self) -> ty.Optional[Register]:
    """
      Return register read/written by this operand.
      If the returned value is none no register has been assigned yet.
      I.e. this is a virtual register that still needs to be allocated.
    """

  @abc.abstractmethod
  def with_register(self, reg: Register) -> 'Register_Operand':
    """
      Return a copy of :self: where the register read/written by this operand
      has been changed to :reg:.
      If :reg: is not a member of self.register_class a TypeError is raised.
    """

  @property
  @abc.override
  def is_virtual(self) -> bool:
    return self.register is None

  def __repr__(self):
    txt = self.name + ":"

    if self.register:
      txt += self.register.name
    else:
      if type(self.register_class.name) is not str:
        print('!', type(self), self.name, self.register_class, self.register_class.name)
      txt += self.register_class.name

    txt += self.use_def.repr_name()
    return txt


Immediate_Value = ty.Union[int, float]


class Immediate_Operand(Operand):
  """
    Constant operand encoded directly into the instruction.
  """

  @abc.abstractproperty
  def num_bits(self) -> int:
    """
      Number of bits used to represent immediate
    """

  @property
  @abc.override
  @abc.final
  def use_def(self) -> Use_Def:
    return Use_Def.USE

  @abc.abstractproperty
  def value(self) -> ty.Optional[Immediate_Value]:
    ...

  @abc.abstractmethod
  def with_value(self, val: Immediate_Value) -> 'Immediate_Operand':
    """
      Return a copy of :self: where value has been changed to :val:.
      If :val: is not a valid constant for self a TypeError is raised.
    """

  @property
  @abc.override
  def is_virtual(self) -> bool:
    return self.value is None

  def __repr__(self):
    if self.value is not None:
      txt = str(self.value)
    else:
      txt = type(self).__name__

    return self.name + ':' + txt + '?'


class Flags_Operand(Operand):
  """
    Read/Write to a flags register.
    Since reads/writes to flags are usually partial & highly optimized
    we don't count it as a normal register operand.
    I.e. if instruction A writes flag FA and instruction B reads flag FB
    there is usually no data dependence between the two.
    If we treated this like a normal read/write to the FLAGS register we'd get
    wrong dependencies.

    TODO: put the flags_read/flags_written properties on the instruction?
  """

  @abc.abstractproperty
  def register(self) -> 'Register':
    return NotImplemented

  @abc.abstractproperty
  def flags_read(self) -> ty.Set['Flag']:
    return NotImplemented

  @abc.abstractproperty
  def flags_written(self) -> ty.Set['Flag']:
    return NotImplemented


class Composite_Operand(Operand):
  """
    An operand composed of multiple sub-operands.
  """

  @abc.abstractproperty
  def sub_operands(self) -> [Operand]:
    ...

  @abc.abstractmethod
  def update_sub_operand(self, idx_or_name: ty.Union[int, str],
                         fn: ty.Callable[['Operand'], 'Operand']) -> 'Composite_Operand':
    ...


class Memory_Operand(Composite_Operand):
  """
    Operand reads/writes to a memory address.
  """

  @abc.abstractproperty
  def address_width(self) -> int:
    return NotImplemented

  @abc.abstractproperty
  def memory_width(self) -> int:
    return NotImplemented


class Address_Operand(Composite_Operand):
  """
    Operand calculates a memory address, but does not access it. (LEA, BND, ...)
  """

  @abc.abstractproperty
  def address_width(self) -> int:
    return NotImplemented


class Base_Displacement_Operand:
  @abc.abstractproperty
  def base(self) -> Register_Operand:
    return NotImplemented

  @abc.abstractproperty
  def displacement(self) -> Immediate_Operand:
    return NotImplemented

  @abc.abstractmethod
  def with_base(self, base_reg: Register) -> 'Base_Displacement_Operand':
    return NotImplemented

  @abc.abstractmethod
  def with_displacement(self, disp: int) -> 'Base_Displacement_Operand':
    return NotImplemented


class Base_Displacement_Address_Operand(Address_Operand, Base_Displacement_Operand):
  """
    Operand that computes a memory address specified by a base register + a constant displacement.
  """


class Base_Displacement_Memory_Operand(Memory_Operand, Base_Displacement_Operand):
  """
    Operand reads/writes to a memory address specified by a base register + a constant displacement.
  """


################################################################################
##### IR HELPERS

class Label(ty.NamedTuple):
  """
    A reference to a symbol in the program.
    Can be used as value for Immediate_Operands
    This is only a helper, in real machine code labels are encoded as Constants.
  """

  name: str


class Loop(ty.NamedTuple):
  """A helper for writing loops."""

  name: str
  head: Label
  exit: Label
