#! /usr/bin/env python3
# vim: et:ts=4:sw=4:fenc=utf-8

"""
  server for the port reverse engineering framework from Saarbruecken.
"""

import argparse
import os
import random
import rpyc
import tempfile
from typing import *

from pipedream.benchmark       import throughput_benchmark
from pipedream.benchmark       import show_stats
from pipedream.benchmark.types import Benchmark_Spec
from pipedream.asm             import ir

WHITE_LIST = [
  'add-i64-rr',
  'sub-i64-rr',
  'andn-b64-rrr',
  'shr-b64-ir',
  'AES_encode',
]


def main():
  random.seed(43)

  throughput_benchmark.set_scheduler_params()

  argparser = argparse.ArgumentParser(description='Erlangen-based experiment provider server')

  group = argparser.add_argument_group('server options')
  group.add_argument('--port', metavar='PORT', type=int, default="42424",
                     help='the port to listen for requests')
  group.add_argument('--sslpath', metavar='PATH', default=".",
                     help='the path to a folder containing an SSL key, certificate and ca file')

  group = argparser.add_argument_group('benchmark options')
  # group.add_argument('--sslpath', metavar='PATH', default=".",
  #                    help='the path to a folder containing an SSL key, certificate and ca file')

  args = argparser.parse_args()

  arch = ir.Architecture.for_name('x86')

  lleval = Experiment_Evaluator(arch=arch, instruction_whitelist=WHITE_LIST)
  print("Starting server on port {}".format(args.port))
  start_server(lleval, port=args.port, folder=args.sslpath)


class Experiment_Evaluator:
  NUM_PORTS = 8

  def __init__(self, arch: ir.Architecture, instruction_whitelist: List[str]):
    self.arch                  = arch
    self.instruction_whitelist = instruction_whitelist
    self.instruction_set       = arch.instruction_set()

  def get_insns(self) -> Iterable[str]:
    insts = []

    for inst in self.instruction_set.benchmark_instructions():
      assert type(inst.name) is str

      if self.instruction_whitelist and inst.name not in self.instruction_whitelist:
        continue

      insts.append(inst.name)

    return insts

  def get_num_ports(self) -> int:
    return self.NUM_PORTS

  # def run_experiment(self, exp) -> Dict[str, object]:
  def run_experiment(self, exp: Iterable[str]) -> int:
    dynamic_instructions = 10_000_000
    unroll_factor        = 100

    class Args:
      pass

    args = Args()
    args.architecture = self.arch
    args.debug        = True
    args.output = Args()
    args.output.tmp_dir = tempfile.mkdtemp(prefix='pipedream.', dir='/tmp/')
    args.run = Args()
    args.run.num_iterations        = 50
    args.run.num_warmup_iterations = 10

    benchmark = Benchmark_Spec.from_instructions(
      arch                     = self.arch,
      num_dynamic_instructions = dynamic_instructions,
      unroll_factor            = unroll_factor,
      instructions             = [self.instruction_set.instruction_for_name(inst) for inst in exp],
    )

    benchmark_specs = [benchmark]
    perf_counters   = throughput_benchmark.Perf_Counter_Spec()

    total_num_iterations = args.run.num_iterations + args.run.num_warmup_iterations

    bench_lib = throughput_benchmark.gen_benchmark_lib(
      benchmark_specs,
      architecture   = args.architecture,
      num_iterations = total_num_iterations,
      debug          = args.debug,
      tmp_dir        = args.output.tmp_dir,
    )

    results = throughput_benchmark.run_benchmark(
      perf_counters         = perf_counters,
      benchmark_lib         = bench_lib,
      benchmark             = benchmark,
      num_iterations        = args.run.num_iterations,
      num_warmup_iterations = args.run.num_warmup_iterations,
      outlier_low           = 0,
      outlier_high          = 100,
    )

    num_kernel_iterations = benchmark.kernel_iterations * benchmark.unroll_factor
    cycles                = results.cycles.min
    uops                  = results.muops.min

    throughput_benchmark.info.commit()

    show_stats.pretty_print_benchmark_run(results)

    return {
      'cycles': cycles / num_kernel_iterations,
      'uops':   uops // num_kernel_iterations,
    }


################################################################################
### HELPER CODE COPIED FROM SAARBRUECKEN REPO (SERVER/CLIENT INTERFACE)

class BenchmarkingService(rpyc.Service):
    def __init__(self, lleval):
        self.lleval = lleval

    def on_connect(self, conn):
        # code that runs when a connection is created
        # (to init the service, if needed)
        print("Opened connection")

    def on_disconnect(self, conn):
        # code that runs after the connection has already closed
        # (to finalize the service, if needed)
        print("Closed connection")

    def exposed_get_insns(self):
        print("  handling request for instruction list")
        for i in self.lleval.get_insns():
            print("    {}".format(i))
        return self.lleval.get_insns()

    def exposed_get_num_ports(self):
        print("  handling request for port number")
        print("    self.lowleveleval.get_num_ports()")
        return self.lleval.get_num_ports()

    def exposed_run_experiment(self, insnseq):
        print("  handling request for running experiment ", insnseq)
        res = self.lleval.run_experiment(insnseq)
        print("    {} cycles".format(res))
        return res


def start_server(lleval, certfile=None, keyfile=None, cafile=None, folder=None, port=42424):
    from rpyc.utils.authenticators import SSLAuthenticator
    from rpyc.utils.server import ThreadedServer

    if folder != None:
        assert certfile == keyfile == cafile == None
        certfile = os.path.join(folder, "cert.pem")
        keyfile = os.path.join(folder,  "key.pem")
        cafile = os.path.join(folder, "ca_file.pem")

    if not os.path.isfile(certfile):
        print("Certfile missing!")
        exit(1)

    if not os.path.isfile(keyfile):
        print("Keyfile missing!")
        exit(1)

    if not os.path.isfile(cafile):
        print("CA file missing!")
        exit(1)

    service = BenchmarkingService(lleval)

    authenticator = SSLAuthenticator(keyfile=keyfile, certfile=certfile, ca_certs=cafile)
    t = ThreadedServer(service, port=port, authenticator=authenticator)
    t.start()


if __name__ == '__main__':
  main()
