import itertools as it
import math as m
from gurobipy import *
from copy import deepcopy
from fractions import Fraction
import time

import cProfile, pstats, io

class BiGraph:
    def __init__(self, nb_muops, nb_ports, link_list):
        assert len(link_list) == nb_muops
        assert all(len(link_list[u]) == nb_ports for u in range(nb_muops))
        self.nb_muops = nb_muops
        self.nb_ports = nb_ports
        self.link_list = link_list
        self.sub_throughputs = None

    def reset_through(self):
        self.sub_throughputs = None
    
    def __str__(self):
        return "BiGraph: \n\t{} muops \n\t{} ports\n\tconnections: {}".format(self.nb_muops, self.nb_ports, self.link_list)

    # Returns a BiGraph corresponding to these muops and the ports
    # which they are connected to, with the proper connection
    def get_subgraph(self, muops):
        sorted_muops = sorted(muops)
        ports = [ p for p in range(self.nb_ports) if any(self.link_list[m][p] for m in muops)]
        links_muops_ports = []
        for nm, m in enumerate(sorted_muops):
            links_muops_ports.append([])
            for p in ports:
                links_muops_ports[nm].append(self.link_list[m][p])
        return BiGraph(len(muops), len(ports), links_muops_ports)

    def port_connected_to_muop(self, port, muops):
        for u in muops:
            if not self.link_list[u][port]:
                return False
        return True

    # Here we return a dictionary containing a caracterisation
    # of the graph. It contains the number of ports shared by every possible combination
    # of muops, indexed by these sets of muops
    def get_carac(self):
        ret = dict()
        for n in range(self.nb_muops):
            for cb in it.combinations(range(self.nb_muops), n + 1):
                ret[cb] = len([port for port in range(self.nb_ports) if self.port_connected_to_muop(port, cb)])
        return ret
                
    # Returns True if the other graph is the same graph with reordered nodes
    def is_isomorph(self, other):
        if self.nb_muops != other.nb_muops or self.nb_ports != other.nb_ports:
            return False
        s_carac = self.get_carac()
        o_carac = other.get_carac()

        # For a given combination, that is, a given mapping between a subpart
        # of graph1 and a subpart of graph 2 Return true if the number of ports
        # connected to these subparts match
        def determine(l, cb):
            # We have to use a sorted tuple
            other_cb = tuple(sorted([l[i] for i in cb]))
            return s_carac[cb] == o_carac[other_cb]
        perm_list = list(it.permutations(list(range(self.nb_muops))))
        for n in range(self.nb_muops):
            if not perm_list:
                return False
            for cb in it.combinations(range(self.nb_muops), n + 1):
                if not perm_list:
                    return False
                perm_list[:] = [ l for l in perm_list if determine(l, cb)]

        # Return True only if list is not empty - there is still valid candidate
        # after all tests
        return bool(perm_list)

    def contains(self, other):
        if self.nb_muops < other.nb_muops or self.nb_ports < other.nb_ports:
            return False
        s_carac = self.get_carac()
        o_carac = other.get_carac()

        # For a given combination, that is, a given mapping between a subpart
        # of graph1 and a subpart of graph 2 Return true if the number of ports
        # connected to these subparts match
        def determine(l, cb):
            # We have to use a sorted tuple
            self_cb = tuple(sorted([l[i] for i in cb]))
            return s_carac[self_cb] >= o_carac[cb]
        perm_list = list(it.permutations(list(range(self.nb_muops)), other.nb_muops))
        for n in range(other.nb_muops):
            if not perm_list:
                return False
            for cb in it.combinations(range(other.nb_muops), n + 1):
                if not perm_list:
                    return False
                perm_list[:] = [ l for l in perm_list if determine(l, cb)]

        # Return True only if list is not empty - there is still valid candidate
        # after all tests
        return bool(perm_list)
    # A very simple version of get_throughput All muops must have the same
    # throughput
    def get_throughput(self):
        m = Model("throughput")
        m.setParam('OutputFlag', 0)
        ins_var = m.addVar(vtype=GRB.CONTINUOUS)

        up_vars = []
        port_set = set()
        for nu in range(self.nb_muops):
            for p in range(self.nb_ports):
                if self.link_list[nu][p]:
                    port_set.add(p)

        for nu in range(self.nb_muops):
            up_vars.append([])
            for  p in port_set:
                up_vars[nu].append(m.addVar(vtype= GRB.CONTINUOUS))


        m.update()
        # Constraints
        for nu in range(self.nb_muops):
            for np, p in enumerate(port_set):
                if not self.link_list[nu][p]:
                    m.addConstr(up_vars[nu][np] == 0)




        for nu in range(self.nb_muops):
            m.addConstr(ins_var == quicksum(up_vars[nu]))

        for np, p in enumerate(port_set):
            m.addConstr(quicksum(up_vars[nu][np] for nu in range(self.nb_muops)) <= 1)


        # Objective
        m.setObjective(ins_var, GRB.MAXIMIZE)
        m.optimize()

        if m.status == GRB.OPTIMAL:
            return ins_var.X
        else:
            print("Model unfeasible")
            raise ValueError

    # Here we test adding some resources to a set of muops
    # and see if the throughput get better
    def get_augmented_throughput(self, augmented_muops):
        epsilon = 0.1

        m = Model("throughput")
        m.setParam('OutputFlag', 0)
        ins_var = m.addVar(vtype=GRB.CONTINUOUS)

        up_vars = []
        port_set = set()
        for nu in range(self.nb_muops):
            for p in range(self.nb_ports):
                if self.link_list[nu][p]:
                    port_set.add(p)

        eps_vars = []
        for nu in range(self.nb_muops):
            up_vars.append([])
            eps_vars.append(m.addVar(vtype= GRB.CONTINUOUS))
            for  p in port_set:
                up_vars[nu].append(m.addVar(vtype= GRB.CONTINUOUS))



        m.update()
        # Constraints
        for nu in range(self.nb_muops):
            for np, p in enumerate(port_set):
                if not self.link_list[nu][p]:
                    m.addConstr(up_vars[nu][np] == 0)



        for nu in range(self.nb_muops):
            m.addConstr(ins_var == quicksum(up_vars[nu]) + eps_vars[nu])
            if nu in augmented_muops:
                m.addConstr(eps_vars[nu] <= epsilon)
            else:
                m.addConstr(eps_vars[nu] == 0)

        for np, p in enumerate(port_set):
            m.addConstr(quicksum(up_vars[nu][np] for nu in range(self.nb_muops)) <= 1)


        # Objective
        m.setObjective(ins_var, GRB.MAXIMIZE)
        m.optimize()

        if m.status == GRB.OPTIMAL:
            return ins_var.X
        else:
            print("Model unfeasible")
            raise ValueError


    def get_sub_throughput(self, comb):
        if self.sub_throughputs is None:
            self.sub_throughputs = dict()
        if comb in self.sub_throughputs:
            return self.sub_throughputs[comb]
        order = len(comb)
        connected_ports = [p for p in range(self.nb_ports) if any(self.link_list[m][p] for m in comb)]
        max_through = 1000
        if order == 1:
            return len(connected_ports)
        for cb in it.combinations(list(comb), order - 1):
            max_through = min(self.get_sub_throughput(cb), max_through)
        max_through = min(max_through, len(connected_ports) / order)
        self.sub_throughputs[comb] = max_through
        return max_through

    # We want to decide if this throughput is the right one
    def is_throughput_right(self, throughput):
        max_order = 4
        eps = 0.1
        for order in range(1, min(self.nb_muops + 1, max_order)):
            for cb in it.combinations(range(self.nb_muops), order):
                sub_thr = self.get_sub_throughput(cb)
                if sub_thr < throughput:
                    return False
                if sub_thr * order / self.nb_muops > throughput:
                    return False
                # Test if we have converged. Not sure that this happens often
                if sub_thr - eps < sub_thr * order / self.nb_muops <= sub_thr + eps:
                    return True
        # Maybe we could avoid memory leak with that
        # We still need the object - so garbage collection is not allowed
        # but we don't need every intermediate results anymore
        self.reset_through()
        right = throughput - eps <= self.get_throughput() <= throughput + eps
        return right

    def is_bottleneck(self, cb):
        subgraph = self.get_subgraph(cb)
        return self.get_augmented_throughput(cb) > self.get_throughput() and subgraph.get_throughput() == self.get_throughput()


    def get_bottlenecks(self):
        bottlenecks = []
        for n in range(self.nb_muops):
            for cb in it.combinations(range(self.nb_muops), n + 1):
                if self.is_bottleneck(cb):
                    bottlenecks = list(cb)
                    return bottlenecks
        raise ValueError

# Generate the list of all possible digraph 
# with exactly nb_muops and nb_ports
# Plus, we prune the possibilities by checking
# that the graph has the right throughput
def generate(l, i, j, nb_muops, nb_ports, throughput, digraphs, stop_at_first_success=False):
    eps = 0.001
    # Testing if the graph contains another graph which we know contains
    # a graph realizing the throughput
    def arch_present(l):
        ltemp = deepcopy(l)
        for ii in range(nb_muops):
            for jj in range(nb_ports):
                if ltemp[ii][jj] is None:
                    ltemp[ii][jj] = False
        temp_graph = BiGraph(nb_muops, nb_ports, ltemp)
        return any(temp_graph.contains(g) for g in digraphs)


    l1 = deepcopy(l)
    l1[i][j] = True
    l2 = deepcopy(l)
    l2[i][j] = False
    # Check if the graph is able to get at least the throughput
    if not stop_at_first_success and (arch_present(l1) or arch_present(l2)):
        return digraphs
    if (i == nb_muops - 1) and (j == nb_ports - 1):
        l1_able = any(sum(l1[ii]) >= throughput for ii in range(nb_muops))
        l2_able = any(sum(l2[ii]) >= throughput for ii in range(nb_muops))
        ret_list = []
        gr1 = BiGraph(nb_muops, nb_ports, l1)
        if l1_able and gr1.is_throughput_right(throughput):
            if stop_at_first_success:
                return [gr1]
            ret_list.append(gr1)
        gr2 = BiGraph(nb_muops, nb_ports, l2)
        if l2_able and gr2.is_throughput_right(throughput):
            if stop_at_first_success:
                return [gr2]
            ret_list.append(gr2)
        
        return ret_list + digraphs
    elif (j == nb_ports - 1):
        return generate(l2, i + 1, 0,  nb_muops, nb_ports, throughput, generate(l1, i + 1, 0, nb_muops, nb_ports, throughput, digraphs)) 
    else:
        return generate(l2, i, j + 1, nb_muops, nb_ports, throughput, generate(l1, i, j + 1, nb_muops, nb_ports, throughput, digraphs))
        
# Returns the list of all possible graphs 
# Realizing the given throughput in the limits
# of max_muops and max_ports
def generate_digraphs_throughput(throughput, max_muops, max_ports):
    # We need at least p ports if throughput > p
    nb_min_port = m.ceil(throughput)

    # We approximate the throughput with a fraction
    # The denominator is a lower bound to the number
    # of muops
    f = Fraction(throughput).limit_denominator(10)
    nb_min_muop = f.denominator
    digraphs = []
    tested_tuples = []
    for nu in range(nb_min_muop, max_muops + 1):
        for np in range(nb_min_port, max_ports + 1):
            print("trying {} muops {} ports".format(nu, np))
            init_l = [[ None for _ in range(np)] for _ in range(nu)]
            stop_at_first = True
            gen_list = generate(init_l, 0, 0, nu, np, throughput, digraphs, stop_at_first_success=stop_at_first)
            if stop_at_first and bool(gen_list):
                return gen_list
            digraphs = gen_list
    return digraphs

if __name__ == "__main__":
    g1 =  BiGraph(1, 1, [[True]])
    g2 =  BiGraph(1, 2, [[True,  False]])
    g3 =  BiGraph(2, 2, [[True,  False], [True, False]])
    gBig=  BiGraph(3, 3, [[True,  False, False], [True, False, True], [False, True, False]])
    print(g2.contains(g1))
    print(g2.is_throughput_right(1))
    print(g2.get_sub_throughput((0,)))
    print(g3.get_sub_throughput((0,1)))
    print("--------------------")
    print("GENERATION")
    pr = cProfile.Profile()
    t = time.time()
    pr.enable()
    list_graphs =generate_digraphs_throughput(3, 5, 5) 
    pr.disable()
    total_time = time.time() - t
    for g in list_graphs:
        print(g)
    s = io.StringIO()
    sortby = 'cumulative'
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    #ps.print_stats()
    print(s.getvalue())  
    print("generation took {} seconds".format(total_time))
