import copy

"""""""""""""""""""""""""""""""""""""""""""""""""""
Given a list of instruction and a number of constraints
on them consisting in how many ports they are 
accessing individually and how many ports each
pair shares, we build and return all possible
graphs that would match these constraints
"""""""""""""""""""""""""""""""""""""""""""""""""""

class Port:
    def __init__(self, nb):
        self.nb = nb

    def __str__(self):
        string = "Port n° {} ".format(self.nb )
        return string


class Inst:
    def __init__(self, name):
        self.name = name


    @property
    def nb_port(self):
        return len(self.port_set)

class Graph:
    def __init__(self, port_list, inst_list, link_list):
        assert all(isinstance(port, Port) for port in port_list)
        assert all(isinstance(inst, Inst) for inst in inst_list)
        for (port, inst) in link_list:
            assert port in port_list
            assert inst in inst_list
        self.ports = port_list
        self.insts = inst_list
        self.edges = link_list

    def copy(self):
        return Graph(self.port_list, self.inst_list, self.link_list)

    def __str__(self):
        st = ""
        for port in self.ports:
            st += "port {} connected to {}\n".format(port.nb, [ins.name for ins in self.get_insts_from_port(port)])
        return st
    
    def get_insts_from_port(self, port):
        lst = []
        for (p, inst) in self.edges:
            if port is p:
                lst.append(inst)
        return lst


    def get_ports_from_inst(self, inst):
        lst = []
        for (port, i) in self.edges:
            if inst is i:
                lst.append(port)
        return lst

    # Create a port whose name is not in current graph
    # Don't add it to the graph as we want immutable objects
    def create_port(self):
        pmax = max([p.nb for p in self.ports] + [-1])
        return Port(pmax + 1)


    def get_common_ports(self, inst1, inst2):
        p_l1 = self.get_ports_from_inst(inst1)
        p_l2 = self.get_ports_from_inst(inst2)
        ret_list = []
        for p in p_l1 + p_l2:
            if p in p_l1 and p in p_l2 and p not in ret_list:
                ret_list.append(p)
        return ret_list

    def is_equal(self, other_graph):
        # Condition on equality is:
        # For any port in graph1, there is a port in graph 2 which supports the exact same instructions
        # Obviously the reverse condition must true.
        return (all(any(all(i in other_graph.get_insts_from_port(p2) for i in
            self.get_insts_from_port(p1)) for p2 in other_graph.ports) for p1 in self.ports) and
            all(any(all(i in self.get_insts_from_port(p1) for i in
                other_graph.get_insts_from_port(p2)) for p1 in self.ports) for
                p2 in other_graph.ports))


# Build a graph corresponding to given constraints
# inst_names are the names of the instructions
# inst_dict contains, for each instruction, the number 
# of ports they can access.
# shared_dict contains every pair of instruction that share a port
def construct_graphs(inst_names, inst_dict, shared_dict):
    insts = []
    inst_ports_to_assign = []
    inst_links_to_assign = []
    # This dict is supposed to be immutable
    # it is accessed everywhere across the call graph
    inst_links_dict = dict()
    inst_ports_dict = dict()
    # Initializing
    for name in inst_names:
        inst = Inst(name)
        insts.append(inst)
    for inst in insts:
        inst_ports_to_assign.append(inst_dict[inst.name])
        inst_ports_dict[inst] = inst_dict[inst.name]
        name_count_dict = shared_dict[inst.name]
        inst_links_to_assign.append([])
        inst_links_dict[inst] = dict()
        for (name, c) in name_count_dict:
            other_inst = next(i for i in insts if i.name == name)
            inst_links_to_assign[-1].append((other_inst, c))
            inst_links_dict[inst][other_inst] = c
    # This is a recursive intermediate function that returns a list of graph corresponding
    # to the constraints
    def inter_graphs(graph, insts_to_treat, inst_links_to_assign, inst_ports_to_assign):
        # This helper function returns true if inst can be connected to port without 
        # breaking any restriction
        def is_port_suitable(port, inst, graph):
            inst_links = inst_links_dict[inst]
            inst_list = graph.get_insts_from_port(port)
            suitable = all(ins in inst_links_dict[inst] and len(graph.get_common_ports(inst, ins)) < inst_links_dict[inst][ins]   for ins in inst_list )
            return suitable
        # We are given a set of ports 
        # As we consider the choice of two ports with the same 
        # instructions connections to be equivalent, we prune
        # the number of possible choices
        def get_distinct_choices(port_set):
            ret_list = []
            for p in port_set:
                if not any(set(graph.get_insts_from_port(p)) == set(graph.get_insts_from_port(port)) for port in ret_list):
                    ret_list.append(p)
            return ret_list

        # If no instruction to treat
        # Then the algorithm has come to an end
        if not insts_to_treat:
            return [graph]

        inst = insts_to_treat[0]
        # Create as many ports as needed for inst
        list_new_port = []
        list_new_edges = []
        list_new_insts = []
        current_inst_ports = graph.get_ports_from_inst(inst)
        nb_ports_to_assign = inst_ports_to_assign[0] - len(current_inst_ports) 
        # First case : not enough port on this inst
        # creating a new port, attaching it to inst
        # and recall the function with the new graph
        if nb_ports_to_assign > 0:
            new_port = graph.create_port()
            list_new_port.append(new_port)
            if inst not in graph.insts:
                list_new_insts.append(inst)
            list_new_edges.append((new_port, inst))

            new_graph = Graph(graph.ports + list_new_port, graph.insts + list_new_insts, graph.edges + list_new_edges)
            return inter_graphs(new_graph, insts_to_treat, inst_links_to_assign, inst_ports_to_assign)


        # If all links to inst are treated We have finsihed treating inst,
        # recalling the function while decrementing (functional style) the
        # parameters
        elif not inst_links_to_assign[0] :
            new_graph = Graph(graph.ports + list_new_port, graph.insts + [inst], graph.edges + list_new_edges)
            return inter_graphs(new_graph, insts_to_treat[1::], inst_links_to_assign[1::], inst_ports_to_assign[1::])
        other_inst, count = inst_links_to_assign[0][0]
        common_ports = graph.get_common_ports(inst, other_inst)

        # For this other instruction, every link has already been created
        # Getting to next instruction
        if count == len(common_ports):
            inst_asgn = inst_links_to_assign.copy()
            inst_asgn[0] = inst_links_to_assign[0][1::]
            inst_prts_asgn = inst_ports_to_assign.copy()
            new_graph = Graph(graph.ports + list_new_port, graph.insts + [inst], graph.edges + list_new_edges)
            return inter_graphs(new_graph, insts_to_treat, inst_asgn, inst_prts_asgn)
        # There is a link to create here
        else: 
            # List of ports connected to inst that could be connected to
            # other_inst
            suitable_ports = [ p for p in graph.get_ports_from_inst(inst) if not p in common_ports and is_port_suitable( p, other_inst, graph)]
            choices = get_distinct_choices(suitable_ports)

            ret_list = []
            # Here is a kind of "fork" on every possibility In practice, a lot
            # of these possible graph will merge later but it is not always the
            # case
            for p in choices:
                list_edgs = list_new_edges.copy()
                # "Linking" other inst to port p 
                list_edgs.append((p, other_inst))
                inst_lnk_asgn = inst_links_to_assign.copy()
                inst_prts_asgn = inst_ports_to_assign.copy()
                new_graph = Graph(graph.ports + list_new_port, graph.insts + [other_inst], graph.edges + list_edgs)
                ret_list.extend(inter_graphs(new_graph, insts_to_treat, inst_lnk_asgn, inst_prts_asgn))
            return ret_list



    graph = Graph([], [], [])
    gs = inter_graphs(graph, insts, inst_links_to_assign, inst_ports_to_assign)
    return_list = []
    for g in gs:
        if not any(g.is_equal(gr) for gr in return_list):
            return_list.append(g)
    return return_list

            

#inst_names = ["add", "sub", "div"]
#inst_dict = dict()
#inst_dict["sub"] = 2
#inst_dict["add"] = 2
#inst_dict["div"] = 2
#for k, v in inst_dict.items():
#    print("{} {}".format(k, v))
#shared_dict = dict()
#shared_dict["sub"] = [ ("div", 1), ("add", 1)]
#shared_dict["add"] = [ ("div", 1), ("sub", 1)]
#shared_dict["div"] = [("add", 1), ("sub", 1)]
#
#SECOND TEST
#inst_names = ["a", "b", "c", "d"]
#inst_dict = dict()
#inst_dict["a"] = 3
#inst_dict["b"] = 3
#inst_dict["c"] = 1
#inst_dict["d"] = 3
#for k, v in inst_dict.items():
#    print("{} {}".format(k, v))
#shared_dict = dict()
#shared_dict["a"] = [ ("b", 2), ("d", 1)]
#shared_dict["b"] = [ ("a", 2), ("c", 1), ("d", 1)]
#shared_dict["c"] = [  ("b", 1)]
#shared_dict["d"] = [  ("a", 1), ("b", 1)]


# REAL TEST
inst_names = ["alu", "div", "br", "fp", "vlu", "agu", "load"]
inst_dict = dict()
inst_dict["alu"] = 4
inst_dict["div"] = 1
inst_dict["br"] = 2
inst_dict["fp"] = 2
inst_dict["vlu"] = 3
inst_dict["agu"] = 3
inst_dict["load"] = 2
for k, v in inst_dict.items():
    print("{} {}".format(k, v))
shared_dict = dict()
shared_dict["alu"] = [ ("div", 1), ("br", 2), ("fp", 2), ("vlu", 3)]
shared_dict["div"] = [ ("alu", 1), ("br", 1), ("fp", 1), ("vlu", 1)]
shared_dict["br"] = [  ("alu", 2), ("div", 1), ("fp", 1), ("vlu", 1)]
shared_dict["fp"] = [  ("alu", 2), ("div", 1),  ("br", 1), ("vlu", 2)]
shared_dict["vlu"] = [  ("alu", 3), ("div", 1),  ("br", 1), ("fp", 2)]
shared_dict["agu"] = [("load", 2)]
shared_dict["load"] = [("agu", 2)]
gs = construct_graphs(inst_names, inst_dict, shared_dict)

print("printing solutions, {} graphs found".format(len(gs)))
print("-------------")
print("-------------")
print("\n\n")
for g in gs:
    print("-------------")
    print(g)
print("\n\n")
print("-------------")
print("-------------")

