import copy


class Port:
    def __init__(self, nb):
        self.nb = nb
        self.inst_set = set()

    def __str__(self):
        string = "Port n° {} connected to {}".format(self.nb, [s.name for s in self.inst_set])
        return string

    def add_inst(self, inst):
        assert isinstance(inst, Inst)
        self.inst_set.add(inst)

class Inst:
    def __init__(self, name):
        self.name = name
        self.port_set = set()

    def add_port(self, port):
            assert isinstance(port, Port)
            self.port_set.add(port)

    @property
    def nb_port(self):
        return len(self.port_set)

class Graph:
    def __init__(self):
        self.port_set = set()
        self.inst_set = set()
        self.edge_set = set()
        self.nb_ports = 0
    
    def create_port(self):
        port = Port(self.nb_ports)
        self.port_set.add(port)
        self.nb_ports += 1
        return port

    def add_inst(self, inst):
        assert isinstance(inst, Port)
        self.inst_set.add(inst)

    def add_edge(self, inst, port):
        assert inst in self.inst_set
        assert port in self.port_set
        inst.add_port(port)
        port.add_inst(inst)
        self.edge_set.add((inst, port))

    def is_edge_in_graph(self, inst, port):
        assert inst in self.inst_set
        assert port in self.port_set
        return (inst, port) in self.edge_set

    def get_common_ports(self, inst1, inst2):
        return [ port for port in self.port_set if inst1 in port.inst_set and inst2 in port.inst_set]


# Build a graph corresponding to given constraints
# inst_names are the names of the instructions
# inst_dict contains, for each instruction, the number 
# of ports they can access.
# shared_dict contains every pair of instruction that share a port
def construct_graph(inst_names, inst_dict, shared_dict):
    insts = []
    inst_ports_to_assign = dict()
    inst_links_to_assign = dict()
    return_list = []
    # Initializing
    for name in inst_names:
        inst = Inst(name)
        insts.append(inst)
        inst_ports_to_assign[inst] = inst_dict[name]
        inst_links_to_assign[inst] = dict()
        list_links = shared_dict[name]
        for other, count in list_links:
            inst_links_to_assign[inst][other] = count
        graph = Graph()
    def inter_graphs(graph, inst_links_to_assign, inst_ports_to_assign):
        def is_port_suitable(port, inst):
            inst_links = inst_links_to_assign[inst]
            inst_names = [inst.name for inst in port.inst_set]
            suitable = all(in_name in [name for name, count in inst_links.items() if count > 0] for in_name in inst_names)
            return suitable
        def get_distinct_choices(port_set):
            ret_list = []
            for p in port_set:
                if not any(p.inst_set == port.inst_set for port in ret_list):
                    ret_list.append(p)
            return ret_list

        for inst in insts:
            list_links = shared_dict[inst.name]
            # If our instruction is not yet connected to enough ports
            if inst_ports_to_assign[inst] > 0:
                for i in range(inst_ports_to_assign[inst]):
                    port = graph.create_port()
                    inst.add_port(port)
                    port.add_inst(inst)
                    inst_ports_to_assign[inst] -=  1
            for name, count in list_links:
                other_inst = next(i for i in insts if i.name == name)
                common_ports = graph.get_common_ports(inst, other_inst)
                if count == len(common_ports):
                    continue
                else:
                    # Must connect other_inst to one of our instructions
                    for to_assign in range(count - len(common_ports)):
                        ports_to_assign = get_distinct_choices([ p for p in (inst.port_set - set(common_ports)) if is_port_suitable(p, other_inst)])
                        print("-------")
                        print("There are {} choices of ports".format(len(ports_to_assign)))
                        for p in ports_to_assign:
                            print(p)
                        print("-------")
                        #print(ports_to_assign)
                        # Now there is a choice to make : to which port are we going to assign other inst ?
                        # For now, just choose the port with the maximum instructions 
                        #assigned_port = max(ports_to_assign, key=lambda p: len(p.inst_set))
                        assigned_port = min(ports_to_assign, key=lambda p: len(p.inst_set))
                        for instr in [i for i in assigned_port.inst_set if i.name != inst.name]:
                            inst_links_to_assign[other_inst][instr.name] -=  1
                            inst_links_to_assign[instr][other_inst.name] -=  1
                        inst_ports_to_assign[other_inst] -=  1
                        inst_links_to_assign[inst][other_inst.name] -= 1
                        inst_links_to_assign[other_inst][inst.name] -= 1
                        other_inst.add_port(assigned_port)
                        assigned_port.add_inst(other_inst)
        assert all(inst_ports_to_assign[inst] == 0 for inst in insts)
        for inst_d in inst_links_to_assign.values():
            assert all(count == 0 for count in inst_d.values())
        return graph
    return inter_graphs(graph, inst_links_to_assign, inst_ports_to_assign)

            

inst_names = ["add", "sub", "div"]
inst_dict = dict()
inst_dict["sub"] = 2
inst_dict["add"] = 2
inst_dict["div"] = 3
for k, v in inst_dict.items():
    print("{} {}".format(k, v))
shared_dict = dict()
shared_dict["sub"] = [ ("div", 2)]
shared_dict["add"] = [ ("div", 1)]
shared_dict["div"] = [("add", 1), ("sub", 2)]
g = construct_graph(inst_names, inst_dict, shared_dict)
for p in g.port_set:
    #print([s.name for s in p.inst_set])
    print(p)
