#!/usr/bin/python3

import typing as ty
import jinja2
import sys
import types
import itertools
import numbers

T = ty.TypeVar('T')


class Freezable:
  SENTINEL = '__Freezable_is_frozen'

  @property
  def frozen(self):
    return hasattr(self, Freezable.SENTINEL)

  def __setattr__(self, attr, value):
    if self.frozen:
      raise TypeError('tried to set attribute {0!r} of frozen object', attr)
    object.__setattr__(self, attr, value)

  def freeze(self):
    setattr(self, Freezable.SENTINEL, True)

    for v in self.__dict__.values():
      if isinstance(v, Freezable):
        v.freeze()


class Muop(Freezable):
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name  = name

  def __str__(self):
    return 'Muop({0!r})'.format(self.name)


class Port:
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name = name

  def __contains__(self, muop):
    return (self, muop) in self.model.delta

  def __str__(self):
    return 'Port({0!r})'.format(self.name)


class Sequence:
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name  = name

  @property
  def total_throughput(self):
    return self.model.total_throughput[self]

  @total_throughput.setter
  def total_throughput(self, value):
    self.model.total_throughput[self] = value

  def __str__(self):
    return 'Sequence({0!r})'.format(self.name)


class Value:
  def __init__(self, value):
    self._value = value

  @property
  def value(self):
    return self._value

  @property
  def is_unspecified(self):
    return self.value is None

  @property
  def is_specified(self):
    return self.value is not None


class Delta(Value):
  FORBID      = False
  FORCE       = True
  UNSPECIFIED = None

  def __init__(self, value):
    if value is None:
      super().__init__(None)
    elif isinstance(value, numbers.Number) and bool(value) == value:
      super().__init__(value)
    else:
      raise TypeError('Expected None, 0, or 1 (or something equivalent)')

  def __str__(self):
    v = self.value
    return '???' if v is None else str(int(v))


class Throughput(Value):
  UNSPECIFIED = None

  def __init__(self, value):
    value = value if value is None else float(value)

    super().__init__(value)

  def __str__(self):
    v = self.value
    return '???' if v is None else str(v)


class Dimension(ty.Generic[T]):
  def __init__(self, type: type):
    self.type = type
    self.list = []

  def append(self, elem):
    assert elem not in self.list
    self.list.append(elem)

  def __iter__(self):
    return iter(self.list)

  def __contains__(self, elem):
    return elem in self.list

  def __len__(self):
    return len(self.list)


class Relation(dict):
  sentinel = object()

  def __init__(self, name: str, dimensions: ty.Sequence[Dimension], type: type, default = None):
    self.name       = name
    self.dimensions = tuple(dimensions)
    self.domain     = tuple(d.type for d in dimensions)
    self.type       = type
    self.default    = default

  def __getitem__(self, key):
    if type(key) is not tuple:
      key = (key,)

    assert len(key) == len(self.domain)
    # TODO: implement slicing

    for i, k, t in zip(itertools.count(), key, self.domain):
      assert type(k) == t

    return super().__getitem__(key)

  def __setitem__(self, key, value):
    value = self.convert_value(value)

    if type(key) is not tuple:
      key = (key,)

    assert len(key) == len(self.domain)

    for i, k, t in zip(itertools.count(), key, self.domain):
      assert type(k) == t, 'Dimension {0} has type {1.__name__}, key[{0}] contains {2.__name__}'.format(i, t, type(k))

    if value is self.default:
      self.pop(key, None)
    else:
      super().__setitem__(key, value)

  def __missing__(self, key):
    assert type(key) is tuple

    sentinel = object()
    attr     = 'default_' + self.name

    while len(key):
      value = getattr(key[-1], attr, sentinel)

      if value is not sentinel:
        return value

      key = key[:-1]

    return self.default

  @property
  def default(self):
    return self._default

  @default.setter
  def default(self, value):
    self._default = self.convert_value(value)

  def convert_value(self, value):
    try:
      return self.type(value)
    except TypeError as e:
      raise TypeError("Want a {0.type.__name__}, have {1}".format(self, value)) from e

  def __iter__(self):
    return itertools.product(*self.dimensions)


class Model:
  def __init__(self):
    self.ports     = Dimension(Port)
    self.muops     = Dimension(Muop)
    self.sequences = Dimension(Sequence)

    self.instruction_mix = Relation('instruction_mix', [self.sequences, self.muops], int, 0)

    self.delta            = Relation('delta', [self.muops, self.ports], Delta, Delta.FORBID)
    self.throughput       = Relation('throughput', [self.sequences, self.muops, self.ports], Throughput)
    self.total_throughput = Relation('total_throughput', [self.sequences], Throughput)

  def new_port(self, name, *muops):
    assert name not in (p.name for p in self.ports), "Tried to create port {0!r} twice".format(name)

    port = Port(self, len(self.ports), name)
    self.ports.append(port)

    for muop in muops:
      self.delta[muop, port] = Delta.FORCE

    return port

  def new_muop(self, name):
    assert name not in (m.name for m in self.muops), "Tried to create muop {0!r} twice".format(name)

    muop = Muop(self, len(self.muops), name)
    self.muops.append(muop)
    return muop

  def new_sequence(self, *, total_throughput: float = None, instruction_mix: ty.Mapping[Muop, int] = None):
    seq = Sequence(self, len(self.sequences), str(len(self.sequences)))
    self.sequences.append(seq)

    if total_throughput is not None:
      seq.total_throughput = total_throughput

    if instruction_mix:
      for muop, ratio in instruction_mix.items():
        assert type(ratio) is int
        assert ratio >= 0

        self.instruction_mix[seq, muop] = ratio

    return seq

  @property
  def num_sequences(self):
    return len(self.sequences)

  @property
  def num_ports(self):
    return len(self.ports)

  @property
  def num_muops(self):
    return len(self.muops)


def getitem(collection, key, default = None):
  def index(collection, key, default):
    try:
      idx = collection.index(key)
      return collection[idx]
    except IndexError:
      return default

  if isinstance(collection, dict):
    getter = type(collection).get
  elif isinstance(collection, (list, tuple, str, bytes)):
    getter = index
  else:
    getter = getattr

  return getter(collection, key, default)


def sort_by_index(seq):
  return sorted(seq, key=lambda idx: tuple(x.index for x in idx))


### muops
model = Model()

model.delta.default            = None
model.throughput.default       = None
model.total_throughput.default = 0

ALU   = model.new_muop("ALU")
SHIFT = model.new_muop("SHIFT")

port0 = model.new_port('port0', ALU, SHIFT)
port1 = model.new_port('port1', ALU)
# port2 = model.new_port('port2')
# port3 = model.new_port('port3')
# port4 = model.new_port('port4')
port5 = model.new_port('port5', ALU)
port6 = model.new_port('port6', ALU, SHIFT)
# port7 = model.new_port('port7')

model.new_sequence(
  total_throughput = 1,
  instruction_mix = {
    ALU: 1,
  }
)
model.new_sequence(
  total_throughput = 2,
  instruction_mix = {
    ALU: 2,
  }
)

model.new_sequence(
  total_throughput = 1,
  instruction_mix = {
    SHIFT: 1,
  }
)

model.new_sequence(
  total_throughput = 1.92,
  instruction_mix = {
    ALU:   1,
    SHIFT: 1,
  }
)


model.new_sequence(
  total_throughput = 3.1,
  instruction_mix = {
    ALU:   2,
    SHIFT: 2,
  }
)

s = model.new_sequence(
  total_throughput = None,
  instruction_mix = {
    ALU:   4,
    SHIFT: 2,
  }
)


template_string = """
//##############################################################################
//##### PARAMETERS

int num_sequences = {{model.num_sequences}};
int num_muops     = {{model.num_muops}};
int num_ports     = {{model.num_ports}};
int max_num_edges = num_muops * num_ports;

tuple Sequence {
  key string name;
};
tuple Muop {
  key string name;
};
tuple Port {
  key string name;
};

{Sequence} Seq = {
#% for s in model.sequences
  <"{{ s.name }}">{% if not loop.last %},{% endif %}
#% endfor
};
{Muop} Src = {
#% for m in model.muops
  <"{{ m.name }}">{% if not loop.last %},{% endif %}
#% endfor
};
{Port} Dst = {
#% for p in model.ports
  <"{{ p.name }}">{% if not loop.last %},{% endif %}
#% endfor
};

/// edge in graph (delta[i][j] = 1 iff there is an edge from Src i to Dst j)
dvar int delta[Src][Dst] in 0..1;

/// throughput per edge (i,j) in sequence s
dvar float throughput[Seq][Src][Dst];

/// total system throughput (uops/cycle)
dvar float total_throughput[Seq];


//##############################################################################
//##### HELPERS

/// throughput for muop type i in sequence s
dexpr float muop_throughput[s in Seq][i in Src] = sum(j in Dst) throughput[s][i][j];

/// mean throughput for a given muop type in a sequence
dexpr float mean[s in Seq][i in Src]     = (1 / num_ports) * sum(j in Dst) throughput[s][i][j];

/// variance of throughput for a given muop type in a sequence
dexpr float variance[s in Seq][i in Src] = (1 / num_ports) *
  sum(j in Dst) pow(throughput[s][i][j], 2) - pow(mean[s][i], 2);


//##############################################################################
//##### OBJECTIVE FUNCTION

dexpr float all_throughputs = sum(s in Seq) total_throughput[s];

/// number of edges present in the machine
dexpr float num_edges = sum(i in Src, j in Dst) delta[i][j];

dexpr float all_variances = sum(s in Seq, i in Src) variance[s][i];

maximize 500 * (sum(s in Seq) total_throughput[s] - num_sequences * num_edges) - 1 * all_variances;


//##############################################################################
//##### CONSTRAINTS

subject to {
  forall(s in Seq, i in Src, j in Dst) {
    throughput_bounds:
      0 <= throughput[s][i][j];
  }
  forall(s in Seq, i in Src, j in Dst) {
    throughputs_limited_by_delta:
      throughput[s][i][j] <= delta[i][j];
  }

  forall(s in Seq, j in Dst) {
    port_bandwidth:
      sum(i in Src) throughput[s][i][j] <= 1;
  }

  forall(s in Seq) {
    total_throughput_limited_by_ports:
      total_throughput[s] <= num_ports;
  }

  forall(s in Seq) {
    total_throughput_is_sum_of_throughputss:
      total_throughput[s] == sum(i in Src) muop_throughput[s][i];
  }

#// prints constraints set with OPL label iff constraint text is not empty
#% macro constraint_set(name)
#%   set X = caller().strip()
#%   if X and not X.isspace()
  {{ name }}:
    {{ X }}
#%   endif
#% endmacro

#// print delta
#% call constraint_set('port_layout')
#%   for i, j in model.delta | sort_by_index
#%     set edge = model.delta[i, j]
#%     if edge.is_specified
    delta[{{ i | tuple }}][{{ j | tuple }}] == {{ edge }};
#%     endif
#%   endfor
#% endcall

#// print total throughput per sequence
#% call constraint_set('observed_total_throughput')
#%   for seq in model.total_throughput | sort_by_index
#%     set total_throughput = model.total_throughput[seq]
#%     if total_throughput.is_specified
    total_throughput[{{ seq[0] | tuple }}] == {{ total_throughput }};
#%     endif
#%   endfor
#% endcall

#// print throughput per muop/port pair
#% call constraint_set('observed_throughputs')
#%   for s, i, j in model.throughput | sort_by_index
#%     set throughput = model.throughput[s, i, j]
#%     if throughput.is_specified
    throughput[{{ s | tuple }}][{{ i | tuple }}][{{ j | tuple }}] == {{ throughput }};
#%     endif
#%   endfor
#% endcall

#// print instruction ratios in sequences
#% call constraint_set('inst_mix')
#%   set mix = model.instruction_mix
#%   for s, i in mix
#%     set ratio = mix[s, i]
#%     if ratio == 0
         #// can't use muop_throughput here because that would use tuple patterns
         #// and we already use non-pattern constraints elsewhere
    sum(j in Dst) throughput[{{ s | tuple }}][{{ i | tuple }}][j] == 0;
#%     endif
#%   endfor
#%   for s in model.sequences
#%     for i1, i2 in model.muops | combinations(2)
#%       set r1 = mix[s, i1]
#%       set r2 = mix[s, i2]
#%       if r1 and r2
    {{ r1 }} * sum(j in Dst) throughput[{{ s | tuple }}][{{ i1 | tuple }}][j] == {{ r2 }} * sum(j in Dst) throughput[{{ s | tuple }}][{{ i2 | tuple }}][j];
#%       endif
#%     endfor
#%   endfor
#% endcall
}

//##############################################################################
//##### OUTPUT

execute {
  writeln("##### BGN DATA");
  writeln('all_throughputs = ', all_throughputs);
  writeln('num_edges       = ', num_edges);
  writeln('all_variances   = ', all_variances);
  writeln();
  for(i in Src) {
    for(j in Dst) {
      writeln('delta[', i.name, '][', j.name, '] = ', delta[i][j]);
    }
  }
  for (s in Seq) {
    writeln();
    writeln('### SEQ ', '"', s.name, '"');
    writeln();
    writeln('total_throughput = ', total_throughput[s]);
    writeln()
    for(i in Src) {
      for(j in Dst) {
        writeln('throughput[', s.name, '][', i.name, '][', j.name, '] = ', throughput[s][i][j]);
      }
    }
    writeln()
    for(i in Src) {
        writeln('muop_throughput[', s.name, '][', i.name, '] = ', muop_throughput[s][i]);
    }
    writeln()
    for(i1 in Src) {
      for(i2 in Src) {
        if (i1.name < i2.name) {
          writeln('muop_throughput[', s.name, '][', i1.name, '] / muop_throughput[', s.name, '][', i2.name, '] = ',
                  muop_throughput[s][i1] / muop_throughput[s][i2]);
        }
      }
    }
  }
  writeln("##### END DATA");
}
"""


jinja2_env = jinja2.Environment(
  # # remove whitespace introduced by jinja2 blocks
  # trim_blocks   = True,
  # lstrip_blocks = True,
  line_statement_prefix='#%',
  line_comment_prefix='#//',
)

jinja2_env.filters['sort_by_index'] = sort_by_index
jinja2_env.filters['repr'] = repr
jinja2_env.filters['strip'] = str.strip
jinja2_env.filters['tuple'] = lambda x: '<"' + x.name + '">'
jinja2_env.filters['combinations'] = itertools.combinations

template = jinja2_env.from_string(template_string)

template_args = dict(model.__dict__)
template_args['model'] = model
template_args['enumerate'] = enumerate

if __name__ == '__main__':
  import argparse
  import os.path
  parser = argparse.ArgumentParser(os.path.basename(sys.argv[0]))

  parser.add_argument('-o', '--output', dest='FILE', help='Output file')

  args = parser.parse_args()

  txt = template.render(template_args)

  if not args.FILE:
    output = sys.stdout
  else:
    output = open(args.FILE, 'w')

  print(txt, file=output)
