import itertools

import contextlib
import gurobipy

from pipedream.ilp import types as ilp

__all__ = [
  'solve',
  'solve_for_delta',
  'solve_for_alpha',
]

GRB = gurobipy.GRB
Σ   = gurobipy.quicksum


def solve(model: ilp.Model):
  solve_for_delta(model)
  solve_for_alpha(model)


def solve_for_delta(model: ilp.Model):
  #####################
  ## build gurobi model

  grb = _Gurobi_Model(model)

  num_im_edges = Σ(grb.delta_im[:, :])
  num_mp_edges = Σ(grb.delta_mp[:, :])

  avg_port_throughput = Σ(grb.alpha_kp[:, :]) #/ len(model.kernels)

  error = Σ(grb.rho_opt[k] - grb.rho_mes[k] for k in model.kernels)

  grb.gurobi.update()

  # grb.gurobi.setObjective(num_mp_edges - port_throughput, GRB.MINIMIZE)
  # grb.gurobi.setObjective(avg_port_throughput - num_im_edges - num_mp_edges, GRB.MAXIMIZE)

  grb.gurobi.ModelSense = GRB.MINIMIZE

  i = 1
  def add_obj(expr):
    nonlocal i
    grb.gurobi.setObjectiveN(expr, i, i)
    i += 1

  add_obj(avg_port_throughput)
  # add_obj(num_im_edges)
  # add_obj(num_mp_edges)

  # add_obj(num_im_edges)
  # add_obj(num_mp_edges)
  # add_obj(-avg_port_throughput)

  #####################
  ## solve gurobi model

  ## Param:  FeasibilityTol
  ## Type:   double
  ## Default value:  1e-6
  ## Minimum value:  1e-9
  ## Maximum value:  1e-2
  # FeasibilityTol = 1e-9
  # grb.gurobi.setParam('FeasibilityTol', FeasibilityTol)

  grb.gurobi.write('/tmp/delta.lp')
  grb.gurobi.optimize()

  ## http://www.gurobi.com/documentation/6.5/refman/optimization_status_codes.html

  if grb.gurobi.status == GRB.INF_OR_UNBD:
    ## DualReductions
    ## Controls dual reductions
    ## Type:   int
    ## Default value:  1
    ## Minimum value:  0
    ## Maximum value:  1
    ## ... You should disable these reductions if you received an optimization status of INF_OR_UNBD ...

    grb.gurobi.setParam('DualReductions', 0)
    grb.gurobi.optimize()

  if grb.gurobi.status == GRB.UNBOUNDED:
    raise RuntimeError("could not solve for delta, solution is UNBOUNDED")

  if grb.gurobi.status == GRB.INFEASIBLE:
    print('The model is infeasible; computing IIS')

    grb.gurobi.computeIIS()
    print('\nThe following constraint(s) cannot be satisfied:')
    for c in grb.gurobi.getConstrs():
      if c.IISConstr:
        print('%s' % c.constrName)

  print(grb.gurobi.status)

  if grb.gurobi.status != GRB.OPTIMAL:
    raise RuntimeError("could not solve for delta, no optmial solution found")

  print('OBJECTIVE FUNCTION:', grb.gurobi.ObjVal)
  print()

  #######################################
  ## back propagate values to input model

  grb.back_propagate_relation('delta_im', model)
  grb.back_propagate_relation('delta_mp', model)

  #HACK:
  grb.back_propagate_relation('alpha_ki', model)
  grb.back_propagate_relation('alpha_km', model)
  grb.back_propagate_relation('alpha_kp', model)
  grb.back_propagate_relation('alpha_kim', model)
  grb.back_propagate_relation('alpha_kmp', model)

  grb.back_propagate_relation('rho_opt', model)


def solve_for_alpha(model: ilp.Model):
  kernels = model.kernels
  muops   = model.muops

  grb = _Gurobi_Model(model)

  alpha_kmp = grb.alpha_kmp

  ##################

  ## variance of throughput for a given muop type in a kernel
  variance = dict()

  def compute_variance(vars):
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Na%C3%AFve_algorithm
    n      = 0
    sum    = 0
    sum_sq = 0

    for x in vars:
      n      += 1
      sum    += x
      sum_sq += x * x

    return (sum_sq - (sum * sum) / n) / (n - 1)

  for k in kernels:
    for m in muops:
      variance[k, m] = compute_variance(alpha_kmp[k, m, :])

  sum_of_variances = Σ(variance.values())

  grb.gurobi.update()

  grb.gurobi.setObjective(sum_of_variances, gurobipy.GRB.MINIMIZE)

  #####################
  ## solve gurobi model

  grb.gurobi.optimize()
  grb.gurobi.write('/tmp/alpha.lp')

  if grb.gurobi.status != gurobipy.GRB.OPTIMAL:
    raise RuntimeError("could not solve for delta, no optmial solution found")

  #######################################
  ## back propagate values to input model

  grb.back_propagate_relation('alpha_ki', model)
  grb.back_propagate_relation('alpha_km', model)
  grb.back_propagate_relation('alpha_kp', model)
  grb.back_propagate_relation('alpha_kim', model)
  grb.back_propagate_relation('alpha_kmp', model)

  grb.back_propagate_relation('rho_opt', model)


def solve_for_throughput(model: ilp.Model):
  #####################
  ## build gurobi model

  grb = _Gurobi_Model(model)

  total_throughput = Σ(grb.rho_opt[:])

  grb.gurobi.update()

  # grb.gurobi.setObjectiveN(num_edges, 1, 1)
  # grb.gurobi.setObjectiveN(-total_throughput, 2, 2)

  grb.gurobi.setObjective(total_throughput, gurobipy.GRB.MAXIMIZE)

  #####################
  ## solve gurobi model

  grb.gurobi.optimize()
  grb.gurobi.write('/tmp/rho.lp')

  if grb.gurobi.status != gurobipy.GRB.OPTIMAL:
    raise RuntimeError("could not solve for throughput, no optmial solution found")

  #######################################
  ## back propagate values to input model

  grb.back_propagate_relation('rho_opt', model)


class _Gurobi_Model:
  """
    Little helper for building a gurobipy.Model
  """

  def __init__(self, src: ilp.Model):
    """
      Tranform our representation of a ILP model to a gurobi model.
      Does not set the objective function.
    """

    #############################
    ## pick apart pipedream.ilp.Model

    self.src = src

    kernels = src.kernels
    insts   = src.insts
    muops   = src.muops
    ports   = src.ports

    #############################
    ## create gurobi Model object

    self.gurobi = gurobipy.Model(src.name)

    self.gurobi.setParam('OutputFlag', 0)

    ############################
    ## create decision variables

    Kernel = ilp.Kernel
    Inst   = ilp.Inst
    Muop   = ilp.Muop
    Port   = ilp.Port

    relation = self._mk_relation

    ## MODEL INPUT

    self.occurences = occurences = src.occurences
    self.rho_mes    = rho_mes    = src.rho_mes

    ## MODEL OUTPUT

    self.rho_opt = rho_opt = relation('rho_opt', [Kernel], GRB.CONTINUOUS)

    self.alpha_ki = alpha_ki = relation('alpha_ki', [Kernel, Inst], GRB.CONTINUOUS)
    self.alpha_km = alpha_km = relation('alpha_km', [Kernel, Muop], GRB.CONTINUOUS)
    self.alpha_kp = alpha_kp = relation('alpha_kp', [Kernel, Port], GRB.CONTINUOUS)

    self.alpha_kim = alpha_kim = relation('alpha_kim', [Kernel, Inst, Muop], GRB.CONTINUOUS)
    self.alpha_kmp = alpha_kmp = relation('alpha_kmp', [Kernel, Muop, Port], GRB.CONTINUOUS)

    self.delta_im = delta_im = relation('delta_im', [Inst, Muop], GRB.BINARY)

    # FIXME: this should be an integer variable, to give number of each type
    #        of muop we decompose into, but that creates a quadratic equality
    #        constraint (see below), which Gurobi can't handle.
    # HACK:  always decompose into one muop per type, which will give us
    #        `duplicate` muops, we hope we can merge them in a post-pass.
    # delta_im = relation('delta_im', [Inst, Muop], GRB.INTEGER)

    ## whether a muop can be executed on a port
    self.delta_mp = delta_mp = relation('delta_mp', [Muop, Port], GRB.BINARY)

    constraint     = self.constraint
    constraint_set = self.constraint_set

    #####################
    ## helper expressions

    ########################
    ## constraint playground

    # stop model from being unbounded

    # for k, i in cross(kernels, insts):
    #   constraint(alpha_ki[k, i] <= 42, name='bound_alpha_ki', indices=[k, i])

    # for k, m in cross(kernels, muops):
    #   constraint(alpha_km[k, m] <= 42, name='bound_alpha_km', indices=[k, m])

    # for k, p in cross(kernels, ports):
    #   constraint(alpha_kp[k, p] <= 42, name='bound_alpha_kp', indices=[k, p])

    #####################
    ## create constraints

    ## optimal throughput must be better than measured one
    # for k in kernels:
    #   constraint(rho_opt[k] >= rho_mes[k], name='opt >= mes', indices=[k])

      #HACK:
      # constraint((rho_opt[k] + 1) <= rho_mes[k], name='opt =~= mes', indices=[k])

    ## optimal throughput is average throughput of all instructions
    # for k in kernels:
      # num_instructions = sum(occurences[k, :])

      # constraint(
      #   rho_opt[k] == (
      #     Σ(occurences[k, i] * alpha_ki[k, i] for i in insts) /
      #     num_instructions
      #   ),
      #   name='avg(K->I)', indices=[k]
      # )

      # constraint(
      #   rho_opt[k] == (
      #     Σ(alpha_ki[k, i] for i in insts) / num_instructions
      #   ),
      #   name='avg(alpha_ki)', indices=[k]
      # )

      # constraint(
      #   rho_opt[k] == (
      #     Σ(alpha_ki[k, i] for i in insts)
      #   ),
      #   name='avg(alpha_ki)', indices=[k]
      # )

    ## Fabrices' all instructions have same throughput constraint
    # for k in kernels:
    #   for i1, i2 in itertools.combinations(insts, 2):
    #     if occurences[k, i1] and occurences[k, i2]:
    #       constraint(occurences[k, i2] * alpha_ki[k, i1] == occurences[k, i1] * alpha_ki[k, i2],
    #                  name='fab-all-same', indices=[k, i1, i2])

    ## propagate down throughput from inst to muop

    for k, i, m in cross(kernels, insts, muops):
      ## FIXME: should be:
      ##     constraint(alpha_ki[k, i] * delta_im[i, m] == alpha_kim[k, i, m])
      ##   but that gives 'Quadratic equality constraints', can only do inequalities (<=, >=)

      ## HACK: for now, leave out inst to multiple muop decomposition

      constraint((delta_im[i, m] == 1) >> (alpha_ki[k, i] == alpha_kim[k, i, m]))

    ## hacky way to say: (alpha_ki > 0) implies (alpha_kim > 0)
    ##   gives lower bounds to alpha_kim, otherwise ILP leaves them at zero,
    ##   even if alpha_ki is non-zero.
    ## FIXME: does it work?

    for k, i in cross(kernels, insts):
      constraint(alpha_ki[k, i] <= Σ(alpha_kim[k, i, :]))

    ## throughput on port is sum of throughputs of all incoming muops
    ## (propagate down throughput from muop to port)

    with constraint_set('port-sum-incoming') as c:
      for k, p in cross(kernels, ports):
        c([k, p], alpha_kp[k, p] == Σ(alpha_kmp[k, :, p]))

    ## throughput on muop is sum of throughputs of all incoming insts
    ## (propagate throughput back up from muop to inst)

    with constraint_set('muop-sum-incoming') as c:
      for k, m in cross(kernels, muops):
        c([k, m], alpha_km[k, m] == Σ(alpha_kim[k, :, m]))

    ## throughput on muop is sum of throughputs of all ports it can go to
    ## (propagate throughput back up from port to muop)

    with constraint_set('muop-sum-outgoing') as c:
      for k, m in cross(kernels, muops):
        c([k, m], alpha_km[k, m] == Σ(alpha_kmp[k, m, :]))

    ## a port can take at most one muop per cycle

    with constraint_set('port-limit') as c:
      for k, p in cross(kernels, ports):
        c([k, p], alpha_kp[k, p] <= 1)

    ## an instruction that does not appear in a benchmark has no throughput

    with constraint_set('does-not-occur') as c:
      for k, i in cross(kernels, insts):
        if occurences[k, i] == 0:
          c(
            [k, i],
            alpha_ki[k, i] == 0
          )

    ## every pair of instructions share the total throughput according
    ## to their relative number of occurences in the kernel.

    for k in kernels:
      for i1, i2 in itertools.combinations(insts, 2):
        if occurences[k, i1] == 0 or occurences[k, i2] == 0:
          continue

        constraint(occurences[k, i2] * alpha_ki[k, i1] == occurences[k, i1] * alpha_ki[k, i2])

    ## can only have edge from muop to port if there is an edge from
    ## some inst to that muop

    # muop_has_inedge = self._mk_relation('muop_has_inedge', [Muop], GRB.BINARY)

    # for m in muops:
    #   constraint(muop_has_inedge[m] == gurobipy.or_(delta_im[:, m]))

    # for m, p in cross(muops, ports):
    #   constraint((muop_has_inedge[m] == 0) >> (Σ(delta_mp[m, :]) == 0))

    # ## a muop must go to some port

    # for m in muops:
    #   constraint((muop_has_inedge[m] == 0) >> (Σ(delta_mp[m, :]) >= 0))

    ## can only have throughput over existing edges

    for k, i, m in cross(kernels, insts, muops):
      constraint((delta_im[i, m] == 0) >> (alpha_kim[k, i, m] == 0))
      # constraint(alpha_kim[k, i, m] <= (occurences[k, i] * delta_im[i, m]))

    for k, m, p in cross(kernels, muops, ports):
      # constraint((delta_mp[m, p] == 0) >> (alpha_kmp[k, m, p] == 0))
      constraint(alpha_kmp[k, m, p] <= delta_mp[m, p])

    #####################################
    ## fill in known variables from model

    for i, m in src.delta_im:
      constraint(delta_im[i, m] == int(src.delta_im[i, m]))

    for k, i, m in src.alpha_kim:
      constraint(alpha_kim[k, i, m] == src.alpha_kim[k, i, m])

    for i, m in src.delta_mp:
      constraint(delta_mp[i, m] == int(src.delta_mp[i, m]))

    for k, i, m in src.alpha_kmp:
      constraint(alpha_kmp[k, i, m] == src.alpha_kmp[k, i, m])

  ### helpers for building the model

  def bool_var(self, name='', *indices):
    return self.var(gurobipy.GRB.BINARY, name, indices)

  def float_var(self, name='', *indices):
    return self.var(gurobipy.GRB.CONTINOUS, name, indices)

  def var(self, vtype, name='', *indices):
    name = self._var_name(name, indices)

    var = self.gurobi.addVar(name=name, vtype=vtype)

    self.gurobi.update()

    return var

  def constraint(self, constraint, *, name='', indices=()):
    name = self._var_name(name, indices)

    self.gurobi.addConstr(constraint, name=name)

  @contextlib.contextmanager
  def constraint_set(self, name):
    c = self.Constraint_Set(self, name)
    yield c

  class Constraint_Set:
    def __init__(self, model: '_Gurobi_Model', name):
      self.model = model
      self.name  = name

    def __call__(self, indices: list, expr):
      self.model.constraint(
        expr,
        name    = self.name,
        indices = indices,
      )

  @staticmethod
  def _var_name(name: str, indices):
    assert (not indices) or name

    if indices:
      name = name + '[' + ','.join(i.name for i in indices) + ']'

    return name

  def _mk_relation(self, name, domain, var_type) -> ilp.Relation:
    rel = ilp.Relation(name, domain, gurobipy.Var)

    TYPE_TO_DOMAIN = {
      ilp.Kernel: self.src.kernels,
      ilp.Inst:   self.src.insts,
      ilp.Muop:   self.src.muops,
      ilp.Port:   self.src.ports
    }

    for key in cross(*[TYPE_TO_DOMAIN[t] for t in domain]):
      rel[key] = self.var(var_type, name, *key)

    return rel

  ### misc

  def back_propagate_relation(self, name: str, dst: ilp.Model):
    src_rel = getattr(self, name)
    dst_rel = getattr(dst, name)

    for key, var in src_rel.items():
      dst_rel[key] = gurobi_to_python(var)


def gurobi_to_python(var: gurobipy.Var, default=None):
  """
    Convert gurobi value to python value (bool, int, float)
  """

  assert isinstance(var, gurobipy.Var)

  vtype = var.VType

  try:
    value = var.X
  except AttributeError:
    if default is None:
      raise

    value = default

  if vtype is GRB.BINARY:
    ## don't do this in general,
    ## here we know the possible values of the RHS
    eps = 0.00001

    if abs(value - 0.0) < eps:
      return False

    if abs(value - 1.0) < eps:
      return True

    raise ValueError(str(value) + ' not bool-ish')

  if vtype in (GRB.INTEGER):
    assert int(value) == value

    return int(value)

  if vtype in (GRB.CONTINUOUS):
    return value

  raise TypeError('Invalid Gurobi type ' + repr(vtype))


def cross(*iterables):
  return itertools.product(*iterables)
