#!/usr/bin/python3

import typing as ty
import jinja2
import sys
import types
import itertools
import numbers
from gurobi_model import resolve_model_through, resolve_model_inst_mix

T = ty.TypeVar('T')


class Freezable:
  SENTINEL = '__Freezable_is_frozen'

  @property
  def frozen(self):
    return hasattr(self, Freezable.SENTINEL)

  def __setattr__(self, attr, value):
    if self.frozen:
      raise TypeError('tried to set attribute {0!r} of frozen object', attr)
    object.__setattr__(self, attr, value)

  def freeze(self):
    setattr(self, Freezable.SENTINEL, True)

    for v in self.__dict__.values():
      if isinstance(v, Freezable):
        v.freeze()


class Muop(Freezable):
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name  = name

  def __str__(self):
    return 'Muop({0!r})'.format(self.name)


class Port:
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name = name

  def __contains__(self, muop):
    return (self, muop) in self.model.delta

  def __str__(self):
    return 'Port({0!r})'.format(self.name)


class Sequence:
  def __init__(self, model, index: int, name: str):
    self.model = model
    self.index = index
    self.name  = name

  @property
  def total_throughput(self):
    return self.model.total_throughput[self]

  @total_throughput.setter
  def total_throughput(self, value):
    self.model.total_throughput[self] = value

  def __str__(self):
    return 'Sequence({0!r})'.format(self.name)


class Value:
  def __init__(self, value):
    self._value = value

  @property
  def value(self):
    return self._value

  @property
  def is_unspecified(self):
    return self.value is None

  @property
  def is_specified(self):
    return self.value is not None


class Delta(Value):
  FORBID      = False
  FORCE       = True
  UNSPECIFIED = None

  def __init__(self, value):
    if value is None:
      super().__init__(None)
    elif isinstance(value, numbers.Number) and bool(value) == value:
      super().__init__(value)
    else:
      raise TypeError('Expected None, 0, or 1 (or something equivalent)')

  def __str__(self):
    v = self.value
    return '???' if v is None else str(int(v))


class Throughput(Value):
  UNSPECIFIED = None

  def __init__(self, value):
    value = value if value is None else float(value)

    super().__init__(value)

  def __str__(self):
    v = self.value
    return '???' if v is None else str(v)


class Dimension(ty.Generic[T]):
  def __init__(self, type: type):
    self.type = type
    self.list = []

  def append(self, elem):
    assert elem not in self.list
    self.list.append(elem)

  def __iter__(self):
    return iter(self.list)

  def __contains__(self, elem):
    return elem in self.list

  def __len__(self):
    return len(self.list)


class Relation(dict):
  sentinel = object()

  def __init__(self, name: str, dimensions: ty.Sequence[Dimension], type: type, default = None):
    self.name       = name
    self.dimensions = tuple(dimensions)
    self.domain     = tuple(d.type for d in dimensions)
    self.type       = type
    self.default    = default

  def __getitem__(self, key):
    if type(key) is not tuple:
      key = (key,)

    assert len(key) == len(self.domain)
    # TODO: implement slicing

    for i, k, t in zip(itertools.count(), key, self.domain):
      assert type(k) == t

    return super().__getitem__(key)

  def __setitem__(self, key, value):
    value = self.convert_value(value)

    if type(key) is not tuple:
      key = (key,)

    assert len(key) == len(self.domain)

    for i, k, t in zip(itertools.count(), key, self.domain):
      assert type(k) == t, 'Dimension {0} has type {1.__name__}, key[{0}] contains {2.__name__}'.format(i, t, type(k))

    if value is self.default:
      self.pop(key, None)
    else:
      super().__setitem__(key, value)

  def __missing__(self, key):
    assert type(key) is tuple

    sentinel = object()
    attr     = 'default_' + self.name

    while len(key):
      value = getattr(key[-1], attr, sentinel)

      if value is not sentinel:
        return value

      key = key[:-1]

    return self.default

  @property
  def default(self):
    return self._default

  @default.setter
  def default(self, value):
    self._default = self.convert_value(value)

  def convert_value(self, value):
    try:
      return self.type(value)
    except TypeError as e:
      raise TypeError("Want a {0.type.__name__}, have {1}".format(self, value)) from e

  def __iter__(self):
    return itertools.product(*self.dimensions)


class Model:
  def __init__(self):
    self.ports     = Dimension(Port)
    self.muops     = Dimension(Muop)
    self.sequences = Dimension(Sequence)

    self.instruction_mix = Relation('instruction_mix', [self.sequences, self.muops], int, 0)

    self.delta            = Relation('delta', [self.muops, self.ports], Delta, Delta.FORBID)
    self.throughput       = Relation('throughput', [self.sequences, self.muops, self.ports], Throughput)
    self.total_throughput = Relation('total_throughput', [self.sequences], Throughput)

  def new_port(self, name, *muops):
    assert name not in (p.name for p in self.ports), "Tried to create port {0!r} twice".format(name)

    port = Port(self, len(self.ports), name)
    self.ports.append(port)

    for muop in muops:
      self.delta[muop, port] = Delta.FORCE

    return port

  def new_muop(self, name):
    assert name not in (m.name for m in self.muops), "Tried to create muop {0!r} twice".format(name)

    muop = Muop(self, len(self.muops), name)
    self.muops.append(muop)
    return muop

  def new_sequence(self, *, total_throughput: float = None, instruction_mix: ty.Mapping[Muop, int] = None):
    seq = Sequence(self, len(self.sequences), str(len(self.sequences)))
    self.sequences.append(seq)

    if total_throughput is not None:
      seq.total_throughput = total_throughput

    if instruction_mix:
      for muop, ratio in instruction_mix.items():
        assert type(ratio) is int
        assert ratio >= 0

        self.instruction_mix[seq, muop] = ratio

    return seq

  @property
  def num_sequences(self):
    return len(self.sequences)

  @property
  def num_ports(self):
    return len(self.ports)

  @property
  def num_muops(self):
    return len(self.muops)


def getitem(collection, key, default = None):
  def index(collection, key, default):
    try:
      idx = collection.index(key)
      return collection[idx]
    except IndexError:
      return default

  if isinstance(collection, dict):
    getter = type(collection).get
  elif isinstance(collection, (list, tuple, str, bytes)):
    getter = index
  else:
    getter = getattr

  return getter(collection, key, default)


def sort_by_index(seq):
  return sorted(seq, key=lambda idx: tuple(x.index for x in idx))

# Resolve this model with gurobi, assuming we only know 
# the number and name of instructions, the number of ports
# and the total throughput for each sequence
# This function update the values of the given model
def resolve_model(model):
    assert isinstance(model, Model)
    # Resolve_model_through take only dictionaries as input
    # We must convert our custom class into dict
    total_throughputs = dict()
    for seq in model.sequences.list:
        val = model.total_throughput[seq].value
        assert val is not None
        total_throughputs[seq] = val
    # We get the results of the model as three dicts
    #delta = resolve_model_through(total_throughputs, model.sequences.list, model.muops.list, model.ports.list)
    delta = resolve_model_inst_mix(total_throughputs, model.instruction_mix, model.sequences.list, model.muops.list, model.ports.list)

    # Now updating the models from the values we got
    for (inst, port), value in delta.items():
        model.delta[(inst, port)] =  value

### muops
model = Model()

model.delta.default            = None
model.throughput.default       = None
model.total_throughput.default = 0

ALU   = model.new_muop("ALU")
SHIFT = model.new_muop("SHIFT")

port0 = model.new_port('port0')
port1 = model.new_port('port1')
#port5 = model.new_port('port5')
#port6 = model.new_port('port6')
#port0 = model.new_port('port0', ALU, SHIFT)
#port1 = model.new_port('port1', ALU)
# port2 = model.new_port('port2')
# port3 = model.new_port('port3')
# port4 = model.new_port('port4')
#port5 = model.new_port('port5', ALU)
#port6 = model.new_port('port6', ALU, SHIFT)
# port7 = model.new_port('port7')

#model.new_sequence(
#  total_throughput = 1,
#  instruction_mix = {
#    ALU: 1,
#  }
#)
model.new_sequence(
  total_throughput = 2,
  instruction_mix = {
    ALU: 1,
  }
)

model.new_sequence(
  total_throughput = 2,
  instruction_mix = {
    SHIFT: 1,
  }
)

model.new_sequence(
  total_throughput = 0.91,
  instruction_mix = {
    ALU:   2,
    SHIFT: 1,
  }
)


#model.new_sequence(
#  total_throughput = 3.1,
#  instruction_mix = {
#    ALU:   2,
#    SHIFT: 2,
#  }
#)
#
#s = model.new_sequence(
#  total_throughput = None,
#  instruction_mix = {
#    ALU:   4,
#    SHIFT: 2,
#  }
#)



if __name__ == '__main__':
  import argparse
  import os.path
  parser = argparse.ArgumentParser(os.path.basename(sys.argv[0]))

  parser.add_argument('-o', '--output', dest='FILE', help='Output file')

  args = parser.parse_args()
  resolve_model(model)
  for muops, port in model.delta.__iter__():
      value = model.delta.__getitem__((muops, port)).value
      if value :
          print(str(muops) +" associated with " + str( port))



