from gurobipy import *
import itertools as it

"""
Modelizing the requests on memory hierarchy
For now only doing it for loads
"""

class Request:
    def __init__(self, name):
        self.name = name

class Sequence:
    def __init__(self, name, inst_mix, throughput = 1):
        self.name = name
        self.throughput = throughput
        self.inst_mix = inst_mix

# level is the level of cache in which data are (now we assume that we retrieve
# data from only one level of cache)
# cache retains the throughputs between each levels (so register/L1, L1/L2, L2/L3)
def resolve_model_memory_up(level, cache, locks):
    # For now we only deal with 3 level of hierarchy
    assert(len(cache) > 1 and len(cache) < 5)
    m = Model("throughput")
    # vars representing the throughputs obtained for a given sequence and link
    a = m.addVar(vtype=GRB.CONTINUOUS)
    # Constraints

    # Some levels of cache cannot be used at the same time
    # to implement this, we get cache as a list of tuples which represent
    # a lock. For example, (0, 1 ) means that the links between the lowest and
    # the second lowest level of cache cannot be used at the same time.
    # A lock can exist among any number of cache level (there could be 
    # a shared resource between all levels of cache)
    for n in range(level + 1):
        for comb in it.combinations(range(level), n):
            if any(all(i in l for i in comb) for l in locks):
                m.addConstr(quicksum([a / cache[i] for i in comb]) <= 1)
        

    # END Constraints

    m.update()
    # Objective
    # We want the number of connections in the graph
    m.setObjective(a, GRB.MAXIMIZE)
    m.optimize()
    if m.status == GRB.OPTIMAL:
        print("--------------")
        print("Optimal solution found !")

    return a.X

def resolve_arch_up_2levels(seq_list, throughputs, request_list,  nb_rsrc, nb_locks):
    # For now we only deal with 3 level of hierarchy
    m = Model("throughput")
    """
     Variables
    """
    # d var are binary vars that represents the fact that there is a link or not
    # between a request type and a resource
    dreq_res = dict()
    # preq_res represents the actual throughput between a request "entry" and a 
    # resource
    for r in request_list:
        dreq_res[r] = []
        for _ in range(nb_rsrc):
            dreq_res[r].append(m.addVar(vtype=GRB.BINARY))




    dres_locks = []
    for r in range(nb_rsrc):
        dres_locks.append([])
        for _ in range(nb_locks):
            dres_locks[r].append(m.addVar(vtype=GRB.BINARY))

    preq_res = [None] * len(seq_list)
    pres_locks = [[]] * len(seq_list)
    for s, seq in enumerate(seq_list):
        preq_res[s] = dict()
        for r in request_list:
            preq_res[s][r] = []
            for _ in range(nb_rsrc):
                preq_res[s][r].append(m.addVar(vtype=GRB.CONTINUOUS))



        for r in range(nb_rsrc):
            pres_locks[s].append([])
            for _ in range(nb_locks):
                pres_locks[s][r].append(m.addVar(vtype=GRB.CONTINUOUS))


    """
     Constraints
    """

    for s, seq in enumerate(seq_list):
        for r in request_list:
            req_ratio = seq.inst_mix[r] / sum(seq.inst_mix.values())
            for i in range(nb_rsrc):
                m.addConstr(preq_res[s][r][i]  <= dreq_res[r][i])
                #m.addConstr(quicksum(preq_res[s][r][i] for i in range(nb_rsrc)) >= req_ratio * throughputs[seq])
                #m.addConstrs(preq_res[s][r][i] >= req_ratio * throughputs[seq] if (dreq_res[r][i] == 1))
                m.addConstr( (dreq_res[r][i] == 1) >> (preq_res[s][r][i] >= req_ratio * throughputs[seq]))


        for l in range(nb_rsrc):
            for lock in range(nb_locks):
                # gurobi is cryptic with min
                # Apparently you cannot compose a sum and a min directly
                # So we use temporary variable and artificial constraints
                min_temp2 = m.addVar(vtype=GRB.CONTINUOUS)
                add_temp = m.addVar(vtype=GRB.CONTINUOUS)
                m.addConstr(add_temp == quicksum(preq_res[s][r][l] for r in request_list))
                m.addConstr(min_temp2 == min_(add_temp, dres_locks[l][lock]))
                m.addConstr(pres_locks[s][l][lock] == min_temp2)

        for l in range(nb_rsrc):
            m.addConstr(quicksum(preq_res[s][r][l] for r in request_list) <= 1)
        for lock in range(nb_locks):
            m.addConstr(quicksum(pres_locks[s][i][lock] for i in range(nb_rsrc)) <= 1)

        #m.addConstr(quicksum(preq_res[s][r][i] for r in request_list for i in
        #    range(nb_rsrc)) >= throughputs[seq])


    """
     Objectives
    """
    # We want to maximize the constraints on the graph
    # That is, get a graph that allows the througput we gave it as input,
    # but no more
    # So we minimize the number of links in first level and maximize the 
    # number of locks
    # Not sure about how the fact that resources can be concurrent and that
    # not complementary is well represented in a three level model though
    m.setObjective( quicksum(dreq_res[r][i] for r in request_list for i in
        range(nb_rsrc))  -
        quicksum(dres_locks[j][lock] for j in range(nb_rsrc) for lock
            in range(nb_locks)) , GRB.MINIMIZE)
    m.update()
    # We want the number of connections in the graph
    m.optimize()
    if m.status == GRB.OPTIMAL:
        print("--------------")
        print("Optimal solution found !")
        print("requests -- level_1 resources")
        for r in request_list:
            for i in range(nb_rsrc):
                print("request {} resource {} : {}".format(r.name, i, dreq_res[r][i].X))
        print(" locks")
        for lock in range(nb_locks):
            l_res = []
            for j in range(nb_rsrc):
                if dres_locks[j][lock].X > 0.2:
                    l_res.append(j)
            print("lock  {} : {}".format( lock, l_res))

        if True:
            for s, seq in enumerate(seq_list):
                print(seq.name)
                print(seq.inst_mix)
                for r in request_list:
                    for i in range(nb_rsrc):
                        print("request {} resource {} : {}".format(r.name, i, preq_res[s][r][i].X))
                print("resources -- locks")
                for j in range(nb_rsrc):
                    for lock in range(nb_locks):
                        print("resource {} lock {} : {}".format(j, lock, pres_locks[s][j][lock].X))

# Here we modelize memory with a 3 level graph
# Some constraints become more natural to exress - to be verified
def resolve_arch_up_3levels(seq_list, throughputs, request_list,  nb_level_1_rsrc, nb_level_2_rsrc, nb_locks):
    m = Model("throughput")
    """
     Variables
    """
    # d var are binary vars that represents the fact that there is a link or not
    # between a request type and a resource
    dreq_res1 = dict()
    basename = "dlta_req_res1_"
    for r in request_list:
        dreq_res1[r] = []
        for i in range(nb_level_1_rsrc):
            dreq_res1[r].append(m.addVar(vtype=GRB.BINARY,name= basename + r.name + "_" + str(i) ))

    dres1_res2 = []
    basename2 = "dlta_res1_res2_"
    for r in range(nb_level_1_rsrc):
        dres1_res2.append([])
        for _ in range(nb_level_2_rsrc):
            dres1_res2[r].append(m.addVar(vtype=GRB.BINARY, name= basename2 + str(r) + "_" + str(i)))



    dres2_locks = []
    basename3 = "dres2_locks_"
    dres2 = []
    for r in range(nb_level_2_rsrc):
        dres2_locks.append([])
        dres2.append(m.addVar(vtype=GRB.BINARY))
        for l in range(nb_locks):
            dres2_locks[r].append(m.addVar(vtype=GRB.BINARY, name= basename2 + str(r) + "_" + str(l)))

    dlocks = []
    for l in range(nb_locks):
        dlocks.append(m.addVar(vtype=GRB.BINARY))


    preq_res1 = [None] * len(seq_list)
    pres1_res2 = [[]] * len(seq_list)
    pres2_locks = [[]] * len(seq_list)
    for s, seq in enumerate(seq_list):
        preq_res1[s] = dict()
        bname = "preq_res1_{}_".format(seq.name)
        for r in request_list:
            preq_res1[s][r] = []
            for i in range(nb_level_1_rsrc):
                preq_res1[s][r].append(m.addVar(vtype=GRB.CONTINUOUS, name = bname + r.name + "_" + str(i)))

        bname2 = "pres1_res2_{}_".format(seq.name)
        for r in range(nb_level_1_rsrc):
            pres1_res2[s].append([])
            for i in range(nb_level_2_rsrc):
                pres1_res2[s][r].append(m.addVar(vtype=GRB.CONTINUOUS, name= bname2 + str(r) + "_" + str(i)))



        bname3 = "pres2_lock_{}_".format(seq.name)
        for r in range(nb_level_2_rsrc):
            pres2_locks[s].append([])
            for l in range(nb_locks):
                pres2_locks[s][r].append(m.addVar(vtype=GRB.CONTINUOUS, name= bname3 + str(r) + "_" + str(l)))




    """
     Constraints
    """

    for s, seq in enumerate(seq_list):
        for r in request_list:
            req_ratio = seq.inst_mix[r] / sum(seq.inst_mix.values())
            print("REQ_RATIO: {}".format(req_ratio))
            print("THROUGHPUT: {}".format(throughputs[seq]))
            print("REQUEST THROUGHPUT: {}".format(req_ratio * throughputs[seq]))
            if( req_ratio > 0):
                m.addConstr(quicksum(dreq_res1[r][i] for i in range(nb_level_1_rsrc)) >= 1)
            for i in range(nb_level_1_rsrc):

                m.addConstr( (dreq_res1[r][i] == 1) >> (preq_res1[s][r][i] >= req_ratio * throughputs[seq]))

        for l1 in range(nb_level_1_rsrc):
            m.addConstr(quicksum(preq_res1[s][r][l1] for r in request_list) ==
                    quicksum(pres1_res2[s][l1][j] for j in
                        range(nb_level_2_rsrc)))
            for l2 in range(nb_level_2_rsrc):
                m.addConstr(pres1_res2[s][l1][l2]  <= dres1_res2[l1][l2])

        for l2 in range(nb_level_2_rsrc):
            m.addConstr(quicksum(pres1_res2[s][i][l2] for i in range(nb_level_1_rsrc)) <= 1)

        for l2 in range(nb_level_2_rsrc):
            for lock in range(nb_locks):
                # gurobi is cryptic with min
                # Apparently you cannot compose a sum and a min directly
                # So we use temporary variable and artificial constraints 
                min_temp2 = m.addVar(vtype=GRB.CONTINUOUS)
                add_temp = m.addVar(vtype=GRB.CONTINUOUS)
                m.addConstr(add_temp == quicksum(pres1_res2[s][l1][l2] for l1 in range(nb_level_1_rsrc)))
                m.addConstr(min_temp2 == min_(add_temp, dres2_locks[l2][lock]))
                m.addConstr(pres2_locks[s][l2][lock] == min_temp2)

        for lock in range(nb_locks):
            m.addConstr(quicksum(pres2_locks[s][i][lock] for i in range(nb_level_2_rsrc)) <= 1)
    for l2 in range(nb_level_2_rsrc):
        m.addGenConstrOr(dres2[l2], [dres1_res2[l1][l2] for l1 in range(nb_level_1_rsrc)])
        m.addConstr( (dres2[l2] == 0) >> (quicksum(dres2_locks[l2][lock] for lock in range(nb_locks)) == 0))



    for lock in range(nb_locks):
        m.addConstr( (dlocks[lock] == 1) >> (quicksum(dres2_locks[i][lock] for i in range(nb_level_2_rsrc)) >= 2))
        m.addConstr( (dlocks[lock] == 0) >> (quicksum(dres2_locks[i][lock] for i in range(nb_level_2_rsrc)) == 0))


    """
     Objectives
    """
    # We want to maximize the constraints on the graph
    # That is, get a graph that allows the througput we gave it as input,
    # but no more
    # So we minimize the number of links in first level and maximize the 
    # number of locks
    m.setObjective( - quicksum( dreq_res1[r][i] for r in request_list for i in
        range(nb_level_1_rsrc)) + quicksum(dres1_res2[i][j] for i in
            range(nb_level_1_rsrc) for j in range(nb_level_2_rsrc)) -
        quicksum(dres2_locks[j][lock] for j in range(nb_level_2_rsrc) for lock
            in range(nb_locks)) , GRB.MINIMIZE)
    m.update()
    # Objective
    # We want the number of connections in the graph
    m.optimize()
    if m.status == GRB.OPTIMAL:
        print("--------------")
        print("Optimal solution found !")
        print("requests -- level_1 resources")
        for r in request_list:
            for i in range(nb_level_1_rsrc):
                print("request {} resource {} : {}".format(r.name, i, dreq_res1[r][i].X))
        print("level 1 resources -- level 2 resources")
        for i in range(nb_level_1_rsrc):
            for j in range(nb_level_2_rsrc):
                print("resource_1 {} resource_2 {} : {}".format(i, j, dres1_res2[i][j].X))
        print("locks:")
        for lock in range(nb_locks):
            l_res = []
            for j in range(nb_level_2_rsrc):
                if dres2_locks[j][lock].X > 0.2:
                    l_res.append(j)
            print("lock  {} : {}".format( lock, l_res))
        if False:
            for s, seq in enumerate(seq_list):
                print(seq.name)
                print([(dic[0].name, dic[1]) for dic in seq.inst_mix.items()])
                print("GLOBAL THROUGHPUT: {}".format(throughputs[seq]))
                for r in request_list:
                    for i in range(nb_level_1_rsrc):
                        print("request {} resource {} : {}".format(r.name, i, preq_res1[s][r][i].X))
                print("level 1 resources -- level 2 resources")
                for i in range(nb_level_1_rsrc):
                    for j in range(nb_level_2_rsrc):
                        print("resource_1 {} resource_2 {} : {}".format(i, j, pres1_res2[s][i][j].X))
                print("level 2 resources -- locks")
                for j in range(nb_level_2_rsrc):
                    for lock in range(nb_locks):
                        print("resource_2 {} lock {} : {}".format(j, lock, pres2_locks[s][j][lock].X))
    # Model is not feasible, we write a log
    else:
        m.computeIIS()
        m.update()
        m.write("arch3.ilp")




if __name__ == "__main__":
    cache = [ 1, 1, 1]
    lock = [(0, 1, 2)]
    #res = resolve_model_memory_up(2, cache, lock)
    req_l1 = Request("L1")
    req_l2 = Request("L2")
    req_l3 = Request("L3")
    req_list = [req_l1, req_l2, req_l3]
    seq1 = Sequence("seq1",{ req_l1: 2, req_l2: 0, req_l3: 0})
    seq2 = Sequence("seq2",{ req_l1: 0, req_l2: 1, req_l3: 0})
    seq3 = Sequence("seq3",{ req_l1: 0, req_l2: 0, req_l3: 1})
    seq4 = Sequence("seq4",{ req_l1: 1, req_l2: 1, req_l3: 0})
    seq5 = Sequence("seq5",{ req_l1: 1, req_l2: 0, req_l3: 1})
    seq6 = Sequence("seq6",{ req_l1: 0, req_l2: 1, req_l3: 1})
    seq7 = Sequence("seq7",{ req_l1: 1, req_l2: 1, req_l3: 1})
    seqs_throughputs = {seq1: 2, seq2: 1, seq3: 1, seq4: 1 ,seq5: 2, seq6: 1, seq7: 1}
    #seq_list = [seq1, seq2, seq3, seq4, seq5, seq6, seq7]
    #seq_list = [seq1, seq2, seq4]
    seq_list = [seq1, seq2, seq3, seq7]
    #seq_list = [seq1]
    resolve_arch_up_3levels(seq_list, seqs_throughputs, req_list,  3, 3, 4)
    #print("--------------------------------------")
    #print("STARTING 2 LEVEL")
    #print("--------------------------------------")
    #resolve_arch_up_2levels(seq_list, seqs_throughputs, req_list,   2, 4)

