"""""""""""""""
This file intends to test our graph generation
We want to generate random graphs, calculate a throughput
for some sequences and see if we can rebuild the original
graph from the fake measurement
"""""""""""""""
from gurobipy import *
import random as rnd

class Instruction:
    def __init__(self, name):
        self.name = name


class CPUGraph:
    def __init__(self, insts, nb_muops, nb_ports, links_iu, links_up):
        self.insts = insts
        self.nb_insts = len(insts)
        self.nb_muops = nb_muops
        self.nb_ports = nb_ports
        self.links_iu = links_iu
        self.links_up = links_up

    def get_throughput(self, sequence_mix):
        assert all(i in self.insts for i in sequence_mix)
        m = Model("throughput")
        m.setParam('OutputFlag', 0)
        iu_vars = []
        ins_var = []
        muop_set = set()
        for ni, i in enumerate(sequence_mix):
            indi = self.insts.index(i)
            for u in range(self.nb_muops):
                if self.links_iu[indi][u]:
                    muop_set.add(u)

        for ni, i in enumerate(sequence_mix):
            ins_var.append(m.addVar(vtype=GRB.CONTINUOUS))
            iu_vars.append([])
            for nu, u in enumerate(muop_set):
                iu_vars[ni].append(m.addVar(vtype= GRB.CONTINUOUS))
        up_vars = []
        port_set = set()
        for nu, u in enumerate(muop_set):
            for p in range(self.nb_ports):
                if self.links_up[u][p]:
                    port_set.add(p)

        for nu, u in enumerate(muop_set):
            up_vars.append([])
            for p in enumerate(port_set):
                up_vars[nu].append(m.addVar(vtype= GRB.CONTINUOUS))

        unit_throughput = m.addVar(vtype= GRB.CONTINUOUS)

        m.update()
        # Constraints
        for ni, i in enumerate(sequence_mix):
            #for nu, u in enumerate(muop_set):
            indi = self.insts.index(i)
            for nu, u in enumerate(muop_set):
                if self.links_iu[indi][u]:
                    m.addConstr(iu_vars[ni][nu] == ins_var[ni])
                else:
                    # Making sure that if there is no link
                    # Then the throughput is 0
                    m.addConstr(iu_vars[ni][nu] == 0)
        for nu, u in enumerate(muop_set):
            for np, p in enumerate(port_set):
                if not self.links_up[u][p]:
                    m.addConstr(up_vars[nu][np] == 0)

        for ni, v in enumerate(sequence_mix.values()):
            m.addConstr(ins_var[ni] == v * unit_throughput)



        for nu, u in enumerate(muop_set):
            m.addConstr(quicksum(iu_vars[ni][nu] for ni in range(len(sequence_mix))) == quicksum(up_vars[nu]))

        for np, p in enumerate(port_set):
            m.addConstr(quicksum(up_vars[nu][np] for nu in range(len(muop_set))) <= 1)


        # Objective
        m.setObjective(quicksum(ins_var), GRB.MAXIMIZE)
        m.optimize()

        if m.status == GRB.OPTIMAL:
            total_throughput = 0
            return sum([ins_var[ni].X for ni in range(len(sequence_mix))])
        else:
            print("Model unfeasible")
            raise ValueError
        def __str__(self):
            return " insts: {}, {} muops, {} ports,\n links iu: {},\n links up: {}\n".format(self.insts, self.nb_muops, self.nb_ports,[i.name + " " +str(u) + " "+ str(self.iu[i][u]) for i in self.insts for u in range(self.nb_ports)] , self.links_up)
            

def gen_arch(insts, nb_muops, nb_ports):
    nb_insts = len(insts)
    def check_arch(iu, up):
        return all([any([(iu[i][u] and any(up[u][p] for p in range(nb_ports))) for u in range(nb_muops) ]) for i in range(nb_insts)]) and all(any(up[u]) for u in range(nb_muops))
    while True:
        di_u = [[(rnd.randint(0,1) == 0) for u in range(nb_muops)] for i in range(nb_insts)]
        du_p = [[(rnd.randint(0,1) == 0) for p in range(nb_ports)] for u in range(nb_muops)]
        if check_arch(di_u, du_p):
            break
    #print(di_u)
    #print(du_p)
    return (di_u, du_p)

if __name__ == "__main__":
    di_u, du_p = gen_arch(["add", "mul"], 2, 2)
    insts = ["add", "mul"]
    diu = [[False, False, False, True], [True, True, True, False]]
    dup = [[False, True, False, True], [True, False, True, True],[False,False, True, True],[True,True, True, True ]]
    graph = CPUGraph(insts, 4, 4, diu, dup)
    seq_mix = { "mul": 1}
    print(graph.get_throughput(seq_mix))
    #graph.get_throughput(seq_mix)



