from z3 import *
import math as m

class Instruction:
    def __init__(self, name):
        self.name = name

class Sequence:
    def __init__(self, inst_mix, throughput):
        self.inst_mix = inst_mix
        self.throughput = throughput

def resolve_model_inst_mix( seqs, insts, nb_muops, nb_ports, multiplicities):
     # Delta variables representing instructions and micro-instructions
    dinst_muop = dict()
    basename = "diu_"
    for inst in insts:
        dinst_muop[inst] = []
        for u in range(nb_muops):
            dinst_muop[inst].append(Bool(basename + inst.name + "_" + str(u)))

    dmuop_port = []

    # Links between muops and  ports
    basename = "dup_"
    for u in range(nb_muops):
        dmuop_port.append([])
        for p in range(nb_ports):
            dmuop_port[u].append(Bool(basename + str(u) + "_" + str(p)))

    # Is inst connected to port ?
    dinst_port = dict()
    basename = "dip_"
    for inst in insts:
        dinst_port[inst] = []
        for p in range(nb_ports):
            dinst_port[inst].append(Int(basename + inst.name + "_" + str(p)))

    # Is inst connected to port ?
    iinst_muop = dict()
    basename = "iiu_"
    for inst in insts:
        iinst_muop[inst] = []
        for u in range(nb_muops):
            iinst_muop[inst].append(Int(basename + inst.name + "_" + str(u)))

    imuop_port = []
    basename ="iup_"
    for u in range(nb_muops):
        imuop_port.append([])
        for p in range(nb_ports):
            imuop_port[u].append(Int(basename + str(u) + "_" + str(p)))
    # Throughputs
    pinst_muop = []
    pmuop_port = []
    for s, seq in enumerate(seqs):
        pinst_muop.append([])
        pinst_muop[s] = dict()
        basename = "piu_{}_".format(s)
        for inst in insts:
            pinst_muop[s][inst] = []
            for u in range(nb_muops):
                pinst_muop[s][inst].append(Real(basename + inst.name + "_" + str(u)))

        pmuop_port.append([])

        # Links between muops and  ports
        basename = "pup_{}_".format(s)
        for u in range(nb_muops):
            pmuop_port[s].append([])
            for p in range(nb_ports):
                pmuop_port[s][u].append(Real(basename + str(u) + "_" + str(p)))

    def connect(i, p):
        return Or([And(dinst_muop[i][u], dmuop_port[u][p]) for u in range(nb_muops)])


    def multiplicity(inst, mul):
        """
         Simplest and least realistic constraint (also most constraining)
         Makes every muop linked to inst links to exactly the number of ports
         to get the required througput
        """
        epsilon = 0.5
        #return (Implies( mul > 0, (Sum(dinst_port[inst]) / Sum(iinst_muop[inst]))   >= mul))
        return (Implies( mul > 0, And((Sum(dinst_port[inst]) / Sum(iinst_muop[inst]))   >= mul - epsilon, (Sum(dinst_port[inst]) / Sum(iinst_muop[inst]))   <= mul + epsilon)))

    def relax_multiplicity(inst, mul):
        """
        a more complicated constraint
        Imposes that every muop can support at least the required througput
        """
        C1 = (And([Implies(dinst_muop[inst][u], Sum(imuop_port[u]) >= mul) for u in range(nb_muops)]))
        """
        Another, imposing that at least one muop
        is linked to exactly the right number of port to get the required throughput
        """
        #C2 = (Or([Implies(dinst_muop[inst][u],
        #        Sum(imuop_port[u]) <= m.ceil(mul)) for u in range(nb_muops)]))
        C2 = Or([And(dinst_muop[inst][u], Sum(imuop_port[u]) <= m.ceil(mul)) for u in range(nb_muops)])

        C12 = And(C1, C2)
        #return (And( And([Implies(dinst_muop[inst][u], Sum(imuop_port[u]) >=
        #    mul) for u in range(nb_muops)])), Or([Implies(dinst_muop[inst][u],
        #        Sum(imuop_port[u]) == mul) for u in range(nb_muops)]))
        #return C1
        return C12

    # Initializing Solver
    #solv = Solver()
    solv = Optimize()


    """
     Constraints
    """

    # This one is a dummy constraint
    # Just put dinst_port[inst][p] at 1 if inst is indeed connected to port
    # 0 elsewise
    for i in insts:
        for p in range(nb_ports):
            solv.add(If(connect(i, p), dinst_port[i][p] == 1, dinst_port[i][p] == 0))

    # Same as above, we need to manipulate integer at some point instead of Bool
    for i in insts:
        solv.add(Or([dinst_muop[i][u] for u in range(nb_muops)]))
        for u in range(nb_muops):
            solv.add(If(dinst_muop[i][u], iinst_muop[i][u] == 1, iinst_muop[i][u] == 0))

    # Now For muop:ports
    for u in range(nb_muops):
        for p in range(nb_ports):
            solv.add(If(dmuop_port[u][p], imuop_port[u][p] == 1, imuop_port[u][p] == 0))

    for i in insts:
        solv.add(multiplicity(i, multiplicities[i]))
        solv.add(relax_multiplicity(i, multiplicities[i]))

    for s, seq in enumerate(seqs):
        # All througputs are positive
        for i in insts:
            for u in range(nb_muops):
                solv.add(pinst_muop[s][i][u] >= 0)

        # A link can have a throughput strictly greater than 0 only if there is
        # indeed a link there
        for i in insts:
            for u in range(nb_muops):
                solv.add(Implies(Not(dinst_muop[i][u]), pinst_muop[s][i][u] == 0))

        # All througputs are positive
        for u in range(nb_muops):
            for p in range(nb_ports):
                solv.add(pmuop_port[s][u][p] >= 0)

        # A link can have a throughput strictly greater than 0 only if there is
        # indeed a link there
        for u in range(nb_muops):
            for p in range(nb_ports):
                #solv.add(Implies(pmuop_port[s][u][p] > 0, dmuop_port[u][p]))
                solv.add(Implies(Not(dmuop_port[u][p]), pmuop_port[s][u][p] == 0))

        for i in insts:
            ratio = float(seq.inst_mix[i]) / sum(seq.inst_mix.values()) if i in seq.inst_mix else 0
            for u in range(nb_muops):
                solv.add(Implies(dinst_muop[i][u], pinst_muop[s][i][u] == ratio * seq.throughput ))

        for u in range(nb_muops):
            solv.add(Sum([pinst_muop[s][i][u] for i in insts]) == Sum([pmuop_port[s][u][p] for p in range(nb_ports)]))

        for p in range(nb_ports):
            solv.add(Sum([pmuop_port[s][u][p] for u in range(nb_muops)]) <= 1)

    # Objective
    # Which one should be the first ? No obvious answer
    solv.minimize(Sum([imuop_port[u][p] for u in range(nb_muops) for p in range(nb_ports)]))
    solv.maximize(Sum([iinst_muop[i][u] for i in insts for u in range(nb_muops)]))

    if solv.check() == sat:
        #print(["inst {} muop {} : {}".format(i.name, u, solv.model()[dinst_muop[i][u]]) for i in insts for u in range(nb_muops)])
        #print(["muop {} port {} : {}".format(u, p, solv.model()[dmuop_port[u][p]]) for p in range(nb_ports) for u in range(nb_muops)])
        diu_ret = [ [solv.model()[dinst_muop[i][u]] for u in range(nb_muops)] for i in insts]
        dup_ret = [ [solv.model()[dmuop_port[u][p]] for p in range(nb_ports)] for u in range(nb_muops)]
        return diu_ret, dup_ret
    else:
        print("No solution !")
        raise ValueError

# End resolve_model_inst_mix


if __name__ == "__main__":
    i1 = Instruction("add")
    i2 = Instruction("mul")
    multiple = dict()
    multiple[i1] = 2
    multiple[i2] = 1
    s1 = Sequence({i1: 1}, 1)
    s2 = Sequence({i2: 1}, 1)
    s3 = Sequence({i1: 1, i2: 1}, 2)
    seqs = [s1, s2, s3]
    insts = [i1, i2]
    resolve_model_inst_mix(seqs, insts, 2, 2, multiple )
