///
/// Helpers for emitting instruction descriptions pipedream can understand
///

#pragma once

#include "XED_Info.hpp"
#include <map>
#include <string>
#include <set>
#include <optional>

struct Pipedream_Operand_Template;

using Pipedream_Operand_Vector = std::vector<Pipedream_Operand_Template>;

struct Pipedream {
  /// @name types
  /// @{

  enum Type {
    /// floating point
    F16,
    F32,
    F64,
    F80,

    /// signed int
    I1,
    I8,
    I16,
    I32,
    I64,

    /// unsigned int
    U4,
    U8,
    U16,
    U32,
    U64,
    U128,
    U256,

    /// binary coded decimal
    BCD80,

    /// too complex or variable sized
    OTHER,
  };

  static Type type(xed_operand_element_xtype_enum_t xtype, xed_operand_element_type_enum_t element_type, unsigned element_bits) {
    auto int_for_bits = [element_bits]() -> Type {
      switch (element_bits) {
        case 1:  return I1;
        case 8:  return I8;
        case 16: return I16;
        case 32: return I32;
        case 64: return I64;
      }
      abort();
    };
    auto uint_for_bits = [element_bits]() -> Type {
      switch (element_bits) {
        case 8:   return U8;
        case 16:  return U16;
        case 32:  return U32;
        case 64:  return U64;
        case 128: return U128;
        case 256: return U256;
      }
      abort();
    };

    switch (xtype) {
      case XED_OPERAND_XTYPE_B80: return BCD80;

      case XED_OPERAND_XTYPE_F16: return F16;
      case XED_OPERAND_XTYPE_F32: return F32;
      case XED_OPERAND_XTYPE_F64: return F64;
      case XED_OPERAND_XTYPE_F80: return F80;

      case XED_OPERAND_XTYPE_I1:  return I1;
      case XED_OPERAND_XTYPE_I8:  return I8;
      case XED_OPERAND_XTYPE_I16: return I16;
      case XED_OPERAND_XTYPE_I32: return I32;
      case XED_OPERAND_XTYPE_I64: return I64;
      case XED_OPERAND_XTYPE_INT: return int_for_bits();

      case XED_OPERAND_XTYPE_U8:   return U8;
      case XED_OPERAND_XTYPE_U16:  return U16;
      case XED_OPERAND_XTYPE_U32:  return U32;
      case XED_OPERAND_XTYPE_U64:  return U64;
      case XED_OPERAND_XTYPE_U128: return U128;
      case XED_OPERAND_XTYPE_U256: return U256;
      case XED_OPERAND_XTYPE_UINT: return uint_for_bits();

      case XED_OPERAND_XTYPE_STRUCT:
      case XED_OPERAND_XTYPE_VAR:
        return OTHER;

      case XED_OPERAND_XTYPE_INVALID:
        switch (element_type) {
          case XED_OPERAND_ELEMENT_TYPE_UINT:       return uint_for_bits();
          case XED_OPERAND_ELEMENT_TYPE_INT:        return int_for_bits();
          case XED_OPERAND_ELEMENT_TYPE_SINGLE:     return F32;
          case XED_OPERAND_ELEMENT_TYPE_DOUBLE:     return F64;
          case XED_OPERAND_ELEMENT_TYPE_LONGDOUBLE: return F80;
          case XED_OPERAND_ELEMENT_TYPE_FLOAT16:    return F16;
          case XED_OPERAND_ELEMENT_TYPE_LONGBCD:    return BCD80;

          case XED_OPERAND_ELEMENT_TYPE_STRUCT:
          case XED_OPERAND_ELEMENT_TYPE_VARIABLE:
            return OTHER;

          case XED_OPERAND_ELEMENT_TYPE_INVALID:
          case XED_OPERAND_ELEMENT_TYPE_LAST:
          default:
            abort();
        }

      case XED_OPERAND_XTYPE_LAST:
      default:
        abort();
    }
  }

  static const char *type_name(Type t) {
    switch (t) {
      case F16:   return "f16";
      case F32:   return "f32";
      case F64:   return "f64";
      case F80:   return "f80";
      case I1:    return "i1";
      case I8:    return "i8";
      case I16:   return "i16";
      case I32:   return "i32";
      case I64:   return "i64";
      case U4:    return "u4";
      case U8:    return "u8";
      case U16:   return "u16";
      case U32:   return "u32";
      case U64:   return "u64";
      case U128:  return "u128";
      case U256:  return "u256";
      case BCD80: return "bcd80";
      case OTHER: return "OTHER";
    }
    abort();
  }

  static unsigned type_bits(Type t) {
    switch (t) {
      case F16:   return 16;
      case F32:   return 32;
      case F64:   return 64;
      case F80:   return 80;
      case I1:    return 1;
      case I8:    return 8;
      case I16:   return 16;
      case I32:   return 32;
      case I64:   return 64;
      case U4:    return 4;
      case U8:    return 8;
      case U16:   return 16;
      case U32:   return 32;
      case U64:   return 64;
      case U128:  return 128;
      case U256:  return 256;
      case BCD80: return 80;
      case OTHER: abort();
    }
    abort();
  }

  /// @}

  /// @name CPU flags
  /// @{

  /// flags for CPU flags register
  enum class Flag : uint64_t {
    CF   = 0x0000'0000'0001, /// Carry flag
    PF   = 0x0000'0000'0004, /// Parity flag
    AF   = 0x0000'0000'0010, /// Adjust flag
    ZF   = 0x0000'0000'0040, /// Zero flag
    SF   = 0x0000'0000'0080, /// Sign flag
    TF   = 0x0000'0000'0100, /// Trap flag
    IF   = 0x0000'0000'0200, /// Interrupt enable flag
    DF   = 0x0000'0000'0400, /// Direction flag
    OF   = 0x0000'0000'0800, /// Overflow flag
    IOPL = 0x0000'0000'3000, /// I/O privilege level (286+ only)
    NT   = 0x0000'0000'4000, /// Nested task flag (286+ only)
    /// EFLAGS
    RF   = 0x0000'0001'0000, /// Resume flag (386+ only)
    VM   = 0x0000'0002'0000, /// Virtual 8086 mode flag (386+ only)
    AC   = 0x0000'0004'0000, /// Alignment check (486SX+ only)
    VIF  = 0x0000'0008'0000, /// Virtual interrupt flag (Pentium+)
    VIP  = 0x0000'0010'0000, /// Virtual interrupt pending (Pentium+)
    ID   = 0x0000'0020'0000, /// Able to use CPUID instruction (Pentium+)
    /// RFLAGS

    /// X87 FLAGS
    FC0 = 0x0001'0000'0001,
    FC1 = 0x0001'0000'0002,
    FC2 = 0x0001'0000'0004,
    FC3 = 0x0001'0000'0008,
  };

  using Flag_Set = uint64_t;

  static const char *flag_name(Flag flag) {
    switch (flag) {
      case Flag::CF:   return "CF";
      case Flag::PF:   return "PF";
      case Flag::AF:   return "AF";
      case Flag::ZF:   return "ZF";
      case Flag::SF:   return "SF";
      case Flag::TF:   return "TF";
      case Flag::IF:   return "IF";
      case Flag::DF:   return "DF";
      case Flag::OF:   return "OF";
      case Flag::IOPL: return "IOPL";
      case Flag::NT:   return "NT";
      case Flag::RF:   return "RF";
      case Flag::VM:   return "VM";
      case Flag::AC:   return "AC";
      case Flag::VIF:  return "VIF";
      case Flag::VIP:  return "VIP";
      case Flag::ID:   return "ID";
      case Flag::FC0:  return "FC0";
      case Flag::FC1:  return "FC1";
      case Flag::FC2:  return "FC2";
      case Flag::FC3:  return "FC3";
    }
    abort();
  }

  static Flag flag_first() { return Flag::CF; }

  static std::optional<Flag> flag_next(Flag flag) {
    switch (flag) {
      case Flag::CF:   return Flag::PF;
      case Flag::PF:   return Flag::AF;
      case Flag::AF:   return Flag::ZF;
      case Flag::ZF:   return Flag::SF;
      case Flag::SF:   return Flag::TF;
      case Flag::TF:   return Flag::IF;
      case Flag::IF:   return Flag::DF;
      case Flag::DF:   return Flag::OF;
      case Flag::OF:   return Flag::IOPL;
      case Flag::IOPL: return Flag::NT;
      case Flag::NT:   return Flag::RF;
      case Flag::RF:   return Flag::VM;
      case Flag::VM:   return Flag::AC;
      case Flag::AC:   return Flag::VIF;
      case Flag::VIF:  return Flag::VIP;
      case Flag::VIP:  return Flag::ID;
      case Flag::ID:   return Flag::FC0;
      case Flag::FC0:  return Flag::FC1;
      case Flag::FC1:  return Flag::FC2;
      case Flag::FC2:  return Flag::FC3;
      case Flag::FC3:  return std::nullopt;
    }
    abort();
  }

  static void print_flag_set(Flag_Set flags, std::ostream &O, const char *name_prefix = "") {
    std::vector<const char *> names;

    for (std::optional<Flag> it = Pipedream::flag_first(); it; it = Pipedream::flag_next(*it)) {
      if (Pipedream::Flag_Set(*it) & flags) {
        names.push_back(Pipedream::flag_name(*it));
      }
    }

    if (names.empty()) {
      O << "0";
    } else {
      O << name_prefix << names.front();

      for (auto it = names.begin() + 1, end = names.end(); it != end; it++) {
        O << " | " << name_prefix << *it;
      }
    }
  }

  /// @}

  /// @name registers
  /// @{

  static const std::string &register_class_name(Register_Class reg_class);

  static Span<xed_reg_enum_t[1]> singleton_register_classes();

  static Register_Class singleton_register_class(xed_reg_enum_t reg) ;

  static bool is_vector_register_class(Register_Class reg_class);

  static bool is_scalar_register_class(Register_Class reg_class);

  static bool is_segment_register_class(Register_Class reg_class);

  static const char *register_name(xed_reg_enum_t reg) {
    switch (reg) {
      case XED_REG_ST0: return "ST0";
      case XED_REG_ST1: return "ST1";
      case XED_REG_ST2: return "ST2";
      case XED_REG_ST3: return "ST3";
      case XED_REG_ST4: return "ST4";
      case XED_REG_ST5: return "ST5";
      case XED_REG_ST6: return "ST6";
      case XED_REG_ST7: return "ST7";
      default:          return xed_reg_enum_t2str(reg);
   }
  }

  /// @}

  static std::set<std::string> tags(XED_Decoded_Instruction inst, const Pipedream_Operand_Vector &operands);
};


/// describes one *value* in machine code (a register, an immediate, a memory reference)
struct Machine_Value {
  enum Kind {
    REG,
    IMM,
    /// calculate address but do not access it (LEA, BND, ...)
    ADDR,
    /// load/store
    MEM,
    /// read/write CPU flags
    FLAGS,
  };

  Machine_Value() : _kind{IMM}, _imm_bits{0} {}

  static Machine_Value make_reg(Register_Class reg_class, xed_reg_enum_t reg) {
    Machine_Value v;
    v._kind          = REG;
    v._reg.reg       = reg;
    v._reg.reg_class = reg_class;
    return v;
  }

  static Machine_Value make_reg(Register_Class reg_class) {
    Machine_Value v;
    v._kind          = REG;
    v._reg.reg       = std::nullopt;
    v._reg.reg_class = reg_class;
    return v;
  }

  static Machine_Value make_imm(uint64_t imm_bits) {
    Machine_Value v;
    v._kind     = IMM;
    v._imm_bits = imm_bits;
    return v;
  }

  static Machine_Value make_addr(uint64_t address_bits) {
    Machine_Value v;
    v._kind             = ADDR;
    v._mem.address_bits = address_bits;
    v._mem.memory_bits  = 0;
    v._mem.base_reg     = std::nullopt;
    return v;
  }

  static Machine_Value make_mem(uint64_t address_bits, uint64_t memory_bits, std::optional<xed_reg_enum_t> base) {
    Machine_Value v;
    v._kind             = MEM;
    v._mem.address_bits = address_bits;
    v._mem.memory_bits  = memory_bits;
    v._mem.base_reg     = base;
    return v;
  }

  static Machine_Value make_flags(xed_reg_enum_t reg, Pipedream::Flag_Set read, Pipedream::Flag_Set written) {
    Machine_Value v;
    v._kind        = FLAGS;
    v._flags.reg   = reg;
    v._flags.read  = read;
    v._flags.write = written;
    return v;
  }

  Kind kind() const { return _kind; }

  std::optional<xed_reg_enum_t> reg() const {
    assert(kind() == REG);
    return _reg.reg;
  }
  Register_Class reg_class() const {
    assert(kind() == REG);
    return _reg.reg_class;
  }

  uint64_t imm_bits() const {
    assert(kind() == IMM);
    return _imm_bits;
  }

  uint64_t address_bits() const {
    assert(kind() == MEM || kind() == ADDR);
    return _mem.address_bits;
  }
  uint64_t memory_bits() const {
    assert(kind() == MEM);
    return _mem.memory_bits;
  }
  std::optional<xed_reg_enum_t> base_reg() const {
    assert(kind() == MEM || kind() == ADDR);
    return _mem.base_reg;
  }

  xed_reg_enum_t flag_reg() const {
    assert(kind() == FLAGS);
    return _flags.reg;
  }
  Pipedream::Flag_Set flags_read() const {
    assert(kind() == FLAGS);
    return _flags.read;
  }
  Pipedream::Flag_Set flags_written() const {
    assert(kind() == FLAGS);
    return _flags.write;
  }

  bool operator<(const Machine_Value &that) const {
    return compare(that, std::less<void>());
  }

  bool operator==(const Machine_Value &that) const {
    return compare(that, std::equal_to<void>());
  }

  bool operator!=(const Machine_Value &that) const {
    return !(*this == that);
  }

  friend std::ostream &operator<<(std::ostream &O, const Machine_Value &v) {
    switch (v.kind()) {
      case REG:
        O << "reg(" << Pipedream::register_class_name(v.reg_class());
        if (auto reg = v.reg()) {
          O << ", " << Pipedream::register_name(*reg);
        }
        O << ")";
        return O;
      case IMM:
        return O << "imm(" << v.imm_bits() << ")";
      case ADDR:
        return O << "addr(" << v.address_bits() << ")";
      case MEM:
        O << "mem(" << v.address_bits() << ", " << v.memory_bits();
        if (auto reg = v.base_reg()) {
          O << ", " << Pipedream::register_name(*reg);
        }
        O << ")";
        return O;
      case FLAGS:
        O << "flags(" << Pipedream::register_name(v.flag_reg()) << ", ";

        Pipedream::print_flag_set(v.flags_read(), O);
        O << ", ";
        Pipedream::print_flag_set(v.flags_written(), O);

        O << ")";
        return O;
    }
    return O << "invalid";
  }
private:
  template<typename Op>
  bool compare(const Machine_Value &that, Op &&compare) const {
    if (_kind == that._kind) {
      switch (_kind) {
        case REG:
          return compare(std::tie(_reg.reg_class, _reg.reg), std::tie(that._reg.reg_class, that._reg.reg));
        case IMM:
          return compare(_imm_bits, that._imm_bits);
        case ADDR:
        case MEM:
          return compare(std::tie(_mem.address_bits, _mem.memory_bits, _mem.base_reg), std::tie(that._mem.address_bits, that._mem.memory_bits, that._mem.base_reg));
        case FLAGS:
          return compare(std::tie(_flags.reg, _flags.read, _flags.write), std::tie(that._flags.reg, that._flags.read, that._flags.write));
      }
      abort();
    } else {
      return compare(_kind, that._kind);
    }
  }

  Kind _kind;

  union {
    struct {
      Register_Class reg_class;
      std::optional<xed_reg_enum_t> reg;
    } _reg;
    uint64_t _imm_bits;
    struct {
      /// width of memory address in bits
      uint64_t address_bits;
      /// number of bits loaded/stored
      uint64_t memory_bits;
      std::optional<xed_reg_enum_t> base_reg;
    } _mem;
    struct {
      xed_reg_enum_t reg;
      Pipedream::Flag_Set read;
      Pipedream::Flag_Set write;
    } _flags;
  };
};

/// An operand of an instruction in the pipedream machine description
struct Pipedream_Operand_Template {
  xed_operand_enum_t name;

  xed_operand_visibility_enum_t visibility;

  Machine_Value value;
  xed_operand_action_enum_t action;

  size_t num_elements;
  Pipedream::Type type;

  std::optional<xed_reg_enum_t> reg() const {
    switch (value.kind()) {
      case Machine_Value::REG:   return value.reg();
      case Machine_Value::FLAGS: return value.flag_reg();
      default: return std::nullopt;
    }
  }

  bool operator<(const Pipedream_Operand_Template &that) const {
    return compare(that, std::less<void>());
  }
  bool operator==(const Pipedream_Operand_Template &that) const {
    return compare(that, std::equal_to<void>());
  }
  bool operator!=(const Pipedream_Operand_Template &that) const {
    return compare(that, std::not_equal_to<void>());
  }

  std::string type_name() const {
    std::string name = Pipedream::type_name(type);
    if (num_elements > 1) {
      name += "x";
      name += std::to_string(num_elements);
    }
    return name;
  }

  std::string short_name() const {
    switch (value.kind()) {
      case Machine_Value::REG:
        if (auto reg = value.reg()) {
          return Pipedream::register_name(*reg) + type_name();
        } else {
          return Pipedream::register_class_name(value.reg_class()) + type_name();
        }
      case Machine_Value::IMM:
        return "IMM" + type_name();
      case Machine_Value::ADDR:
        return "ADDR" + std::to_string(value.address_bits()) + type_name();
      case Machine_Value::MEM:
        return "MEM" + std::to_string(value.address_bits()) + type_name();
      case Machine_Value::FLAGS:
        return "FLAGS";
    }
    abort();
  }

  friend std::ostream &operator<<(std::ostream &O, const Pipedream_Operand_Template &op) {
    const char *name       = xed_operand_enum_t2str(op.name);
    const char *visibility = xed_operand_visibility_enum_t2str(op.visibility);
    const char *action     = xed_operand_action_enum_t2str(op.action);
    const char *type       = Pipedream::type_name(op.type);

    return O << "Pipedream_Operand_Template" << std::tie(name, visibility, op.value, action, op.num_elements, type);
  }
private:
  template<typename Op>
  bool compare(const Pipedream_Operand_Template &that, Op &&compare) const {
    const auto tie = [](const Pipedream_Operand_Template &a) {
      return std::tie(a.name, a.visibility, a.value, a.action, a.num_elements, a.type);
    };

    return compare(tie(*this), tie(that));
  }
};

using Pipedream_Operand_Vector = std::vector<Pipedream_Operand_Template>;

/// Wrapper around a XED instruction that we can handle (we managed to find a valid encoding).
struct Pipedream_Instruction {
  /// @name compare by these
  /// @{

  std::string name;
  xed_iclass_enum_t iclass;
  XED_CPU_Mode cpu_mode;

  std::string att_mnemonic;
  std::string intel_mnemonic;

  Pipedream_Operand_Vector operands = {};

  /// @}

  bool operator<(const Pipedream_Instruction &that) const {
    auto tie = [](const Pipedream_Instruction &a){
      return std::tie(a.iclass, a.cpu_mode, a.att_mnemonic, a.operands);
    };

    return tie(*this) < tie(that);
  }
  bool operator!=(const Pipedream_Instruction &that) const {
    auto tie = [](const Pipedream_Instruction &a){
      return std::tie(a.iclass, a.cpu_mode, a.att_mnemonic, a.operands);
    };

    return (tie(*this) < tie(that)) || (tie(that) < tie(*this));
  }
};