///
/// Extract Pipedream instruction descriptions for Intel X86 from Intel XED.
///

/// TODO: print more instruction info (see xed/examples/xed-ex1.c):
///         * flags read/written
///         * rep/repne prefix (F2, F3)
///         * address size prefix (67)
///         * the various 66 prefixes
///         * exceptions
///         * is broadcast
///         * AVX mask behaviour
///         * required CPU info bits
///         * valid chipsets

#include "XED_Info.hpp"
#include "Pipedream_Info.hpp"
#include "utils.hpp"
#include <algorithm>
#include <cassert>
#include <cinttypes>
#include <cstdio>
#include <cstring>
#include <functional>
#include <iomanip>
#include <iostream>
#include <set>
#include <string>
#include <vector>
#include <xed/xed-interface.h>

static const Pipedream pipedream;

struct Debug {
  static void print_operand(std::ostream &O, const XED_Instruction_Operand op) {
    print_operand_common(O, op);

    switch (op.type()) {
      case XED_OPERAND_TYPE_NT_LOOKUP_FN:
        O << "lookup: " << op.get_lookup_fn_name() << " ";
        break;
      case XED_OPERAND_TYPE_REG:
        O << "reg: " << op.fixed_reg_str() << " ";
        break;
      case XED_OPERAND_TYPE_IMM_CONST:
        O << "imm: " << op.imm_const() << " ";
        break;

      case XED_OPERAND_TYPE_IMM:
      case XED_OPERAND_TYPE_NT_LOOKUP_FN4:
      case XED_OPERAND_TYPE_INVALID:
      case XED_OPERAND_TYPE_ERROR:
      case XED_OPERAND_TYPE_LAST:
        break;

      default:
        abort();
    }
  }

  static void print_operand(std::ostream &O, const XED_Decoded_Instruction_Operand op) {
    print_operand_common(O, op.static_op());

    switch (op.type()) {
      case XED_OPERAND_TYPE_NT_LOOKUP_FN:
        O << "lookup: " << op.nonterminal_str() << " ";
        O << "reg: " << op.reg_str() << " ";
        break;
      case XED_OPERAND_TYPE_REG:
        O << "reg: " << op.reg_str() << " ";
        break;
      case XED_OPERAND_TYPE_IMM_CONST:
        O << "imm: " << op.imm_const() << " ";
        break;

      case XED_OPERAND_TYPE_IMM:
      case XED_OPERAND_TYPE_NT_LOOKUP_FN4:
      case XED_OPERAND_TYPE_INVALID:
      case XED_OPERAND_TYPE_ERROR:
      case XED_OPERAND_TYPE_LAST:
        break;

      default:
        abort();
    }
  }

  static void print_operand(std::ostream &O, const xed_encoder_operand_t &op) {
    switch (op.type) {
      case XED_ENCODER_OPERAND_TYPE_REG:
        O << "REG=" << Pipedream::register_name(op.u.reg);
        break;
      case XED_ENCODER_OPERAND_TYPE_MEM:
        O << "MEM" << op.width_bits;
        O << "(";
        if (op.u.mem.base) {
          O << "base="  << Pipedream::register_name(op.u.mem.base) << ", ";
        }
        if (op.u.mem.index) {
          O << "index=" << Pipedream::register_name(op.u.mem.index) << ", ";
        }
        if (op.u.mem.disp.displacement_bits) {
          O << "disp="  << op.u.mem.disp.displacement_bits << "/" << op.u.mem.disp.displacement << ", ";
        }
        O << ")";
        break;
      case XED_ENCODER_OPERAND_TYPE_IMM0:
        O << "IMM" << op.width_bits << "=" << op.u.imm0;
        break;
      case XED_ENCODER_OPERAND_TYPE_IMM1:
        O << "IMM" << op.width_bits << "=" << op.u.imm1;
        break;
      default:
        abort();
    }
  }

  static void print_inst(std::ostream &O, const XED_Instruction inst) {
    O << "  XED_INST\n";
    print_inst_commmon(O, inst);

    O << "    #operands:         " << inst.num_operands() << "\n";
    for(unsigned j = 0; j < inst.num_operands(); j++) {
      O << "      " << j << " ";
      print_operand(O, inst.operand(j));
      O << "\n";
    }
  }

  static void print_inst(std::ostream &O, const XED_Decoded_Instruction dec) {
    O << "  XED_DECODED_INST\n";
    O << "    machine-mode:            " << dec.get_machine_mode_bits() << "\n";
    O << "    stack-address-mode:      " << dec.get_stack_address_bits() << "\n";
    O << "    effective-operand-width: " << dec.effective_operand_width() << "\n";
    O << "    effective-address-width: " << dec.effective_address_width() << "\n";
    print_inst_commmon(O, dec.inst());

    O << "    #operands:         " << dec.num_operands() << "\n";
    for(unsigned j = 0; j < dec.num_operands(); j++) {
      O << "      " << j << " ";
      print_operand(O, dec.operand(j));
      O << "\n";
    }
  }

  static void print_inst_commmon(std::ostream &O, const XED_Instruction inst) {
    O << "    iclass                   " << inst.iclass_str() << " (" << inst.iclass() << ")\n";
    O << "    iform                    " << inst.iform_str() << "\n";
    O << "    category                 " << inst.category_str() << "\n";
    O << "    extension                " << inst.extension_str() << "\n";
    O << "    ISA                      " << inst.isa_set_str() << "\n";

    O << "    attributes:         [";
    for(const char *attr : inst.attribute_strs()) {
      O << attr << ", ";
    }
    O << "]\n";
  }

  static void print_operand_common(std::ostream &O, const XED_Instruction_Operand op) {
    O << "name: " << op.name_str() << " ";
    O << "visibility: " << op.visibility_str() << " ";
    O << "action: " << op.action_str() << " ";
    O << "type: " << op.type_str() << " ";
    O << "xtype: " << op.xtype_str() << " ";
  }

  static void print_encoder_state(std::ostream &O, const XED_Encoder_State &state) {
    O << "XED_ENCODER_STATE:";
    O << "  machine-mode:            " << state.cpu_mode.mode_str() << "\n";
    O << "  stack-address-mode:      " << state.cpu_mode.stack_addr_width_str() << "\n";
    O << "  effective-operand-width: " << state.effective_operand_width << "\n";
    O << "  effective-address-width: " << state.effective_address_width << "\n";
    O << "  allow-rex:               " << (state.allow_rex ? "YES" : "NO") << "\n";
  }
};

struct Convert {
  static std::optional<Machine_Value> machine_value_from_xed(XED_Instruction inst, XED_Instruction_Operand op, const XED_Encoder_State &state, unsigned memory_width) {
    if (op.is_reg_op() || op.is_base_op()) {
      if (auto reg = op.get_reg(state)) {
        assert(reg != XED_REG_INVALID);
        return Machine_Value::make_reg(pipedream.singleton_register_class(*reg), *reg);
      } else if (auto rc = op.get_reg_class(state)) {
        return Machine_Value::make_reg(*rc);
      }
    } else if (op.is_memory_op()) {
      return Machine_Value::make_mem(state.effective_address_width, memory_width, std::nullopt);
    } else if (op.is_address_generate_op()) {
      return Machine_Value::make_addr(state.effective_address_width);
    } else if (op.is_imm0_op()) {
      const unsigned imm_bits = xed_operand_width_bits(op.raw(), state.xed_eosz());

      return Machine_Value::make_imm(imm_bits);
    }

    /// FIXME: can't handle yet
    (void) inst;
//    if (op.is_lookup_fn()) {
//      std::cerr << "# TODO: " << inst.iform_str() << " OP " << op.name_str() << " LOOKUP_FN: " << op.get_lookup_fn_name() << "\n";
//    } else {
//      std::cerr << "# TODO: " << inst.iform_str() << " OP " << op.name_str() << " " << op.type_str() << "\n";
//    }
    return std::nullopt;
  }

  static std::optional<Machine_Value> machine_value_from_xed(XED_Decoded_Instruction inst,
                                                             XED_Decoded_Instruction_Operand dec_op,
                                                             const XED_Encoder_State &state) {
    XED_Instruction_Operand op = dec_op.static_op();

    if (op.is_reg_op()) {
      if (auto reg = op.get_reg(state)) {
        assert(reg != XED_REG_INVALID);

        switch (*reg) {
          case XED_REG_FLAGS:
          case XED_REG_EFLAGS:
          case XED_REG_RFLAGS:
          {
            const auto read  = read_flags_from_xed(inst);
            const auto write = write_flags_from_xed(inst);

            /// TODO: some shifts & rotates never read/write flags depending on machine mode.
            ///       XED only knows this after decoding. Drop the flags operand in this case.
            // assert(read | write);

            return Machine_Value::make_flags(*reg, read, write);
          }
          default:
            return Machine_Value::make_reg(pipedream.singleton_register_class(*reg), *reg);
        }
      } else if (auto rc = op.get_reg_class(state)) {
        return Machine_Value::make_reg(*rc);
      }
    }

    if (op.is_memory_op()) {
      unsigned idx = dec_op.memory_operand_index();

      std::optional<xed_reg_enum_t> base_reg = inst.get_base_reg(idx);

      return Machine_Value::make_mem(dec_op.memory_address_bits(), dec_op.memory_operand_bits(), base_reg);
    } else if (op.is_address_generate_op()) {
      return Machine_Value::make_addr(dec_op.memory_address_bits());
    } else if (op.is_imm0_op()) {
      return Machine_Value::make_imm(dec_op.element_bits());
    }

    return std::nullopt;
  }

  static Pipedream_Operand_Template pipedream_operand_from_xed(XED_Decoded_Instruction inst,
                                                               XED_Decoded_Instruction_Operand xed_dec_op,
                                                               const XED_Encoder_State &state) {
    const XED_Instruction_Operand xed_op = xed_dec_op.static_op();

    const xed_operand_element_xtype_enum_t xed_xtype = xed_dec_op.element_xtype();
    const xed_operand_element_type_enum_t  xed_type  = xed_dec_op.element_type();

    unsigned bits = xed_dec_op.element_bits();
    unsigned num_elements = xed_dec_op.num_elements();

    const std::optional<Machine_Value> value = machine_value_from_xed(inst, xed_dec_op, state);

    if (!value) {
      auto &O = std::cerr;
      O << "TODO: COULD NOT REPRESENT OPERAND:";
      O << " " << inst.inst().iclass_str();
      O << " " << inst.inst().iform_str();
      O << " " << xed_op.name_str();
      O << " " << xed_op.type_str();
      if (xed_op.is_lookup_fn()) {
        O << " " << xed_op.get_lookup_fn_name();
      }
      O << "\n";
      abort();
    }

    Pipedream::Type type = Pipedream::OTHER;
    xed_operand_visibility_enum_t visibility = xed_dec_op.visibility();

    switch (value->kind()) {
      case Machine_Value::REG:
        switch (value->reg().value_or(XED_REG_INVALID)) {
          case XED_REG_FLAGS:
          case XED_REG_EFLAGS:
          case XED_REG_RFLAGS:
            std::cerr << "FLAGS/EFLAGS/RFLAGS in REG operand, should be FLAGS\n";
            abort();
            break;

          case XED_REG_X87STATUS:
          case XED_REG_X87TAG:
          case XED_REG_X87CONTROL:
            type         = Pipedream::I16;
            num_elements = 1; /// X87TAG: actually 8 2-bit tags
            bits         = 16;
            break;

//            case XED_REG_STACKPOP:
//            case XED_REG_STACKPUSH:

          case XED_REG_X87POP:
          case XED_REG_X87PUSH:
            type         = Pipedream::F80;
            num_elements = 1;
            bits         = 80;
            break;
          case XED_REG_X87POP2:
            type         = Pipedream::F80;
            num_elements = 2;
            bits         = 80;
            break;

          case XED_REG_ST0:
          case XED_REG_ST1:
          case XED_REG_ST2:
          case XED_REG_ST3:
          case XED_REG_ST4:
          case XED_REG_ST5:
          case XED_REG_ST6:
          case XED_REG_ST7:
            /// XED sometimes marks the second float stack operand implicit where it should be 'suppressed'
            switch (inst.inst().iclass()) {
              case XED_ICLASS_FCOM:
              case XED_ICLASS_FCOMP:
              case XED_ICLASS_FCOMI:
              case XED_ICLASS_FCOMIP:
              case XED_ICLASS_FCOMPP:
              case XED_ICLASS_FUCOM:
              case XED_ICLASS_FUCOMP:
              case XED_ICLASS_FUCOMI:
              case XED_ICLASS_FUCOMIP:
              case XED_ICLASS_FUCOMPP:
              case XED_ICLASS_FXCH:
              case XED_ICLASS_FLD:
                if (visibility != XED_OPVIS_EXPLICIT) {
                  visibility = XED_OPVIS_SUPPRESSED;
                }
                break;
              default:
                break;
            }
            break;

          case XED_REG_CR0:
            switch (state.cpu_mode.mode()) {
              case XED_MACHINE_MODE_LONG_64:
                /// FIXME: are there any other modes where CRO is 64 bits?
                type         = Pipedream::I64;
                num_elements = 1;
                bits         = 64;
                break;
              default:
                type         = Pipedream::I32;
                num_elements = 1;
                bits         = 32;
                break;
            }
            break;

          case XED_REG_GDTR:
          case XED_REG_LDTR:
          case XED_REG_IDTR:
          case XED_REG_TR:
            /// FIXME: introduce special machine types for these?
            type         = Pipedream::OTHER;
            num_elements = 1;
            break;

          case XED_REG_INVALID:
            /// Could be any register from register class, trust XED type info for operand.
          default:
            type = pipedream.type(xed_xtype, xed_type, bits);
            break;
        }
        break;
      case Machine_Value::ADDR:
        assert(bits == 0);
        bits = xed_dec_op.memory_address_bits();
        type = pipedream.type(xed_xtype, xed_type, bits);
        break;
      case Machine_Value::MEM:
        type = pipedream.type(xed_xtype, xed_type, bits);
        assert(value->memory_bits() == (num_elements * bits));
        break;
      case Machine_Value::IMM:
        type = pipedream.type(xed_xtype, xed_type, bits);
        break;
      case Machine_Value::FLAGS:
        type         = Pipedream::OTHER;
        bits         = 0;
        num_elements = 0;
        break;
    }

    Pipedream_Operand_Template op;

    assert((type == Pipedream::OTHER) || bits == pipedream.type_bits(type));

    op.type = type;
    op.name = xed_op.name();
    op.visibility    = visibility;
    op.action        = xed_dec_op.action();
    op.num_elements  = num_elements;
    op.value = *value;

    return op;
  }

  static Pipedream_Operand_Vector pipedream_operands_from_xed(const XED_Decoded_Instruction &xed_inst, XED_CPU_Mode cpu_mode, bool allow_rex) {
    const Effective_Operand_Width effective_operand_width = xed_inst.effective_operand_width();
    const Effective_Address_Width effective_address_width = xed_inst.effective_address_width();

    const XED_Encoder_State state = {cpu_mode, effective_operand_width, effective_address_width, allow_rex};

    Pipedream_Operand_Vector operands;

    for (unsigned i = 0, end = xed_inst.num_operands(); i < end; i++) {
      const XED_Decoded_Instruction_Operand xed_op = xed_inst.operand(i);

      if (xed_op.static_op().is_memory_addressing_register()) {
        /// they are handled when the corresponding memory op gets handled
        continue;
      }

      auto op = pipedream_operand_from_xed(xed_inst, xed_op, state);

      bool ignore_op = false;

      switch (op.reg().value_or(XED_REG_INVALID)) {
        case XED_REG_STACKPOP:
        case XED_REG_STACKPUSH:
        case XED_REG_X87POP:
        case XED_REG_X87POP2:
        case XED_REG_X87PUSH:
          ignore_op = true;
          break;
        default:
          break;
      }

      if (ignore_op) { continue; }

      operands.push_back(op);
    }

    return std::move(operands);
  }

  /// Compute AT&T asm syntax instruction mnemonic from Pipedream_Instruction
  /// All fields except 'mnemonic' must already be filled in
  static std::string make_att_mnemonic(XED_Decoded_Instruction inst, const Pipedream_Operand_Vector &operands) {
    const std::string att_asm = inst.asm_str(XED_SYNTAX_ATT);

    std::string mnem;

    const size_t pos = att_asm.find(' ');
    if (pos != std::string::npos) {
      mnem = att_asm.substr(0, pos);
    } else {
      mnem = att_asm;
    }

    /// XED sometimes get's AT&T mnemonics wrong
    switch (inst.inst().iclass()) {
      case XED_ICLASS_MOVSXD:
      case XED_ICLASS_MOVSX:
        /// remove trailing 'x', AT&T syntax does NOT use it.
        mnem = "movs";
        goto type_suffix;
      case XED_ICLASS_MOVZX:
        /// remove trailing 'x', AT&T syntax does NOT use it.
        mnem = "movz";
        goto type_suffix;

      type_suffix:
        assert(operands.size() == 2);
        assert(operands[0].value.kind() == Machine_Value::REG);
        {
          /// add type suffix
          const char src = [&]() -> char {
            switch (inst.inst().iform()) {
              case XED_IFORM_MOVSX_GPRv_GPR8:  return 'b';
              case XED_IFORM_MOVSX_GPRv_GPR16: return 'w';
              case XED_IFORM_MOVSX_GPRv_MEMb:  return 'b';
              case XED_IFORM_MOVSX_GPRv_MEMw:  return 'w';

              case XED_IFORM_MOVZX_GPRv_GPR8:  return 'b';
              case XED_IFORM_MOVZX_GPRv_GPR16: return 'w';
              case XED_IFORM_MOVZX_GPRv_MEMb:  return 'b';
              case XED_IFORM_MOVZX_GPRv_MEMw:  return 'w';

              case XED_IFORM_MOVSXD_GPRv_GPR32: return 'l';
              case XED_IFORM_MOVSXD_GPRv_MEMd:  return 'l';

              default: abort();
            }
          }();

          const char dst = [&]() -> char {
            switch (operands[0].type) {
              case Pipedream::I16: return 'w';
              case Pipedream::I32: return 'l';
              case Pipedream::I64: return 'q';
              default: abort();
            }
          }();

          mnem += src;
          mnem += dst;
        }
        break;

      /// FIXME: XEDs type suffix for memory accesses has a meaning here.
      ///        (for example, XED can express a 32-bit address load that loads 64 bits of data)
      ///        But GAS does not understand that.
      case XED_ICLASS_ADDPD:
      case XED_ICLASS_ADDPS:
      case XED_ICLASS_ADDSD:
      case XED_ICLASS_ADDSS:
      case XED_ICLASS_ADDSUBPD:
      case XED_ICLASS_ADDSUBPS:
      case XED_ICLASS_AESDEC:
      case XED_ICLASS_AESDECLAST:
      case XED_ICLASS_AESENC:
      case XED_ICLASS_AESENCLAST:
      case XED_ICLASS_AESIMC:
      case XED_ICLASS_AESKEYGENASSIST:
      case XED_ICLASS_ANDNPD:
      case XED_ICLASS_ANDNPS:
      case XED_ICLASS_ANDPD:
      case XED_ICLASS_ANDPS:
      case XED_ICLASS_BLENDPD:
      case XED_ICLASS_BLENDPS:
      case XED_ICLASS_BLENDVPD:
      case XED_ICLASS_BLENDVPS:
      case XED_ICLASS_CMPPD:
      case XED_ICLASS_CMPPS:
      case XED_ICLASS_CMPSD:
      case XED_ICLASS_CMPSD_XMM:
      case XED_ICLASS_CMPSS:
      case XED_ICLASS_COMISD:
      case XED_ICLASS_COMISS:
      case XED_ICLASS_CVTDQ2PD:
      case XED_ICLASS_CVTDQ2PS:
      case XED_ICLASS_CVTPD2DQ:
      case XED_ICLASS_CVTPD2PI:
      case XED_ICLASS_CVTPD2PS:
      case XED_ICLASS_CVTPI2PD:
      case XED_ICLASS_CVTPI2PS:
      case XED_ICLASS_CVTPS2DQ:
      case XED_ICLASS_CVTPS2PD:
      case XED_ICLASS_CVTPS2PI:
      case XED_ICLASS_CVTSD2SI:
      case XED_ICLASS_CVTSD2SS:
      case XED_ICLASS_CVTSS2SD:
      case XED_ICLASS_CVTSS2SI:
      case XED_ICLASS_CVTTPD2DQ:
      case XED_ICLASS_CVTTPD2PI:
      case XED_ICLASS_CVTTPS2DQ:
      case XED_ICLASS_CVTTPS2PI:
      case XED_ICLASS_CVTTSD2SI:
      case XED_ICLASS_CVTTSS2SI:
      case XED_ICLASS_DIVPD:
      case XED_ICLASS_DIVPS:
      case XED_ICLASS_DIVSD:
      case XED_ICLASS_DIVSS:
      case XED_ICLASS_DPPD:
      case XED_ICLASS_DPPS:
      case XED_ICLASS_EXTRACTPS:
      case XED_ICLASS_HADDPD:
      case XED_ICLASS_HADDPS:
      case XED_ICLASS_HSUBPD:
      case XED_ICLASS_HSUBPS:
      case XED_ICLASS_INSERTPS:
      case XED_ICLASS_LAR:
      case XED_ICLASS_LDDQU:
      case XED_ICLASS_LSL:
      case XED_ICLASS_MASKMOVDQU:
      case XED_ICLASS_MASKMOVQ:
      case XED_ICLASS_MAXPD:
      case XED_ICLASS_MAXPS:
      case XED_ICLASS_MAXSD:
      case XED_ICLASS_MAXSS:
      case XED_ICLASS_MINPD:
      case XED_ICLASS_MINPS:
      case XED_ICLASS_MINSD:
      case XED_ICLASS_MINSS:
      case XED_ICLASS_MOVAPD:
      case XED_ICLASS_MOVAPS:
      case XED_ICLASS_MOVD:
      case XED_ICLASS_MOVDDUP:
      case XED_ICLASS_MOVDIRI:
      case XED_ICLASS_MOVDQA:
      case XED_ICLASS_MOVDQU:
      case XED_ICLASS_MOVHPD:
      case XED_ICLASS_MOVHPS:
      case XED_ICLASS_MOVLPD:
      case XED_ICLASS_MOVLPS:
      case XED_ICLASS_MOVNTDQ:
      case XED_ICLASS_MOVNTDQA:
      case XED_ICLASS_MOVNTI:
      case XED_ICLASS_MOVNTPD:
      case XED_ICLASS_MOVNTPS:
      case XED_ICLASS_MOVNTQ:
      case XED_ICLASS_MOVNTSD:
      case XED_ICLASS_MOVNTSS:
      case XED_ICLASS_MOVQ:
      case XED_ICLASS_MOVSB:
      case XED_ICLASS_MOVSD:
      case XED_ICLASS_MOVSD_XMM:
      case XED_ICLASS_MOVSHDUP:
      case XED_ICLASS_MOVSLDUP:
      case XED_ICLASS_MOVSQ:
      case XED_ICLASS_MOVSS:
      case XED_ICLASS_MOVSW:
      case XED_ICLASS_MOVUPD:
      case XED_ICLASS_MOVUPS:
      case XED_ICLASS_MPSADBW:
      case XED_ICLASS_MULPD:
      case XED_ICLASS_MULPS:
      case XED_ICLASS_MULSD:
      case XED_ICLASS_MULSS:
      case XED_ICLASS_ORPD:
      case XED_ICLASS_ORPS:
      case XED_ICLASS_PABSB:
      case XED_ICLASS_PABSD:
      case XED_ICLASS_PABSW:
      case XED_ICLASS_PACKSSDW:
      case XED_ICLASS_PACKSSWB:
      case XED_ICLASS_PACKUSDW:
      case XED_ICLASS_PACKUSWB:
      case XED_ICLASS_PADDB:
      case XED_ICLASS_PADDD:
      case XED_ICLASS_PADDQ:
      case XED_ICLASS_PADDSB:
      case XED_ICLASS_PADDSW:
      case XED_ICLASS_PADDUSB:
      case XED_ICLASS_PADDUSW:
      case XED_ICLASS_PALIGNR:
      case XED_ICLASS_PADDW:
      case XED_ICLASS_PAND:
      case XED_ICLASS_PANDN:
      case XED_ICLASS_PAVGB:
      case XED_ICLASS_PAVGW:
      case XED_ICLASS_PCMPEQB:
      case XED_ICLASS_PCMPEQD:
      case XED_ICLASS_PCMPEQQ:
      case XED_ICLASS_PCMPEQW:
      case XED_ICLASS_PCMPESTRI:
      case XED_ICLASS_PCMPESTRM:
      case XED_ICLASS_PCMPGTB:
      case XED_ICLASS_PCMPGTD:
      case XED_ICLASS_PCMPGTQ:
      case XED_ICLASS_PCMPGTW:
      case XED_ICLASS_PCMPISTRI:
      case XED_ICLASS_PCMPISTRM:
      case XED_ICLASS_PEXTRB:
      case XED_ICLASS_PEXTRD:
      case XED_ICLASS_PEXTRQ:
      case XED_ICLASS_PEXTRW:
      case XED_ICLASS_PF2ID:
      case XED_ICLASS_PF2IW:
      case XED_ICLASS_PFACC:
      case XED_ICLASS_PFADD:
      case XED_ICLASS_PFECMPEQ:
      case XED_ICLASS_PFECMPGE:
      case XED_ICLASS_PFECMPGT:
      case XED_ICLASS_PHADDD:
      case XED_ICLASS_PHADDSW:
      case XED_ICLASS_PHADDW:
      case XED_ICLASS_PHMINPOSUW:
      case XED_ICLASS_PHSUBD:
      case XED_ICLASS_PHSUBSW:
      case XED_ICLASS_PHSUBW:
      case XED_ICLASS_PINSRB:
      case XED_ICLASS_PINSRD:
      case XED_ICLASS_PINSRQ:
      case XED_ICLASS_PINSRW:
      case XED_ICLASS_PMADDUBSW:
      case XED_ICLASS_PMADDWD:
      case XED_ICLASS_PMAXSB:
      case XED_ICLASS_PMAXSD:
      case XED_ICLASS_PMAXSW:
      case XED_ICLASS_PMAXUB:
      case XED_ICLASS_PMAXUD:
      case XED_ICLASS_PMAXUW:
      case XED_ICLASS_PMINSB:
      case XED_ICLASS_PMINSD:
      case XED_ICLASS_PMINSW:
      case XED_ICLASS_PMINUB:
      case XED_ICLASS_PMINUD:
      case XED_ICLASS_PMINUW:
      case XED_ICLASS_PMOVSXBD:
      case XED_ICLASS_PMOVSXBQ:
      case XED_ICLASS_PMOVSXBW:
      case XED_ICLASS_PMOVSXDQ:
      case XED_ICLASS_PMOVSXWD:
      case XED_ICLASS_PMOVSXWQ:
      case XED_ICLASS_PMOVZXBD:
      case XED_ICLASS_PMOVZXBQ:
      case XED_ICLASS_PMOVZXBW:
      case XED_ICLASS_PMOVZXDQ:
      case XED_ICLASS_PMOVZXWD:
      case XED_ICLASS_PMOVZXWQ:
      case XED_ICLASS_PMULDQ:
      case XED_ICLASS_PMULHRSW:
      case XED_ICLASS_PMULHRW:
      case XED_ICLASS_PMULHUW:
      case XED_ICLASS_PMULHW:
      case XED_ICLASS_PMULLD:
      case XED_ICLASS_PMULLW:
      case XED_ICLASS_PMULUDQ:
      case XED_ICLASS_POR:
      case XED_ICLASS_PSADBW:
      case XED_ICLASS_PSHUFB:
      case XED_ICLASS_PSHUFD:
      case XED_ICLASS_PSHUFHW:
      case XED_ICLASS_PSHUFLW:
      case XED_ICLASS_PSHUFW:
      case XED_ICLASS_PSIGNB:
      case XED_ICLASS_PSIGND:
      case XED_ICLASS_PSIGNW:
      case XED_ICLASS_PSLLD:
      case XED_ICLASS_PSLLDQ:
      case XED_ICLASS_PSLLQ:
      case XED_ICLASS_PSLLW:
      case XED_ICLASS_PSRAD:
      case XED_ICLASS_PSRAW:
      case XED_ICLASS_PSRLD:
      case XED_ICLASS_PSRLQ:
      case XED_ICLASS_PSRLW:
      case XED_ICLASS_PSUBB:
      case XED_ICLASS_PSUBD:
      case XED_ICLASS_PSUBQ:
      case XED_ICLASS_PSUBSB:
      case XED_ICLASS_PSUBSW:
      case XED_ICLASS_PSUBUSB:
      case XED_ICLASS_PSUBUSW:
      case XED_ICLASS_PSUBW:
      case XED_ICLASS_PTEST:
      case XED_ICLASS_PUNPCKHBW:
      case XED_ICLASS_PUNPCKHDQ:
      case XED_ICLASS_PUNPCKHQDQ:
      case XED_ICLASS_PUNPCKHWD:
      case XED_ICLASS_PUNPCKLBW:
      case XED_ICLASS_PUNPCKLDQ:
      case XED_ICLASS_PUNPCKLQDQ:
      case XED_ICLASS_PUNPCKLWD:
      case XED_ICLASS_PXOR:
      case XED_ICLASS_RCPPS:
      case XED_ICLASS_RCPSS:
      case XED_ICLASS_ROUNDPD:
      case XED_ICLASS_ROUNDPS:
      case XED_ICLASS_ROUNDSD:
      case XED_ICLASS_ROUNDSS:
      case XED_ICLASS_RSQRTPS:
      case XED_ICLASS_RSQRTSS:
      case XED_ICLASS_SHUFPD:
      case XED_ICLASS_SHUFPS:
      case XED_ICLASS_SQRTPD:
      case XED_ICLASS_SQRTPS:
      case XED_ICLASS_SQRTSS:
      case XED_ICLASS_SQRTSD:
      case XED_ICLASS_SUBPD:
      case XED_ICLASS_SUBPS:
      case XED_ICLASS_SUBSD:
      case XED_ICLASS_SUBSS:
      case XED_ICLASS_UCOMISD:
      case XED_ICLASS_UCOMISS:
      case XED_ICLASS_UNPCKHPD:
      case XED_ICLASS_UNPCKHPS:
      case XED_ICLASS_UNPCKLPD:
      case XED_ICLASS_UNPCKLPS:
      case XED_ICLASS_VADDPD:
      case XED_ICLASS_VADDPS:
      case XED_ICLASS_VADDSD:
      case XED_ICLASS_VADDSS:
      case XED_ICLASS_VADDSUBPD:
      case XED_ICLASS_VADDSUBPS:
      case XED_ICLASS_VANDNPD:
      case XED_ICLASS_VANDNPS:
      case XED_ICLASS_VANDPD:
      case XED_ICLASS_VANDPS:
      case XED_ICLASS_VBLENDMPD:
      case XED_ICLASS_VBLENDMPS:
      case XED_ICLASS_VBLENDPD:
      case XED_ICLASS_VBLENDPS:
      case XED_ICLASS_VBLENDVPD:
      case XED_ICLASS_VBLENDVPS:
      case XED_ICLASS_VBROADCASTF128:
      case XED_ICLASS_VBROADCASTI128:
      case XED_ICLASS_VBROADCASTSD:
      case XED_ICLASS_VBROADCASTSS:
      case XED_ICLASS_VCMPPD:
      case XED_ICLASS_VCMPPS:
      case XED_ICLASS_VCMPSD:
      case XED_ICLASS_VCMPSS:
      case XED_ICLASS_VCVTDQ2PD:
      case XED_ICLASS_VCVTDQ2PS:
      case XED_ICLASS_VCVTPD2DQ:
      case XED_ICLASS_VCVTPD2PS:
      case XED_ICLASS_VCVTPD2QQ:
      case XED_ICLASS_VCVTPD2UDQ:
      case XED_ICLASS_VCVTPD2UQQ:
      case XED_ICLASS_VCVTPH2PS:
      case XED_ICLASS_VCVTPS2DQ:
      case XED_ICLASS_VCVTPS2PD:
      case XED_ICLASS_VCVTPS2PH:
      case XED_ICLASS_VCVTPS2QQ:
      case XED_ICLASS_VCVTPS2UDQ:
      case XED_ICLASS_VCVTPS2UQQ:
      case XED_ICLASS_VCVTQQ2PD:
      case XED_ICLASS_VCVTQQ2PS:
      case XED_ICLASS_VCVTSD2SI:
      case XED_ICLASS_VCVTSD2SS:
      case XED_ICLASS_VCVTSD2USI:
      case XED_ICLASS_VCVTSI2SD:
      case XED_ICLASS_VCVTSI2SS:
      case XED_ICLASS_VCVTSS2SD:
      case XED_ICLASS_VCVTSS2SI:
      case XED_ICLASS_VCVTSS2USI:
      case XED_ICLASS_VCVTTPD2DQ:
      case XED_ICLASS_VCVTTPD2QQ:
      case XED_ICLASS_VCVTTPD2UDQ:
      case XED_ICLASS_VCVTTPD2UQQ:
      case XED_ICLASS_VCVTTPS2DQ:
      case XED_ICLASS_VCVTTPS2QQ:
      case XED_ICLASS_VCVTTPS2UDQ:
      case XED_ICLASS_VCVTTPS2UQQ:
      case XED_ICLASS_VCVTTSD2SI:
      case XED_ICLASS_VCVTTSD2USI:
      case XED_ICLASS_VCVTTSS2SI:
      case XED_ICLASS_VCVTTSS2USI:
      case XED_ICLASS_VCVTUDQ2PD:
      case XED_ICLASS_VCVTUDQ2PS:
      case XED_ICLASS_VCVTUQQ2PD:
      case XED_ICLASS_VCVTUQQ2PS:
      case XED_ICLASS_VCVTUSI2SD:
      case XED_ICLASS_VCVTUSI2SS:
      case XED_ICLASS_VCOMISS:
      case XED_ICLASS_VCOMISD:
      case XED_ICLASS_VDIVPD:
      case XED_ICLASS_VDIVPS:
      case XED_ICLASS_VDIVSD:
      case XED_ICLASS_VDIVSS:
      case XED_ICLASS_VDPPD:
      case XED_ICLASS_VDPPS:
      case XED_ICLASS_VEXTRACTF128:
      case XED_ICLASS_VEXTRACTF32X4:
      case XED_ICLASS_VEXTRACTF32X8:
      case XED_ICLASS_VEXTRACTF64X2:
      case XED_ICLASS_VEXTRACTF64X4:
      case XED_ICLASS_VEXTRACTI128:
      case XED_ICLASS_VEXTRACTI32X4:
      case XED_ICLASS_VEXTRACTI32X8:
      case XED_ICLASS_VEXTRACTI64X2:
      case XED_ICLASS_VEXTRACTI64X4:
      case XED_ICLASS_VEXTRACTPS:
      case XED_ICLASS_VFMADD132PD:
      case XED_ICLASS_VFMADD132PS:
      case XED_ICLASS_VFMADD132SD:
      case XED_ICLASS_VFMADD132SS:
      case XED_ICLASS_VFMADD213PD:
      case XED_ICLASS_VFMADD213PS:
      case XED_ICLASS_VFMADD213SD:
      case XED_ICLASS_VFMADD213SS:
      case XED_ICLASS_VFMADD231PD:
      case XED_ICLASS_VFMADD231PS:
      case XED_ICLASS_VFMADD231SD:
      case XED_ICLASS_VFMADD231SS:
      case XED_ICLASS_VFMADDPD:
      case XED_ICLASS_VFMADDPS:
      case XED_ICLASS_VFMADDSD:
      case XED_ICLASS_VFMADDSS:
      case XED_ICLASS_VFMADDSUB132PD:
      case XED_ICLASS_VFMADDSUB132PS:
      case XED_ICLASS_VFMADDSUB213PD:
      case XED_ICLASS_VFMADDSUB213PS:
      case XED_ICLASS_VFMADDSUB231PD:
      case XED_ICLASS_VFMADDSUB231PS:
      case XED_ICLASS_VFMADDSUBPD:
      case XED_ICLASS_VFMADDSUBPS:
      case XED_ICLASS_VFMSUB132PD:
      case XED_ICLASS_VFMSUB132PS:
      case XED_ICLASS_VFMSUB132SD:
      case XED_ICLASS_VFMSUB132SS:
      case XED_ICLASS_VFMSUB213PD:
      case XED_ICLASS_VFMSUB213PS:
      case XED_ICLASS_VFMSUB213SD:
      case XED_ICLASS_VFMSUB213SS:
      case XED_ICLASS_VFMSUB231PD:
      case XED_ICLASS_VFMSUB231PS:
      case XED_ICLASS_VFMSUB231SD:
      case XED_ICLASS_VFMSUB231SS:
      case XED_ICLASS_VFMSUBADD132PD:
      case XED_ICLASS_VFMSUBADD132PS:
      case XED_ICLASS_VFMSUBADD213PD:
      case XED_ICLASS_VFMSUBADD213PS:
      case XED_ICLASS_VFMSUBADD231PD:
      case XED_ICLASS_VFMSUBADD231PS:
      case XED_ICLASS_VFMSUBADDPD:
      case XED_ICLASS_VFMSUBADDPS:
      case XED_ICLASS_VFMSUBPD:
      case XED_ICLASS_VFMSUBPS:
      case XED_ICLASS_VFMSUBSD:
      case XED_ICLASS_VFMSUBSS:
      case XED_ICLASS_VFNMADD132PD:
      case XED_ICLASS_VFNMADD132PS:
      case XED_ICLASS_VFNMADD132SD:
      case XED_ICLASS_VFNMADD132SS:
      case XED_ICLASS_VFNMADD213PD:
      case XED_ICLASS_VFNMADD213PS:
      case XED_ICLASS_VFNMADD213SD:
      case XED_ICLASS_VFNMADD213SS:
      case XED_ICLASS_VFNMADD231PD:
      case XED_ICLASS_VFNMADD231PS:
      case XED_ICLASS_VFNMADD231SD:
      case XED_ICLASS_VFNMADD231SS:
      case XED_ICLASS_VFNMADDPD:
      case XED_ICLASS_VFNMADDPS:
      case XED_ICLASS_VFNMADDSD:
      case XED_ICLASS_VFNMADDSS:
      case XED_ICLASS_VFNMSUB132PD:
      case XED_ICLASS_VFNMSUB132PS:
      case XED_ICLASS_VFNMSUB132SD:
      case XED_ICLASS_VFNMSUB132SS:
      case XED_ICLASS_VFNMSUB213PD:
      case XED_ICLASS_VFNMSUB213PS:
      case XED_ICLASS_VFNMSUB213SD:
      case XED_ICLASS_VFNMSUB213SS:
      case XED_ICLASS_VFNMSUB231PD:
      case XED_ICLASS_VFNMSUB231PS:
      case XED_ICLASS_VFNMSUB231SD:
      case XED_ICLASS_VFNMSUB231SS:
      case XED_ICLASS_VFNMSUBPD:
      case XED_ICLASS_VFNMSUBPS:
      case XED_ICLASS_VFNMSUBSD:
      case XED_ICLASS_VFNMSUBSS:
      case XED_ICLASS_VFRCZPD:
      case XED_ICLASS_VFRCZPS:
      case XED_ICLASS_VHADDPD:
      case XED_ICLASS_VHADDPS:
      case XED_ICLASS_VHSUBPD:
      case XED_ICLASS_VHSUBPS:
      case XED_ICLASS_VINSERTF128:
      case XED_ICLASS_VINSERTF32X4:
      case XED_ICLASS_VINSERTF32X8:
      case XED_ICLASS_VINSERTF64X2:
      case XED_ICLASS_VINSERTF64X4:
      case XED_ICLASS_VINSERTI128:
      case XED_ICLASS_VINSERTI32X4:
      case XED_ICLASS_VINSERTI32X8:
      case XED_ICLASS_VINSERTI64X2:
      case XED_ICLASS_VINSERTI64X4:
      case XED_ICLASS_VINSERTPS:
      case XED_ICLASS_VLDDQU:
      case XED_ICLASS_VMASKMOVDQU:
      case XED_ICLASS_VMASKMOVPD:
      case XED_ICLASS_VMASKMOVPS:
      case XED_ICLASS_VMAXPD:
      case XED_ICLASS_VMAXPS:
      case XED_ICLASS_VMAXSD:
      case XED_ICLASS_VMAXSS:
      case XED_ICLASS_VMINPD:
      case XED_ICLASS_VMINPS:
      case XED_ICLASS_VMINSD:
      case XED_ICLASS_VMINSS:
      case XED_ICLASS_VMOVAPD:
      case XED_ICLASS_VMOVAPS:
      case XED_ICLASS_VMOVD:
      case XED_ICLASS_VMOVDDUP:
      case XED_ICLASS_VMOVDQA32:
      case XED_ICLASS_VMOVDQA64:
      case XED_ICLASS_VMOVDQA:
      case XED_ICLASS_VMOVDQU16:
      case XED_ICLASS_VMOVDQU32:
      case XED_ICLASS_VMOVDQU64:
      case XED_ICLASS_VMOVDQU8:
      case XED_ICLASS_VMOVDQU:
      case XED_ICLASS_VMOVHLPS:
      case XED_ICLASS_VMOVHPD:
      case XED_ICLASS_VMOVHPS:
      case XED_ICLASS_VMOVLHPS:
      case XED_ICLASS_VMOVLPD:
      case XED_ICLASS_VMOVLPS:
      case XED_ICLASS_VMOVMSKPD:
      case XED_ICLASS_VMOVMSKPS:
      case XED_ICLASS_VMOVNTDQ:
      case XED_ICLASS_VMOVNTDQA:
      case XED_ICLASS_VMOVNTPD:
      case XED_ICLASS_VMOVNTPS:
      case XED_ICLASS_VMOVQ:
      case XED_ICLASS_VMOVSD:
      case XED_ICLASS_VMOVSHDUP:
      case XED_ICLASS_VMOVSLDUP:
      case XED_ICLASS_VMOVSS:
      case XED_ICLASS_VMOVUPD:
      case XED_ICLASS_VMOVUPS:
      case XED_ICLASS_VMPSADBW:
      case XED_ICLASS_VMULPD:
      case XED_ICLASS_VMULPS:
      case XED_ICLASS_VMULSD:
      case XED_ICLASS_VMULSS:
      case XED_ICLASS_VORPD:
      case XED_ICLASS_VORPS:
      case XED_ICLASS_VPABSB:
      case XED_ICLASS_VPABSD:
      case XED_ICLASS_VPABSQ:
      case XED_ICLASS_VPABSW:
      case XED_ICLASS_VPACKSSDW:
      case XED_ICLASS_VPACKSSWB:
      case XED_ICLASS_VPACKUSDW:
      case XED_ICLASS_VPACKUSWB:
      case XED_ICLASS_VPADDB:
      case XED_ICLASS_VPADDD:
      case XED_ICLASS_VPADDQ:
      case XED_ICLASS_VPADDSB:
      case XED_ICLASS_VPADDSW:
      case XED_ICLASS_VPADDUSB:
      case XED_ICLASS_VPADDUSW:
      case XED_ICLASS_VPADDW:
      case XED_ICLASS_VPALIGNR:
      case XED_ICLASS_VPAND:
      case XED_ICLASS_VPANDN:
      case XED_ICLASS_VPAVGB:
      case XED_ICLASS_VPAVGW:
      case XED_ICLASS_VPBLENDD:
      case XED_ICLASS_VPBLENDVB:
      case XED_ICLASS_VPBLENDW:
      case XED_ICLASS_VPBROADCASTB:
      case XED_ICLASS_VPBROADCASTD:
      case XED_ICLASS_VPBROADCASTQ:
      case XED_ICLASS_VPBROADCASTW:
      case XED_ICLASS_VPCLMULQDQ:
      case XED_ICLASS_VPCMPB:
      case XED_ICLASS_VPCMPD:
      case XED_ICLASS_VPCMPEQB:
      case XED_ICLASS_VPCMPEQD:
      case XED_ICLASS_VPCMPEQQ:
      case XED_ICLASS_VPCMPEQW:
      case XED_ICLASS_VPCMPESTRI:
      case XED_ICLASS_VPCMPESTRM:
      case XED_ICLASS_VPCMPGTB:
      case XED_ICLASS_VPCMPGTD:
      case XED_ICLASS_VPCMPGTQ:
      case XED_ICLASS_VPCMPGTW:
      case XED_ICLASS_VPCMPISTRI:
      case XED_ICLASS_VPCMPISTRM:
      case XED_ICLASS_VPCMPQ:
      case XED_ICLASS_VPCMPUB:
      case XED_ICLASS_VPCMPUD:
      case XED_ICLASS_VPCMPUQ:
      case XED_ICLASS_VPCMPUW:
      case XED_ICLASS_VPCMPW:
      case XED_ICLASS_VPERM2F128:
      case XED_ICLASS_VPERM2I128:
      case XED_ICLASS_VPERMD:
      case XED_ICLASS_VPERMILPD:
      case XED_ICLASS_VPERMILPS:
      case XED_ICLASS_VPERMPD:
      case XED_ICLASS_VPERMPS:
      case XED_ICLASS_VPERMQ:
      case XED_ICLASS_VPEXTRB:
      case XED_ICLASS_VPEXTRD:
      case XED_ICLASS_VPEXTRQ:
      case XED_ICLASS_VPEXTRW:
      case XED_ICLASS_VPHADDBD:
      case XED_ICLASS_VPHADDBQ:
      case XED_ICLASS_VPHADDBW:
      case XED_ICLASS_VPHADDD:
      case XED_ICLASS_VPHADDDQ:
      case XED_ICLASS_VPHADDSW:
      case XED_ICLASS_VPHADDUBD:
      case XED_ICLASS_VPHADDUBQ:
      case XED_ICLASS_VPHADDUBW:
      case XED_ICLASS_VPHADDUDQ:
      case XED_ICLASS_VPHADDUWD:
      case XED_ICLASS_VPHADDUWQ:
      case XED_ICLASS_VPHADDW:
      case XED_ICLASS_VPHADDWD:
      case XED_ICLASS_VPHADDWQ:
      case XED_ICLASS_VPHMINPOSUW:
      case XED_ICLASS_VPHSUBBW:
      case XED_ICLASS_VPHSUBD:
      case XED_ICLASS_VPHSUBDQ:
      case XED_ICLASS_VPHSUBSW:
      case XED_ICLASS_VPHSUBW:
      case XED_ICLASS_VPHSUBWD:
      case XED_ICLASS_VPINSRB:
      case XED_ICLASS_VPINSRD:
      case XED_ICLASS_VPINSRQ:
      case XED_ICLASS_VPINSRW:
      case XED_ICLASS_VPMADD52HUQ:
      case XED_ICLASS_VPMADD52LUQ:
      case XED_ICLASS_VPMADDUBSW:
      case XED_ICLASS_VPMADDWD:
      case XED_ICLASS_VPMASKMOVD:
      case XED_ICLASS_VPMASKMOVQ:
      case XED_ICLASS_VPMAXSB:
      case XED_ICLASS_VPMAXSD:
      case XED_ICLASS_VPMAXSW:
      case XED_ICLASS_VPMAXUB:
      case XED_ICLASS_VPMAXUD:
      case XED_ICLASS_VPMAXUW:
      case XED_ICLASS_VPMINSB:
      case XED_ICLASS_VPMINSD:
      case XED_ICLASS_VPMINSW:
      case XED_ICLASS_VPMINUB:
      case XED_ICLASS_VPMINUD:
      case XED_ICLASS_VPMINUW:
      case XED_ICLASS_VPMOVSDB:
      case XED_ICLASS_VPMOVSDW:
      case XED_ICLASS_VPMOVSQB:
      case XED_ICLASS_VPMOVSQD:
      case XED_ICLASS_VPMOVSQW:
      case XED_ICLASS_VPMOVSWB:
      case XED_ICLASS_VPMOVSXBD:
      case XED_ICLASS_VPMOVSXBQ:
      case XED_ICLASS_VPMOVSXBW:
      case XED_ICLASS_VPMOVSXDQ:
      case XED_ICLASS_VPMOVSXWD:
      case XED_ICLASS_VPMOVSXWQ:
      case XED_ICLASS_VPMOVZXBD:
      case XED_ICLASS_VPMOVZXBQ:
      case XED_ICLASS_VPMOVZXBW:
      case XED_ICLASS_VPMOVZXDQ:
      case XED_ICLASS_VPMOVZXWD:
      case XED_ICLASS_VPMOVZXWQ:
      case XED_ICLASS_VPMULDQ:
      case XED_ICLASS_VPMULHRSW:
      case XED_ICLASS_VPMULHUW:
      case XED_ICLASS_VPMULHW:
      case XED_ICLASS_VPMULLD:
      case XED_ICLASS_VPMULLW:
      case XED_ICLASS_VPMULUDQ:
      case XED_ICLASS_VPOR:
      case XED_ICLASS_VPORD:
      case XED_ICLASS_VPORQ:
      case XED_ICLASS_VPSADBW:
      case XED_ICLASS_VPSHUFB:
      case XED_ICLASS_VPSHUFD:
      case XED_ICLASS_VPSHUFHW:
      case XED_ICLASS_VPSHUFLW:
      case XED_ICLASS_VPSIGNB:
      case XED_ICLASS_VPSIGND:
      case XED_ICLASS_VPSIGNW:
      case XED_ICLASS_VPSLLD:
      case XED_ICLASS_VPSLLDQ:
      case XED_ICLASS_VPSLLQ:
      case XED_ICLASS_VPSLLVD:
      case XED_ICLASS_VPSLLVQ:
      case XED_ICLASS_VPSLLVW:
      case XED_ICLASS_VPSLLW:
      case XED_ICLASS_VPSRAD:
      case XED_ICLASS_VPSRAVD:
      case XED_ICLASS_VPSRAW:
      case XED_ICLASS_VPSRLD:
      case XED_ICLASS_VPSRLQ:
      case XED_ICLASS_VPSRLVD:
      case XED_ICLASS_VPSRLVQ:
      case XED_ICLASS_VPSRLVW:
      case XED_ICLASS_VPSRLW:
      case XED_ICLASS_VPSUBB:
      case XED_ICLASS_VPSUBD:
      case XED_ICLASS_VPSUBQ:
      case XED_ICLASS_VPSUBSB:
      case XED_ICLASS_VPSUBSW:
      case XED_ICLASS_VPSUBUSB:
      case XED_ICLASS_VPSUBUSW:
      case XED_ICLASS_VPSUBW:
      case XED_ICLASS_VPTEST:
      case XED_ICLASS_VPUNPCKHBW:
      case XED_ICLASS_VPUNPCKHDQ:
      case XED_ICLASS_VPUNPCKHQDQ:
      case XED_ICLASS_VPUNPCKHWD:
      case XED_ICLASS_VPUNPCKLBW:
      case XED_ICLASS_VPUNPCKLDQ:
      case XED_ICLASS_VPUNPCKLQDQ:
      case XED_ICLASS_VPUNPCKLWD:
      case XED_ICLASS_VPXOR:
      case XED_ICLASS_VRCPPS:
      case XED_ICLASS_VRCPSS:
      case XED_ICLASS_VROUNDPD:
      case XED_ICLASS_VROUNDPS:
      case XED_ICLASS_VROUNDSD:
      case XED_ICLASS_VROUNDSS:
      case XED_ICLASS_VRSQRT14PD:
      case XED_ICLASS_VRSQRT14PS:
      case XED_ICLASS_VRSQRT14SD:
      case XED_ICLASS_VRSQRT14SS:
      case XED_ICLASS_VRSQRT28PD:
      case XED_ICLASS_VRSQRT28PS:
      case XED_ICLASS_VRSQRT28SD:
      case XED_ICLASS_VRSQRT28SS:
      case XED_ICLASS_VRSQRTPS:
      case XED_ICLASS_VRSQRTSS:
      case XED_ICLASS_VSHUFPD:
      case XED_ICLASS_VSHUFPS:
      case XED_ICLASS_VSQRTPD:
      case XED_ICLASS_VSQRTPS:
      case XED_ICLASS_VSQRTSD:
      case XED_ICLASS_VSQRTSS:
      case XED_ICLASS_VSUBPD:
      case XED_ICLASS_VSUBPS:
      case XED_ICLASS_VSUBSD:
      case XED_ICLASS_VSUBSS:
      case XED_ICLASS_VTESTPD:
      case XED_ICLASS_VTESTPS:
      case XED_ICLASS_VUCOMISD:
      case XED_ICLASS_VUCOMISS:
      case XED_ICLASS_VUNPCKHPD:
      case XED_ICLASS_VUNPCKHPS:
      case XED_ICLASS_VUNPCKLPD:
      case XED_ICLASS_VUNPCKLPS:
      case XED_ICLASS_VXORPD:
      case XED_ICLASS_VXORPS:
      case XED_ICLASS_XORPD:
      case XED_ICLASS_XORPS:

        /// remove trailing type suffix ('l', 'q', 'x', 'y', 'z', AT&T syntax does NOT use it.
        mnem = make_intel_mnemonic(inst);
        break;

      default:
        break;
    }

    return mnem;
  }

  /// Compute AT&T asm syntax instruction mnemonic from Pipedream_Instruction
  /// All fields except 'mnemonic' must already be filled in
  static std::string make_intel_mnemonic(XED_Decoded_Instruction inst) {
    std::string mnem = xed_iform_to_iclass_string_intel(inst.inst().iform());

    for (char &c : mnem) {
      c = std::tolower(c);
    }

    return mnem;
  }

  static Pipedream::Flag_Set flag_set_from_xed(const xed_flag_set_t *p) {
    using Flag     = Pipedream::Flag;
    using Flag_Set = Pipedream::Flag_Set;

    Flag_Set flags = 0;

    if (p->s.of) { flags |= Flag_Set(Flag::OF); }
    if (p->s.sf) { flags |= Flag_Set(Flag::SF); }
    if (p->s.zf) { flags |= Flag_Set(Flag::ZF); }
    if (p->s.af) { flags |= Flag_Set(Flag::AF); }
    if (p->s.pf) { flags |= Flag_Set(Flag::PF); }
    if (p->s.cf) { flags |= Flag_Set(Flag::CF); }
    if (p->s.df) { flags |= Flag_Set(Flag::DF); }
    if (p->s.vif) { flags |= Flag_Set(Flag::VIF); }
    if (p->s.iopl) { flags |= Flag_Set(Flag::IOPL); }
    if (p->s._if) { flags |= Flag_Set(Flag::IF); }
    if (p->s.ac) { flags |= Flag_Set(Flag::AC); }
    if (p->s.vm) { flags |= Flag_Set(Flag::VM); }
    if (p->s.rf) { flags |= Flag_Set(Flag::RF); }
    if (p->s.nt) { flags |= Flag_Set(Flag::NT); }
    if (p->s.tf) { flags |= Flag_Set(Flag::TF); }
    if (p->s.id) { flags |= Flag_Set(Flag::ID); }
    if (p->s.vip) { flags |= Flag_Set(Flag::VIP); }
    if (p->s.fc0) { flags |= Flag_Set(Flag::FC0); }
    if (p->s.fc1) { flags |= Flag_Set(Flag::FC1); }
    if (p->s.fc2) { flags |= Flag_Set(Flag::FC2); }
    if (p->s.fc3) { flags |= Flag_Set(Flag::FC3); }

    return flags;
  }

  static Pipedream::Flag_Set read_flags_from_xed(XED_Decoded_Instruction inst) {
    Pipedream::Flag_Set out = 0;

    if (xed_decoded_inst_uses_rflags(inst.raw())) {
      const xed_simple_flag_t *const rfi = xed_decoded_inst_get_rflags_info(inst.raw());

      const xed_flag_set_t *const read_set = xed_simple_flag_get_read_flag_set(rfi);

      out |= flag_set_from_xed(read_set);
    }

    return out;
  }

  static Pipedream::Flag_Set write_flags_from_xed(XED_Decoded_Instruction inst) {
    Pipedream::Flag_Set out = 0;

    if (xed_decoded_inst_uses_rflags(inst.raw())) {
      const xed_simple_flag_t *const rfi = xed_decoded_inst_get_rflags_info(inst.raw());

      /// write flags includes "undefined" flags
      /// TODO: seperate out undefined flags? what are they?
      const xed_flag_set_t *const write_set = xed_simple_flag_get_written_flag_set(rfi);

      out |= flag_set_from_xed(write_set);
    }

    return out;
  }

};

/// try to find XED instructions we can represent
struct Encode_Decode {
  /// A XED instruction we are able to encode.
  struct Instruction {
    /// the minimum amount of information (should uniquely identify the instruction)
    xed_iclass_enum_t iclass;
    Pipedream_Operand_Vector operands;
    xed_isa_set_enum_t isa;

    /// additional information (try to add as much as possible while still keeping the instruction uniquely identifiable by the keys above)
    xed_extension_enum_t extension;
    xed_category_enum_t category;
    std::set<std::string> tags;

    std::string att_mnemonic;
    std::string intel_mnemonic;

    bool operator<(const Instruction &that) const { return compare(that, std::less<void>{}); }
    bool operator==(const Instruction &that) const { return compare(that, std::equal_to<void>{}); }

    friend std::ostream &operator<<(std::ostream &O, const Instruction &x) {
      O << "  " << xed_iclass_enum_t2str(x.iclass) << "\n";

      for (const auto &op : x.operands) {
        O << "    " << op << "\n";
      }

      O << "    isa " << xed_isa_set_enum_t2str(x.isa) << "\n";
      O << "    ext " << xed_extension_enum_t2str(x.extension) << "\n";
      O << "    cat " << xed_category_enum_t2str(x.category) << "\n";
      O << "    tag";
      for (const std::string &tag : x.tags) {
        O << " " << tag;
      }
      O << "\n";
      O << "    at&t  " << x.att_mnemonic << "\n";
      O << "    intel " << x.intel_mnemonic << "\n";

      O << "\n";
      return O;
    }
  private:
    template<typename Op>
    bool compare(const Instruction &that, Op &&cmp) const {
      const auto tie = [](const Instruction &ei) {
        return std::tie(ei.iclass, ei.operands, ei.isa, ei.extension, ei.category, ei.tags, ei.att_mnemonic, ei.intel_mnemonic);
      };

      return cmp(tie(*this), tie(that));
    }
  };

  struct Instruction_Set {
    using iterator = typename std::set<Instruction>::const_iterator;

    void insert(const Instruction &inst) {
      set.insert(inst);
    }

    void merge(Instruction_Set &&that) {
      set.merge(that.set);
    }

    void operator|=(Instruction_Set &&that) {
      merge(std::move(that));
    }

    void clear() { set.clear(); }

    size_t size() const { return set.size(); }

    iterator begin() const { return set.begin(); }
    iterator end() const { return set.end(); }
  private:
    std::set<Instruction> set;
  };

  static Instruction_Set find_all_encodings(XED_CPU_Mode cpu_mode, Span<XED_Instruction> instructions) {
    Instruction_Set all;

#ifdef _OPENMP
    #pragma omp parallel
#endif
    {
      Instruction_Set chunk;

      const size_t num_instructions = instructions.size();

#ifdef _OPENMP
      #pragma omp for schedule(dynamic, 64)
#endif
      for (size_t i = 0; i < num_instructions; i++) {
        find_all_encodings(cpu_mode, instructions[i], chunk);
      }

#ifdef _OPENMP
      #pragma omp critical
#endif
      {
        all.merge(std::move(chunk));
      }
    }

    return all;
  }

  static Instruction_Set find_all_encodings(XED_CPU_Mode cpu_mode, XED_Instruction inst) {
    Instruction_Set out;
    find_all_encodings(cpu_mode, inst, out);
    return out;
  }
private:
  static void find_all_encodings(XED_CPU_Mode cpu_mode, XED_Instruction inst, Instruction_Set &out) {
    std::cerr << "# " << inst.iform_str() << "\n";

    for (Effective_Operand_Width effective_operand_width : {EOSZ_16, EOSZ_32, EOSZ_64}) {
      for (bool allow_rex : {true, false}) {
        if (inst.has_memory_operands()) {
          for (Effective_Address_Width effective_address_width : {EASZ_16, EASZ_32, EASZ_64}) {
            XED_Encoder_State state = {cpu_mode, effective_operand_width, effective_address_width, allow_rex};

            for (unsigned memory_width : {8, 16, 32, 64, 128, 256, 320, 512}) {
              out |= find_all_encodings(inst, state, memory_width, false);
              out |= find_all_encodings(inst, state, memory_width, true);
            }
          }
        } else {
          XED_Encoder_State state = {cpu_mode, effective_operand_width, EASZ_64, allow_rex};

          out |= find_all_encodings(inst, state, 0, false);
          out |= find_all_encodings(inst, state, 0, true);
        }
      }
    }
  }

  static Instruction_Set find_all_encodings(XED_Instruction inst,
                                            XED_Encoder_State &state,
                                            unsigned memory_width,
                                            bool fill_in_trailing_implicit_operands) {
    if (xed_extractor_cannot_handle(inst, state)) { return {}; }

    struct Encoder_Operand_Variable {
      /// Are we done exploring all options of this variable?.
      bool done() const {
        return curr_reg_class.empty();
      }

      /// advance to the next option.
      void step() {
        *operand = xed_reg(curr_reg_class.front());

        curr_reg_class = curr_reg_class.drop_front();
      }

      void reset() {
        curr_reg_class = reg_class;
      }

      void show(std::ostream &O) const {
        if (!operand) {
          O << "nil";
          return;
        }

        switch (operand->type) {
          case XED_ENCODER_OPERAND_TYPE_IMM0:
            O << "imm" << operand->width_bits << ":" << operand->u.imm0;
            break;
          case XED_ENCODER_OPERAND_TYPE_REG:
            O << "reg: " << Pipedream::register_name(operand->u.reg);
            break;
          case XED_ENCODER_OPERAND_TYPE_MEM:
            O << "mem" << operand->width_bits << ":" << Pipedream::register_name(operand->u.mem.base);
            break;
          default:
            abort();
        }
      }

      xed_encoder_operand_t *operand = nullptr;
      Register_Class curr_reg_class = {};
      Register_Class reg_class = {};
    };

    size_t num_encoder_operands = inst.num_operands();

    /// We don't have to fill in trailing SUPPRESSED operands, encoder guesses them for us.
    /// For SOME instructions we HAVE TO fill in trailing IMPLICIT operands, for MOST we MUST NOT :/
    /// FIXME: In fact if we fill in SUPPRESSED operands we get some weird segfaults in xed_encode (jump to PC 0x0 ?!?)
    while (num_encoder_operands) {
      const xed_operand_visibility_enum_t opvis = inst.operand(num_encoder_operands - 1).visibility();

      if (fill_in_trailing_implicit_operands && (opvis == XED_OPVIS_IMPLICIT)) {
        break;
      }
      if (fill_in_trailing_implicit_operands && opvis == XED_OPVIS_EXPLICIT) {
        break;
      }

      num_encoder_operands--;
    }

    xed_encoder_operand_t encoder_operands[num_encoder_operands];

    std::vector<Encoder_Operand_Variable> vars;

    for (size_t i = 0, end = num_encoder_operands; i < end; i++) {
      XED_Instruction_Operand op = inst.operand(i);

      const std::optional<Machine_Value> value = Convert::machine_value_from_xed(inst, op, state, memory_width);

      if (!value) { return {}; }

      switch (value->kind()) {
        case Machine_Value::REG:
          if (auto reg = value->reg()) {
            encoder_operands[i] = xed_reg(*reg);
          } else {
            Encoder_Operand_Variable var = {encoder_operands + i, value->reg_class(), value->reg_class()};
            vars.push_back(var);
            var.step();
          }
          break;
        case Machine_Value::MEM:
        case Machine_Value::ADDR:
          {
            xed_reg_enum_t base;
            if (auto reg = value->base_reg()) {
              base = *reg;
            } else {
              base = XED_Register_Info::base_reg(state);
            }
            encoder_operands[i] = xed_mem_b(base, memory_width);
          }
          break;
        case Machine_Value::IMM:
          encoder_operands[i] = xed_imm0(0, value->imm_bits());
          break;
        case Machine_Value::FLAGS:
          /// FLAGS operands are always supressed
          abort();
          break;
      }
    }

    auto all_done = [&]() {
      for (const Encoder_Operand_Variable &var : vars) {
        if (!var.done()) { return false; }
      }
      return true;
    };

    auto step_iterators = [&]() -> bool {
      for (Encoder_Operand_Variable &var : vars) {
        if (var.done()) {
          var.reset();
          continue;
        } else {
          var.step();
          return true;
        }
      }

      return false;
    };

    Instruction_Set out;

    for (; !all_done(); step_iterators()) {
      const std::optional<XED_Decoded_Instruction> decoded_inst = encode(state, inst, num_encoder_operands, encoder_operands);

      if (decoded_inst) {
        const XED_Instruction new_inst = decoded_inst->inst();

        if (inst.iclass() != new_inst.iclass()) {
          std::cerr << __func__ << ": ICLASS has changed:\n";
          std::cerr << "  WANT:\n";
          Debug::print_inst(std::cerr, inst);
          std::cerr << "  HAVE:\n";
          Debug::print_inst(std::cerr, *decoded_inst);
          abort();
        }

//        if (decoded_inst->effective_operand_width() != effective_operand_width) {
//          std::cerr << __func__ << ": effective_operand_width has changed:\n";
//          std::cerr << "  WANT: EOSZ=" << effective_operand_width << "\n";
//          Debug::print_inst(std::cerr, inst);
//          std::cerr << "  HAVE: EOSZ=" << decoded_inst->effective_operand_width() << "\n";
//          Debug::print_inst(std::cerr, *decoded_inst);
//          abort();
//        }
//
//        if (decoded_inst->effective_address_width() != effective_address_width) {
//          std::cerr << __func__ << ": effective_address_width has changed:\n";
//          std::cerr << "  WANT: EASZ=" << effective_address_width << "\n";
//          Debug::print_inst(std::cerr, inst);
//          std::cerr << "  HAVE: EASZ=" << decoded_inst->effective_address_width() << "\n";
//          Debug::print_inst(std::cerr, *decoded_inst);
//          abort();
//        }

        /// the IFORM may have changed.

        if (inst.num_operands() != new_inst.num_operands()) {
          std::cerr << __func__ << ": number of operands has changed:\n";
          std::cerr << "  WANT:\n";
          Debug::print_inst(std::cerr, inst);
          std::cerr << "  HAVE:\n";
          Debug::print_inst(std::cerr, *decoded_inst);
          abort();
        }

        Instruction enc;
        enc.iclass         = inst.iclass();
        enc.operands       = Convert::pipedream_operands_from_xed(*decoded_inst, state.cpu_mode, state.allow_rex);
        enc.isa            = inst.isa_set();
        enc.extension      = inst.extension();
        enc.category       = inst.category();
        enc.tags           = Pipedream::tags(*decoded_inst, enc.operands);
        enc.att_mnemonic   = Convert::make_att_mnemonic(*decoded_inst, enc.operands);
        enc.intel_mnemonic = Convert::make_intel_mnemonic(*decoded_inst);

        out.insert(enc);
      } else {
//        auto &O = std::cerr;
//
//        O << __func__ << ": could not encode/decode:\n";
//        Debug::print_inst(O, inst);
//        Debug::print_encoder_state(O, state);
//        O << "  ENCODER PARAMS:\n";
//        for (size_t i = 0; i < num_encoder_operands; i++) {
//          O << "    ";
//          Debug::print_operand(O, encoder_operands[i]);
//          O << "\n";
//        }
//        abort();

        /// bail out
        out.clear();
        break;
      }
    }

    return out;
  }
public:
  static std::optional<XED_Decoded_Instruction> encode(const XED_Encoder_State &state, XED_Instruction inst, size_t noperands, const xed_encoder_operand_t *operands) {
    return encode(state, inst.iclass(), inst.iform(), noperands, operands);
  }

  static std::optional<XED_Decoded_Instruction> encode(const XED_Encoder_State &state, const xed_iclass_enum_t iclass, const xed_iform_enum_t iform, size_t noperands, const xed_encoder_operand_t *operands) {
    /// NOTE: The XED encoder is pretty lenient. The decoder is not.
    ///       More precisely, instead of failing to encode the encoder uses illegal prefix bytes if you ask it to use bad operand/address widths.
    const bool strict = (state.effective_operand_width != 16);

    xed_encoder_instruction_t encoder_instruction;
    xed_inst(&encoder_instruction, state.cpu_mode.raw(), iclass, state.effective_operand_width, noperands, operands);

    xed_encoder_request_t encoder_request;
    xed_encoder_request_zero(&encoder_request);

    const xed_bool_t encode_ok = xed_convert_to_encoder_request(&encoder_request, &encoder_instruction);
    if (!encode_ok) {
      if (strict) {
        std::cerr << "CONVERT TO ENCODER REQUEST FAILED\n";
        exit(1);
      }
      return std::nullopt;
    }

    XED_Decoded_Instruction decoded_inst;
    xed_uint8_t itext[XED_MAX_INSTRUCTION_BYTES];
    unsigned int ilen = XED_MAX_INSTRUCTION_BYTES;
    unsigned int olen = 0;

    {
      const xed_error_enum_t encode_error = xed_encode(&encoder_request, itext, ilen, &olen);
      if (encode_error != XED_ERROR_NONE) {
        /// only ever returns XED_ERROR_GENERAL_ERROR or XED_ERROR_NONE
        return std::nullopt;
      }

      xed_decoded_inst_zero_set_mode(decoded_inst.raw(), &encoder_instruction.mode);
      const xed_error_enum_t decode_error = xed_decode(decoded_inst.raw(), itext, olen);
      switch (decode_error) {
        case XED_ERROR_NONE:
          /// no error. great
          break;
        case XED_ERROR_BAD_LEGACY_PREFIX:
          /// the encoder allowed a legacy prefix on something where it is not actually legal
          return std::nullopt;
        default:
          if (strict) {
            auto &O = std::cerr;

            O << "cannot decode what we just encoded? " << xed_error_enum_t2str(decode_error) << "\n";
            O << " iclass:                  " << xed_iclass_enum_t2str(iclass) << " (" << iclass << ")\n";
            O << " iform:                   " << xed_iform_enum_t2str(iform) << " (" << iform << ")\n";
            O << " src machine-mode:        " << state.cpu_mode.mode_str() << "(" << state.cpu_mode.mode() << ")\n";
            O << " src stack-address-width: " << state.cpu_mode.stack_addr_width_str() << " (" << state.cpu_mode.stack_addr_width() << ")\n";

            XED_CPU_Mode enc_mode{encoder_instruction.mode};

            O << " dst machine-mode:        " << enc_mode.mode_str() << " (" << enc_mode.mode_str() << ")\n";
            O << " dst stack-address-width: " << enc_mode.stack_addr_width_str() << " (" << enc_mode.stack_addr_width() << ")\n";
            O << " effective-operand-width: " << state.effective_operand_width << "\n";
            O << " bytes:                  ";
            for (unsigned i = 0; i < olen; i++) {
              O << " " << std::setw(2) << std::setfill('0') << std::hex << int(itext[i]) << std::dec;
            }
            O << "\n";
            abort();
          }
          return std::nullopt;
      }
    }

    if (decoded_inst.inst().iclass() != iclass) {
      if ((iclass == XED_ICLASS_XCHG) && (decoded_inst.inst().iform() == XED_IFORM_NOP_90)) {
        /// FIXME: "xchg %ax, %ax" (and "xchg %eax, %eax" ...) become XED_IFORM_NOP_90 in XED.
        /// ... actually NOP_90 is not a real instruction :/
      } else if (strict) {
        auto &O = std::cerr;
        O << "inst.iclass != decode(encode(inst)).iclass\n";
        O << "inst:\n";
        O << "  iclass:                  " << xed_iclass_enum_t2str(iclass) << " (" << iclass << ")\n";
        O << "  iform:                   " << xed_iform_enum_t2str(iform) << " (" << iform << ")\n";
        Debug::print_encoder_state(O, state);
        O << "  #operands:               " << noperands << "\n";
        O << "  operands:                ";
        for (unsigned i = 0; i < noperands; i++) {
          O << " ";
          Debug::print_operand(O, operands[i]);
        }
        O << "\n";
        O << "  bytes:                  ";
        for (unsigned i = 0; i < olen; i++) {
          O << " " << std::setw(2) << std::setfill('0') << std::hex << int(itext[i]) << std::dec;
        }
        O << "\n";
        O << "decode(encode(inst)):\n";
        Debug::print_inst(O, decoded_inst);
        abort();
      }
      return std::nullopt;
    }

    return decoded_inst;
  }

  static bool xed_extractor_cannot_handle(XED_Instruction inst, XED_Encoder_State &state) {
//    if (inst.iclass_str()[0] == 'V') { return true; }
//    if (inst.iclass_str()[0] != 'V') { return true; }

    if (state.cpu_mode.mode() != XED_MACHINE_MODE_LONG_64) { return true; }

    if (inst.has_rep_prefix()) {
      /// skip instructions with rep prefix
      /// TODO: add repe/repne support
      return true;
    }
    if (inst.has_lock_prefix()) {
      /// skip instructions with lock prefix
      /// TODO: add lock support
      return true;
    }

    /// drop well known weird instructions
    switch (inst.iclass()) {
      /// Instruction with two memory operands
      /// where the second memory operand is not encoded like a normal memory operand.
      /// This leads to weirdness in XED.
      case XED_ICLASS_MOVDIR64B:

      /// gets decoded as NOP since we don't set some exotic CPU flags
      case XED_ICLASS_ENDBR32:
      case XED_ICLASS_ENDBR64:
      case XED_ICLASS_RDSSPD:
      case XED_ICLASS_RDSSPQ:
      case XED_ICLASS_CLDEMOTE:
        return true;

      /// AMD XOP instructions: have a 4-bit immediate operand, but XED says the immediate has 8 bits.
      /// TODO: patch operand in Pipedream_Operand::from_xed
      case XED_ICLASS_VPERMIL2PS:
      case XED_ICLASS_VPERMIL2PD:
        return true;

      /// multi-byte NOP instructions: in the encoding they have a memory BI operand, but XED says they have two register operands.
      /// TODO: patch operand in Pipedream_Operand::from_xed
      case XED_ICLASS_NOP:
      case XED_ICLASS_NOP2:
      case XED_ICLASS_NOP3:
      case XED_ICLASS_NOP4:
      case XED_ICLASS_NOP5:
      case XED_ICLASS_NOP6:
      case XED_ICLASS_NOP7:
      case XED_ICLASS_NOP8:
      case XED_ICLASS_NOP9:
        return true;

      /// gets decoded as PREFETCHWT1 since we don't set some exotic CPU flags
      case XED_ICLASS_PREFETCH_RESERVED:
        return true;

      default:
        break;
    }

    return false;
  }
};

/// find unique names for instructions
struct Instruction_Namer {
  using Instruction = Encode_Decode::Instruction;

  struct Cannot_Disambiguate {
    Encode_Decode::Instruction a, b;
  };

  using Disambiguated = std::map<std::string, Instruction>;

  /// find unique names for all instructions in @p instructions
  static std::variant<Disambiguated, Cannot_Disambiguate> disambiguate(const Encode_Decode::Instruction_Set &instructions) {
    std::map<std::string, Worklist_Item> disambig;

    std::vector<Worklist_Item> worklist;
    worklist.reserve(instructions.size());
    for (const Instruction &p : instructions) {
      worklist.emplace_back(p);
    }

    assert((worklist.size() + disambig.size()) == instructions.size());

    while (!worklist.empty()) {
      assert((worklist.size() + disambig.size()) == instructions.size());

      Worklist_Item &new_item = worklist.back();

      auto it = disambig.find(new_item.prefix);

      if (it == disambig.end()) {
        /// new element. done
        disambig.emplace_hint(it, new_item.prefix, new_item);
        worklist.pop_back();
        continue;
      }

      Worklist_Item &old_item = it->second;

      if (new_item.payload == old_item.payload) {
        assert(new_item.prefix    == old_item.prefix);
        assert(new_item.remaining == old_item.remaining);

        worklist.pop_back();
        continue;
      }

      if (new_item.remaining.empty() || old_item.remaining.empty()) {
        return Cannot_Disambiguate{new_item.payload, old_item.payload};
      }

      /// old_item is no longer unambiguously identified by its prefix
      /// remove move map and put back on worklist

      new_item.pop_key();
      old_item.pop_key();

      worklist.emplace_back(std::move(old_item));
      disambig.erase(it);
    }

    assert((worklist.size() + disambig.size()) == instructions.size());

    Disambiguated out;

    for (const auto &pair : disambig) {
      out[pair.first] = pair.second.payload;
    }

    return out;
  }
private:
  struct Worklist_Item {
    Worklist_Item(const Worklist_Item &) = default;
    Worklist_Item(Worklist_Item &&) = default;

    Worklist_Item(const Instruction &instruction) : payload{instruction} {
      Span<Pipedream_Operand_Template> explicit_ops, implicit_ops;

      /// every instruction is always at least identified by
      ///  a) its name
      ///  b) its explicit operands

      /// we then try to find the shortest possible sequence of the following
      /// that uniquely identifies it among all instructions:
      ///   a) its implicit operands
      ///   b) its ISA set.

      const Pipedream_Operand_Vector &operands = instruction.operands;

      if (!operands.empty()) {
        size_t num_explicit = operands.size();

        while (num_explicit && operands[num_explicit - 1].visibility != XED_OPVIS_EXPLICIT) {
          num_explicit--;
        }

        explicit_ops = Span{num_explicit, operands.data()};
        implicit_ops = Span{operands.size() - num_explicit, operands.data() + num_explicit};
      }

      prefix = xed_iclass_enum_t2str(instruction.iclass);
      for (const auto &op : explicit_ops) {
        prefix += "_";
        prefix += op.short_name();
      }

      for (const auto &op : implicit_ops) {
        remaining.push_back("_" + op.short_name());
      }

      /// needed to disambiguate some vector instructions (VCVTUSI2SD)
//        if (inst.has_attribute(XED_ATTRIBUTE_MXCSR) || inst.has_attribute(XED_ATTRIBUTE_MXCSR_RD)) {
//          p.remaining.push_back("_MXCSR");
//        }

      remaining.push_back(std::string{"_"} + xed_isa_set_enum_t2str(instruction.isa));
    }

    Instruction payload;
    std::string prefix;
    std::vector<std::string> remaining;

    bool operator<(const Worklist_Item &p) const {
      return payload < p.payload;
    }

    void pop_key() {
      assert(!remaining.empty());

      const std::string &disambiguator = remaining.front();
      prefix += disambiguator;

      remaining.erase(remaining.begin());
    }
  };
};

template<class... Ts> struct overload : Ts... { using Ts::operator()...; };
template<class... Ts> overload(Ts...) -> overload<Ts...>;

struct Main {
  static int main() {
    xed_tables_init();

    std::vector<XED_Instruction> xed_insts;
    for (int i = 0; i < XED_MAX_INST_TABLE_NODES; i++) {
      const XED_Instruction inst{xed_inst_table_base() + i};

      xed_insts.push_back(inst);
    }

    const XED_CPU_Mode cpu_mode = XED_CPU_Mode::make(XED_MACHINE_MODE_LONG_64, XED_ADDRESS_WIDTH_64b);

    Encode_Decode::Instruction_Set instructions = Encode_Decode::find_all_encodings(cpu_mode, xed_insts);

    auto names = Instruction_Namer::disambiguate(instructions);

    std::visit(overload{
      [](const Instruction_Namer::Cannot_Disambiguate &err) {
        std::cerr << "ERROR: CANNOT DISAMBIGUATE\n";
        std::cerr << err.a;
        std::cerr << err.b;
        std::exit(1);
      },
      [](const Instruction_Namer::Disambiguated &names) {
        auto &O = std::cout;

        emit_pipedream_instruction_file_header(O);

        for (const auto &[name, inst] : names) {
          emit_pipedream_instruction_descriptor(O, name, inst);
        }

        emit_pipedream_instruction_file_footer(O);
      }
    }, names);


    return 0;
  }

  static void playground() {
    xed_tables_init();

    const auto try_encode = [](xed_iclass_enum_t iclass, const Effective_Operand_Width EOSZ, bool allow_rex, auto ...ops) {
      const XED_CPU_Mode cpu_mode = XED_CPU_Mode::make(XED_MACHINE_MODE_LONG_64, XED_ADDRESS_WIDTH_64b);

      const size_t noperands = sizeof...(ops);
      const xed_encoder_operand_t operands[noperands] = { ops... };

      const XED_Encoder_State state = {cpu_mode, EOSZ, EASZ_64, allow_rex};

      auto inst = Encode_Decode::encode(state, iclass, XED_IFORM_INVALID, noperands, operands);

      auto &O = std::cerr;

      if (inst) {
        O << "playground" << ": encoded: " << xed_iclass_enum_t2str(iclass) << "\n";
        Debug::print_inst(O, *inst);
      } else {
        O << "playground" << ": could not encode: " << xed_iclass_enum_t2str(iclass) << "\n";
//        Debug::print_inst(O, inst);
        Debug::print_encoder_state(O, state);
        O << "  ENCODER OPERANDS:\n";
        for (size_t i = 0; i < noperands; i++) {
          O << "    ";
          Debug::print_operand(O, operands[i]);
          O << "\n";
        }
        O << "\n";
        abort();
      }
    };

    try_encode(XED_ICLASS_SHLD, EOSZ_64, true, xed_reg(XED_REG_RAX), xed_reg(XED_REG_RBX), xed_reg(XED_REG_CL));

    try_encode(XED_ICLASS_INC, EOSZ_16, true, xed_reg(XED_REG_AL));
    try_encode(XED_ICLASS_INC, EOSZ_16, true, xed_reg(XED_REG_AX));
    try_encode(XED_ICLASS_INC, EOSZ_64, true, xed_reg(XED_REG_RAX));

    try_encode(XED_ICLASS_XOR, EOSZ_16, true, xed_reg(XED_REG_AL), xed_imm0(1, 8));

    try_encode(XED_ICLASS_XOR, EOSZ_16, true, xed_reg(XED_REG_AL), xed_reg(XED_REG_AL));
    try_encode(XED_ICLASS_XOR, EOSZ_16, true, xed_reg(XED_REG_AL), xed_reg(XED_REG_R15B));
  }

  static void emit_pipedream_instruction_file_header(std::ostream &O) {
    O << R"(
## This file is derived from the instruction data files of Intel XED
## The original files are available at https://intelxed.github.io
##
##  Original copyright (c) 2018 Intel Corporation
##  Modified copyright (c) 2019 Fabian Gruber
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##   http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from pipedream.asm.x86 import registers, flags
from pipedream.asm.ir import Use_Def, Operand_Visibility

import enum
from typing import *

## CPU mode
LONG_64 = 'LONG_64'

## read/write
R   = Use_Def.USE
W   = Use_Def.DEF
RW  = Use_Def.USE_DEF
CW  = Use_Def.DEF
RCW = Use_Def.USE_DEF

## operand visibility
EXPLICIT   = Operand_Visibility.EXPLICIT
IMPLICIT   = Operand_Visibility.IMPLICIT
SUPPRESSED = Operand_Visibility.SUPPRESSED

## operand type
i1   = 'i1'
i8   = 'i8'
i16  = 'i16'
i32  = 'i32'
i64  = 'i64'

u8   = 'u8'
u16  = 'u16'
u32  = 'u32'
u64  = 'u64'
u128 = 'u128'
u256 = 'u256'

f16  = 'f16'
f32  = 'f32'
f64  = 'f64'
f80  = 'f80'

OTHER = 'OTHER'

)";

  O << "## ISA\n";
  O << "class ISA(enum.Enum):\n";
  for (xed_isa_set_enum_t isa = xed_isa_set_enum_t(XED_ISA_SET_INVALID + 1); isa < XED_ISA_SET_LAST; isa = xed_isa_set_enum_t(int(isa) + 1)) {
    std::string txt  = xed_isa_set_enum_t2str(isa);
    std::string name = txt;

    if (std::isdigit(txt.front())) { name = '_' + name; }

    O << "  " << std::setw(20) << std::left << name << " = '" << txt << "'\n";
  }
  O << "\n";
  O << "## ISA extension\n";
  O << "class ISA_Extension(enum.Enum):\n";
  for (xed_extension_enum_t ext = xed_extension_enum_t(XED_EXTENSION_INVALID + 1); ext < XED_EXTENSION_LAST; ext = xed_extension_enum_t(int(ext) + 1)) {
    std::string txt  = xed_extension_enum_t2str(ext);
    std::string name = txt;

    if (std::isdigit(txt.front())) { name = '_' + name; }

    O << "  " << std::setw(20) << std::left << name << " = '" << txt << "'\n";
  }
  O << "\n";

  O << R"(
EXT = ISA_Extension

## register classes
class RC:
  GPR8      = registers.GPR8
  GPR8NOREX = registers.GPR8NOREX
  GPR16     = registers.GPR16
  GPR32     = registers.GPR32
  GPR64     = registers.GPR64
  VR64      = registers.VR64
  VR128     = registers.VR128
  VRX128    = registers.VRX128
  VR256     = registers.VR256
  VRX256    = registers.VRX256
  VRX512    = registers.VRX512
  FPST      = registers.FPST
)";
  for (const xed_reg_enum_t *regs : pipedream.singleton_register_classes()) {
    Register_Class reg_class{1, regs};

    switch (*regs) {
      case XED_REG_STACKPUSH:
      case XED_REG_STACKPOP:
      case XED_REG_X87POP:
      case XED_REG_X87PUSH:
        break;
      default:
        O << "  " << pipedream.register_class_name(reg_class) << " = registers.RC_" << pipedream.register_class_name(reg_class) << "\n";
        break;
    }
  }
  O << "\n";
  O << "## registers\n";
  for (const xed_reg_enum_t *regs : pipedream.singleton_register_classes()) {
    Register_Class reg_class{1, regs};

    switch (*regs) {
      case XED_REG_STACKPUSH:
      case XED_REG_STACKPOP:
      case XED_REG_X87POP:
      case XED_REG_X87PUSH:
        break;
      default:
        O << pipedream.register_class_name(reg_class) << " = registers." << pipedream.register_class_name(reg_class)
          << "\n";
        break;
    }
  }

  O << "\n";
  O << "## CPU flags\n";
  O << "F = flags.X86_Flags\n";
  O << "\n";

  O << R"(

def make_instruction_database(*, make_instruction, reg_op, mem_op, addr_op, imm_op, flags_op):
)";
  }

  static void emit_pipedream_instruction_file_footer(std::ostream &O) {
    O << R"(

if __name__ == '__main__':
  def make_instruction(**kwargs):
    print(kwargs)

  make_op = lambda **kwargs: kwargs

  make_instruction_database(
    make_instruction = make_instruction,
    reg_op           = make_op,
    mem_op           = make_op,
    addr_op          = make_op,
    imm_op           = make_op,
    flags_op         = make_op,
  )
)";
  }

  static void emit_pipedream_instruction_operands(std::ostream &O, const Pipedream_Operand_Vector &operands) {
    unsigned num_uses = 0, num_defs = 0, num_use_defs = 0;

    for (const auto &op : operands) {
      switch (op.reg().value_or(XED_REG_INVALID)) {
        case XED_REG_FLAGS:
        case XED_REG_EFLAGS:
        case XED_REG_RFLAGS:
        case XED_REG_IP:
        case XED_REG_EIP:
        case XED_REG_RIP:
          break;
        default:
          switch (op.action) {
          case XED_OPERAND_ACTION_R:
          case XED_OPERAND_ACTION_CR:
            num_uses++;
            break;
          case XED_OPERAND_ACTION_W:
          case XED_OPERAND_ACTION_CW:
            num_defs++;
            break;
          case XED_OPERAND_ACTION_RW:
          case XED_OPERAND_ACTION_RCW:
          case XED_OPERAND_ACTION_CRW:
            num_use_defs++;
            break;
          case XED_OPERAND_ACTION_INVALID:
          case XED_OPERAND_ACTION_LAST:
            abort();
        }
      }
    }

    unsigned curr_use = 0, curr_def = 0, curr_use_def = 0;

    for (const Pipedream_Operand_Template &op : operands) {
//      const Pipedream::Type type = pipedream.type(op.element_xtype, op.element_type, op.element_bits);

//      assert((type == Pipedream::OTHER) || op.element_bits == pipedream.type_bits(type));

      /// operand name

      std::string name;
      switch (op.reg().value_or(XED_REG_INVALID)) {
        case XED_REG_FLAGS:
        case XED_REG_EFLAGS:
        case XED_REG_RFLAGS:
          name = "flags";
          break;
        case XED_REG_IP:
        case XED_REG_EIP:
        case XED_REG_RIP:
          name = "ip";
          break;
        case XED_REG_SP:
        case XED_REG_ESP:
        case XED_REG_RSP:
          name = "sp";
          break;
        default:
          switch (op.action) {
            case XED_OPERAND_ACTION_R:
            case XED_OPERAND_ACTION_CR:
              curr_use++;
              name = "src";
              if (num_uses > 1) { name += std::to_string(curr_use); }
              break;
            case XED_OPERAND_ACTION_W:
            case XED_OPERAND_ACTION_CW:
              curr_def++;
              name = "dst";
              if (num_defs > 1) { name += std::to_string(curr_def); }
              break;
            case XED_OPERAND_ACTION_RW:
            case XED_OPERAND_ACTION_RCW:
            case XED_OPERAND_ACTION_CRW:
              curr_use_def++;
              name = "src-dst";
              if (num_use_defs > 1) { name += std::to_string(curr_use_def); }
              break;
            case XED_OPERAND_ACTION_INVALID:
            case XED_OPERAND_ACTION_LAST:
              abort();
          }
      }

      O << "      ";
      switch (op.value.kind()) {
        case Machine_Value::REG:
          O << "reg_op(";
          break;
        case Machine_Value::ADDR:
          O << "addr_op(";
          break;
        case Machine_Value::MEM:
          O << "mem_op(";
          break;
        case Machine_Value::IMM:
          O << "imm_op(";
          break;
        case Machine_Value::FLAGS:
          O << "flags_op(";
          break;
      }

      O << "name='" << name << "', ";

      switch (op.value.kind()) {
        case Machine_Value::REG:
          if (auto reg = op.value.reg()) {
            O << "reg=" << pipedream.register_name(*reg) << ", ";
          }
          O << "reg_class=RC." << pipedream.register_class_name(op.value.reg_class()) << ", ";
          O << "action=" << xed_operand_action_enum_t2str(op.action) << ", ";
          break;
        case Machine_Value::MEM:
          O << "addr_bits=" << op.value.address_bits() << ", ";
          O << "mem_bits=" << op.value.memory_bits() << ", ";
          if (auto reg = op.value.base_reg()) {
            O << "base=" << pipedream.register_name(*reg) << ", ";
          }
          O << "action=" << xed_operand_action_enum_t2str(op.action) << ", ";
          break;
        case Machine_Value::ADDR:
          O << "addr_bits=" << op.value.address_bits() << ", ";
          break;
        case Machine_Value::IMM:
          O << "imm_bits=" << op.value.imm_bits() << ", ";
          break;
        case Machine_Value::FLAGS:
          O << "reg=" << pipedream.register_name(op.value.flag_reg()) << ", ";
          O << "read="; pipedream.print_flag_set(op.value.flags_read(), O, "F."); O << ", ";
          O << "write="; pipedream.print_flag_set(op.value.flags_written(), O, "F."); O << ", ";
          break;
      }

      if (op.value.kind() != Machine_Value::FLAGS) {
        O << "type=" << pipedream.type_name(op.type) << ", ";
        O << "elems=" << op.num_elements << ", ";
      }
      O << "visibility=" << xed_operand_visibility_enum_t2str(op.visibility); // << ", ";

      O << "),\n";
    }
  }

  static void emit_pipedream_instruction_descriptor(std::ostream &O, const std::string &name, const Encode_Decode::Instruction &inst) {
    std::string isa_set = xed_isa_set_enum_t2str(inst.isa);
    assert(!isa_set.empty());
    if (std::isdigit(isa_set[0])) {
      isa_set = '_' + isa_set;
    }

    std::string isa_extension = xed_extension_enum_t2str(inst.extension);
    assert(!isa_extension.empty());
    if (std::isdigit(isa_extension[0])) {
      isa_extension = '_' + isa_extension;
    }

    O << "  make_instruction(\n";
    O << "    name                    = '" << name << "',\n";
    O << "    intel_mnemonic          = '" << inst.intel_mnemonic << "',\n";
    O << "    att_mnemonic            = '" << inst.att_mnemonic << "',\n";
//    O << "    cpu_mode                = " << inst.cpu_mode.mode_str() << ",\n";
//    O << "    stack_address_width     = " << inst.cpu_mode.stack_addr_width_bits() << ",\n";
//    O << "    effective_operand_width = " << inst.effective_operand_width << ",\n";
//    O << "    effective_address_width = " << inst.effective_address_width << ",\n";
//    O << "    is_nop                  = " << (inst.is_nop ? "True" : "False") << ",\n";
    O << "    isa_set                 = ISA." << isa_set << ",\n";
    O << "    isa_extension           = EXT." << isa_extension << ",\n";
    O << "    tags                    = [";
    {
      bool first = true;
      for (const std::string &tag : inst.tags) {
        if (!first) { O << ", "; }

        first = false;

        O << "'" << tag << "'";
      }
    }
    O << "],\n";
    O << "    operands                = [\n";
    emit_pipedream_instruction_operands(O, inst.operands);
    O << "    ],\n";
    O << "  )\n";
  }
};

int main() {
//  (void) Main::playground();
  return Main::main();
}
